!********************************************************************************
! Educationally-Designed Unstructured 3D (EDU3D) Code
!
!  --- EDU3D Poisson
!
! This module containes subroutines that compute the right state based on
! various boundary conditions. This provides a boundary state for computing
! a boundary flux for implementing weak boundary conditions.
!
! All subroutines are written in terms of the derivative data type
! (automatic differentiation). So, it returns the right state as well as
! the derivative dwR/dwj, where wR is the right state and wj is the interior
! state.
!
! Note: In the node-centered edge-based scheme, the boundary flux is computed
!       at a boundary node. So, we compute a flux with the nodal value, and
!       a right state at the same location. The right state is determined by
!       a boundary condition. That is exactly what is determined here.
!
!
!
!        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!
! the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!
! 
! This F90 program is written and made available for an educational purpose.
!
! This file may be updated in future.
!
! Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module bc_states

  implicit none

 !This module contains the following subroutines:

  public :: set_bc_state        !set BC values (ub,qnb).
  public :: get_right_state     !get the right state by bc

 contains

!********************************************************************************
!
! This subroutine sets the BC values: ub and/or qnb=ux*nx+uy*ny+uz*nz,
! which will be used in the subroutine get_right_state() to define the BC state.
!
!********************************************************************************

 subroutine set_bc_state(u,ux,uy,uz,ib,nb,jb,xb,yb,zb,  ub,qnb)

 use poisson_eqns     , only : exact_sol

 use input_parameter  , only : bc_dirichlet_user_values, &
                               bc_dirichlet            , &
                               bc_neumann_user_values  , &
                               bc_neumann              , &
                               bc_neumann_read_uxuyuz

 use data_module      , only : ux_bc, uy_bc, uz_bc

 implicit none

 integer , parameter :: p2 = selected_real_kind(15) ! Double precision

!Input
 real(p2),                  intent( in) :: u     !Interior solution value
 real(p2),                  intent( in) :: ux    !Interior x-derivative
 real(p2),                  intent( in) :: uy    !Interior y-derivative
 real(p2),                  intent( in) :: uz    !Interior z-derivative
 integer ,                  intent( in) :: ib    !Boundary part number
 integer ,                  intent( in) :: jb    !Boundary node number
 real(p2), dimension(   3), intent( in) :: nb    !Boundary normal vector
 real(p2),                  intent( in) :: xb    !x-corodinate of boundry pt
 real(p2),                  intent( in) :: yb    !y-corodinate of boundry pt
 real(p2),                  intent( in) :: zb    !z-corodinate of boundry pt

!Output
 real(p2),                  intent(out) :: ub    !BC solution value
 real(p2),                  intent(out) :: qnb   !BC normal gradient

!Local
 real(p2), dimension(4)                 :: ue


!-----------------------------------------------------------------------
! (1) Dirichlet condition with a provided value.
!

  if     ( bc_dirichlet_user_values ) then

           ub = bc_dirichlet(ib)                 !Will be used
          qnb = ux*nb(1) + uy**nb(2) + uz**nb(3) !Not used.

!-----------------------------------------------------------------------
! (2) Neumann condition with a user provided value.
!

  elseif ( bc_neumann_user_values   ) then 

          ub  = u               !Not used
          qnb = bc_neumann(ib)  !Will be used

!-----------------------------------------------------------------------
! (3) Neumann condition with user provided data files.
!     Gradient data are stored in ux_bc, uy_bc, uz_bc.
!

  elseif ( bc_neumann_read_uxuyuz   ) then

          ub  = u                                                    !Not used
          qnb = ux_bc(jb)*nb(1) + uy_bc(jb)*nb(2) + uz_bc(jb)*nb(3)  !Will be used

!-----------------------------------------------------------------------
! Otherwise, use the exact solution (a manufactured solution).
!

  else

     call exact_sol( xb, yb, zb, ue(1),ue(2),ue(3),ue(4) )

           ub = ue(1)                                   !Will be used if Dirichlet
          qnb = ue(2)*nb(1) + ue(3)*nb(2) + ue(4)*nb(3) !Will be used if Neumann

  endif

!
!
!-----------------------------------------------------------------------


 end subroutine set_bc_state


!********************************************************************************
! Compute the right state that incorporates a boundary condition.
!
! - Compute wR for given wj.
!
! - nj is the unit normal at the node j, which is used for slip condition.
!
! - This subroutines return both wR and duR/duj. The latter is requried to
!   compute the residual Jacobian exactly.
!
!********************************************************************************

 subroutine get_right_state(bc_state_type,uj,qb,dqdn_b,nj,  duRduj,uR)

 implicit none

 integer , parameter :: p2 = selected_real_kind(15) ! Double precision

!Input
 real(p2), dimension(   4), intent( in) :: uj
 real(p2),                  intent( in) :: qb, dqdn_b
 real(p2), dimension(   3), intent( in) :: nj
 character(80),             intent( in) :: bc_state_type

!Output
 real(p2), dimension(4,4), intent(out) :: duRduj
 real(p2), dimension(  4), intent(out) :: uR

!-----------------------------------------------------------------------
! Below, input is wL = the primitive variabes [rho,u,v,p].
! Output is the right state in wR = [rho,u,v,p] and the Jacobian duR/duL.

  select case(trim(bc_state_type))

   case('dirichlet','dirichlet_strong','dirichlet_strong_all')

     call dirichlet( uj,qb, duRduj,uR)     ! BC is given by qb.

   case('neumann')

     call neumann(uj,dqdn_b,nj, duRduj,uR) ! BC is given by dqdn_b.

   case default

     write(*,*) "Boundary condition=",trim(bc_state_type),"  not implemented."
     write(*,*) " --- Stop at get_right_state() in bc_states.f90..."
     stop

  end select
!-----------------------------------------------------------------------

 end subroutine get_right_state

!********************************************************************************
! Dirichlet with the desired boundary value given by qb_real
!********************************************************************************
 subroutine dirichlet(uL_real,qb_real,duRduL,uR_real)

 use derivative_data_df4

 implicit none

 integer , parameter :: p2 = selected_real_kind(15) ! Double precision
 real(p2), parameter :: two = 2.0_p2

 real(p2)                      , dimension(4)  , intent( in) :: uL_real
 real(p2)                                      , intent( in) :: qb_real !Desired value
 real(p2)                      , dimension(4,4), intent(out) :: duRduL
 real(p2)                      , dimension(4)  , intent(out) :: uR_real

 type(derivative_data_type_df4), dimension(4) :: uL, uR

 integer :: i, j

!-------------------------------------------------------------------------
!-------------------------
!--------------

  uL = uL_real
  call ddt_seed(uL)

!-------------------------------------------------------------------------
!-------------------------
!--------------

   uR    = uL

!  Set uR such that ( uR + uL )/2 = qb_real.

   uR(1) = two*qb_real - uL(1)

!-------------------------------------------------------------------------
!-------------------------
!--------------
! Output

 !(1)Right state

     uR_real = uR%f

 !(2)Derivative

    do i = 1, 4
     do j = 1, 4
      duRduL(i,j) = uR(i)%df(j)
     end do
    end do

 end subroutine dirichlet

!********************************************************************************
! Neumann condition with a specified normal gradient, qn_real.
!********************************************************************************
 subroutine neumann(uL_real,qn_real,nj, duRduL,uR_real)

 use derivative_data_df4

 implicit none

 integer , parameter ::  p2 = selected_real_kind(15) ! Double precision
 real(p2), parameter :: two = 2.0_p2

 real(p2)                      , dimension(4)  , intent( in) :: uL_real
 real(p2)                                      , intent( in) :: qn_real !desired normal grad.
 real(p2)                      , dimension(3)  , intent( in) :: nj
 real(p2)                      , dimension(4,4), intent(out) :: duRduL
 real(p2)                      , dimension(4)  , intent(out) :: uR_real

 type(derivative_data_type_df4), dimension(4) :: uL, uR

 type(derivative_data_type_df4)               :: dudn

 integer :: i, j
 integer :: ix =1, iy =2, iz =3
 integer :: iux=2, iuy=3, iuz=4

!-------------------------------------------------------------------------
!-------------------------
!--------------

  uL = uL_real
  call ddt_seed(uL)

!-------------------------------------------------------------------------
!-------------------------
!--------------

       uR = uL

    dudn  = uL(iux)*nj(ix) + uL(iuy)*nj(iy) + uL(iuz)*nj(iz)

!  Set the gradient on the right side such that the normal of the average
!   of the left and the right gradients is equal to the specified normal gradient.

    uR(iux) = uL(iux) + two*(qn_real-dudn) *nj(ix) !=du/dx
    uR(iuy) = uL(iuy) + two*(qn_real-dudn) *nj(iy) !=du/dy
    uR(iuz) = uL(iuz) + two*(qn_real-dudn) *nj(iz) !=du/dz

!  This could work also: fix the right state.
!
!    uR(2) = qn_real*nj(1) !=du/dx
!    uR(3) = qn_real*nj(2) !=du/dy
!    uR(4) = qn_real*nj(3) !=du/dz

!-------------------------------------------------------------------------
!-------------------------
!--------------
! Output

 !(1)Right state

     uR_real = uR%f

 !(2)Derivative

    do i = 1, 4
     do j = 1, 4
      duRduL(i,j) = uR(i)%df(j)
     end do
    end do

 end subroutine neumann




 end module bc_states
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
