!********************************************************************************
! Educationally-Designed Unstructured 3D (EDU3D) Code
!
!  --- EDU3D Poisson
!
!
! This module containes subroutines that computes numerical fluxes for the
! 3D Poisson equation. The following flux functions are implemented: 
!
!  - Upwind flux for the hyperbolized Poisson equation.
!  - Alpha-damping flux for the Poisson equation.
!
! Note: see edu3d_module_ddt5.f90 for details on how automatic differentiation
!       works.
!
!
! Note: These functions are written based on the derivative data type, meaning
!       that the flux Jacobian is also computed on background. These wil be
!       used also to compute the Jacobian for implicit solvers.
!
! Note: How to use these flux functions to compute the flux or the Jacobian
!       is described in the file edu3d_poisson_fluxjac_interfaces.f90, which
!       contains drivers subroutines to compute the flux and Jacobian.
!
!
!        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!
! the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!
! 
! This F90 program is written and made available for an educational purpose.
!
! This file may be updated in future.
!
! Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module flux_functions_ddt

 !This module contains the following subroutines:

  public ::     upwind_ddt
  public :: alpha_damp_ddt

 contains

!********************************************************************************
!
! -- 3D Upwind Hyperbolic Diffusion Flux Function .
!
! Note: This is an upwind flux for the hyperbolized Poisson equation with 4 eqs.
!
! See the following paper for details:
!
! H. Nishikawa, “A hyperbolic Poisson solver for tetrahedral grids”, Journal of
! Computational Physics, Volume 409, 15 May 2020, 109358, doi:10.1016/j.jcp.2020.109358.
!
!********************************************************************************
 subroutine upwind_ddt(uL, uR, njk, num_flux,dFdU,wsn)

 use derivative_data_df4

 use poisson_eqns   , only : diff_parameters_ddt
 use input_parameter, only : diffusion_type, Lr_factor
 use input_parameter, only : diss_factor

 implicit none
 integer , parameter :: p2 = selected_real_kind(15) ! Double precision

!Input
 type(derivative_data_type_df4), dimension(4), intent( in) :: uL
 type(derivative_data_type_df4), dimension(4), intent( in) :: uR
 real(p2)                      , dimension(3), intent( in) :: njk

!Output
 real(p2), dimension(4)  , intent(out) :: num_flux !Numerical viscous flux
 real(p2), dimension(4,4), intent(out) :: dFdU     !Numerical viscous flux Jacobian
 real(p2),                 intent(out) :: wsn      ! Max wave speed

!Some constants
 real(p2), parameter :: half = 0.50_p2

!Local variables

 type(derivative_data_type_df4) ::  nu
 type(derivative_data_type_df4) :: Lr, Tr
 type(derivative_data_type_df4) :: lambda

 type(derivative_data_type_df4), dimension(4) :: numerical_flux !Numerical flux in ddt
 type(derivative_data_type_df4), dimension(4) :: fL, fR
 type(derivative_data_type_df4), dimension(4) :: diss

 real(p2)                       :: nx, ny, nz             ! Normal vector components

 type(derivative_data_type_df4) :: temp
 type(derivative_data_type_df4) :: pnL, pnR

 integer                        :: i, j
 integer                        :: iu, iux, iuy, iuz

  iu  = 1
  iux = 2
  iuy = 3
  iuz = 4 

! Diffusion coefficient = 1 for Poisson.

  temp = half*( uL(iu) + uR(iu) ) !For a nonlinear problem, nu is compuoted with the averaged solution.

  call diff_parameters_ddt( diffusion_type,Lr_factor, temp,nu,Lr,Tr )

  lambda = nu / Lr !=sqrt(nu/Tr)

! Face normal vector (unit vector)

  nx = njk(1)
  ny = njk(2)
  nz = njk(3)

! Projected gradient variable (it includes the diffusion coefficient, nu)

  pnL = uL(iux)*nx + uL(iuy)*ny + uL(iuz)*nz
  pnR = uR(iux)*nx + uR(iuy)*ny + uR(iuz)*nz

!Compute the physical flux: fL = Fn(UL) and fR = Fn(UR)

  fL(1) = -pnL
  fL(2) = -uL(iu)*nx
  fL(3) = -uL(iu)*ny
  fL(4) = -uL(iu)*nz

  fR(1) = -pnR
  fR(2) = -uR(iu)*nx
  fR(3) = -uR(iu)*ny
  fR(4) = -uR(iu)*nz

! Dissipation term: P^{-1}*|PAn|*(UR-UL)

 diss(1) = diss_factor * lambda * ( uR(iu) - uL(iu) )
 diss(2) = diss_factor * lambda * (   pnR  -    pnL )*nx * Tr/nu
 diss(3) = diss_factor * lambda * (   pnR  -    pnL )*ny * Tr/nu
 diss(4) = diss_factor * lambda * (   pnR  -    pnL )*nz * Tr/nu

! This is the numerical flux: Roe flux = 1/2 *[  Fn(UL)+Fn(UR) - |An|(UR-UL) ]

  numerical_flux = half * (fL + fR - diss)

!--------------
! Output

   ! Normal max wave speed
    temp = lambda
     wsn = temp%f

  do i = 1, 4

   !Numerical flux
    num_flux(i) = numerical_flux(i)%f

   do j = 1, 4

     !Flux derivative
      dFdU(i,j) = numerical_flux(i)%df(j)

   end do

  end do

 end subroutine upwind_ddt
!--------------------------------------------------------------------------------

!********************************************************************************
!
! -- 3D Diffusion Flux Function based on the alpha-damping scheme.
!
! Note: This is to solver the Poisson equation, which is a scalar equation,
!       but it is implemented in a 4-squation format, for the sake of simplicity
!       in implementing both conventional and hyperbolic solvers in one code.
!
!       It would be more efficient to implement the conventional (alpha-damping)
!       solver with a single unknown: e.g., 1 residual and 1x1 Jacobian block 
!       instead of 4 and 4x4.
!
!       But remember that the hyperbolic Poisson solver has advantages of faster
!       convergence and accurate (one-order higher-order) gradients. The hyperbolic
!       solver is hihgly recommended for comutations on non-regular grids.
!
! See the following paper for details of the alpha-damping scheme:
!
! H. Nishikawa, Robust and Accurate Viscous Discretization via Upwind Scheme – I:
! Basic Principle, Computers and Fluids, Volume 49, Issue 1, pp. 62-86, 2011.
!
!********************************************************************************
 subroutine alpha_damp_ddt(uL, uR, njk, e_dot_n,ejk_mag, num_flux,dFdU,wsn)

 use derivative_data_df4

 use poisson_eqns   , only : diff_parameters_ddt
 use input_parameter, only : diffusion_type, Lr_factor, alpha

 implicit none
 integer , parameter :: p2 = selected_real_kind(15) ! Double precision

!Input
 type(derivative_data_type_df4), dimension(4), intent( in) :: uL
 type(derivative_data_type_df4), dimension(4), intent( in) :: uR
 real(p2)                      , dimension(3), intent( in) :: njk
 real(p2)                                    , intent( in) :: e_dot_n, ejk_mag

!Output
 real(p2), dimension(4)  , intent(out) :: num_flux !Numerical viscous flux
 real(p2), dimension(4,4), intent(out) :: dFdU     !Numerical viscous flux Jacobian
 real(p2),                 intent(out) :: wsn      ! Max wave speed

!Some constants
 real(p2), parameter :: zero = 0.0_p2
 real(p2), parameter :: half = 0.50_p2

!Local variables

 type(derivative_data_type_df4) ::  nu, Lr, Tr
 type(derivative_data_type_df4), dimension(4) :: numerical_flux !Numerical flux in ddt
 real(p2)                       :: nx, ny, nz             ! Normal vector components
 type(derivative_data_type_df4) :: temp
 type(derivative_data_type_df4) :: pnL, pnR

 integer                        :: i, j
 integer                        :: iu, iux, iuy, iuz

  iu  = 1
  iux = 2
  iuy = 3
  iuz = 4 

! Diffusion coefficient = 1 for Poisson.

  temp = half*( uL(iu) + uR(iu) )!For a nonlinear problem, nu is compuoted with the averaged solution.

  call diff_parameters_ddt( diffusion_type,Lr_factor, temp,nu,Lr,Tr )

! Face normal vector (unit vector)

   nx = njk(1)
   ny = njk(2)
   nz = njk(3)

! Projected gradient variable: These are already multiplied by nu in
! "subroutine compute_gradient".

  pnL = uL(iux)*nx + uL(iuy)*ny + uL(iuz)*nz
  pnR = uR(iux)*nx + uR(iuy)*ny + uR(iuz)*nz

  numerical_flux    = zero

               temp = half*e_dot_n*ejk_mag

  numerical_flux(1) = -half*( pnL + pnR + (alpha*nu/temp)*( uR(iu) - uL(iu) ) )

! Note: numerical_flux has nonzero value only in the 1st component because here
!       we are solving a scalar Poisson equation.

!--------------
! Output

   ! Normal max wave speed
    temp = alpha*nu/temp
     wsn = temp%f

  do i = 1, 4

   !Numerical flux
    num_flux(i) = numerical_flux(i)%f

   do j = 1, 4

     !Flux derivative
      dFdU(i,j) = numerical_flux(i)%df(j)

   end do

  end do

 end subroutine alpha_damp_ddt
!--------------------------------------------------------------------------------

 end module flux_functions_ddt
