!********************************************************************************
!* Educationally-Designed Unstructured 3D (EDU3D) Code
!*
!*  --- EDU3D Poisson
!*
!* This module containes subroutines that comptute numerical fluxes and flux
!* Jacobians by using flux functions written based on automatic differentiation.
!*
!*
!*        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!*
!* the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!*
!* 
!* This F90 program is written and made available for an educational purpose.
!*
!* This file may be updated in future.
!*
!* Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************
!********************************************************************************

 module fluxjac_interface

  implicit none

 !This module contains the following subroutines:

  public :: interface_flux   ! compute a numerical flux at interior face
  public :: interface_flux_b ! compute a numerical flux at boundary face

  public :: interface_jac    ! compute a flux Jacobian at interior face
  public :: interface_jac_b  ! compute a flux Jacobian at boundary face

 contains

!********************************************************************************
! Compute a numerical flux at an internal face.
!
! The numerical flux is computed at the edge midpoint between node j and k.
!
!               
!   ------------------> ejk(1:3)*ejk_mag:  This is the edge vector.
!
!   o--------x--------o
!   j       mid       k
!
!
! Left and right states linearly extrpolated from nodes to edge midpoint
! for 2nd-order accuracy. No reconstruction in 1st-order scheme.
!
!              |           . |           |
!              |        .  L |.R         |
!              |      .      |   .       |
!              |    .        |      .    |
!              | .           |        .  |
!              |             |           |
!         ------------o------x------o-------
!                     j     mid     k
!
! Note: the location of the face is always at the midpoint of the edge.
!
!********************************************************************************
 subroutine interface_flux(uj,uk,graduj,graduk,njk,ejk,ejk_mag,ar1,ar2, flux, wsn)

 use derivative_data_df4

 use data_module        , only : p2, half, ix, iy ,iz, two, one
 use input_parameter    , only : flux_choice, hyp_scheme_type
 use flux_functions_ddt , only :   upwind_ddt, alpha_damp_ddt
 use input_parameter    , only : accuracy_order, kappa

 use input_parameter    , only : diffusion_type, Lr_factor
 use poisson_eqns       , only : diff_parameters

 implicit none

!Input
 real(p2), dimension(4)  , intent( in) ::     uj  !primitive variables at j
 real(p2), dimension(4)  , intent( in) ::     uk  !primitive variables at k
 real(p2), dimension(4,3), intent( in) :: graduj  !gradient at j
 real(p2), dimension(4,3), intent( in) :: graduk  !gradient at k
 real(p2), dimension(3)  , intent( in) :: njk     !unit drected area vector
 real(p2), dimension(3)  , intent( in) :: ejk     !unit edge vector
 real(p2)                , intent( in) :: ejk_mag !magnitude of the edge vector
 real(p2)                , intent( in) :: ar1, ar2

!Output
 real(p2), dimension(4)  , intent(out) :: flux !output: numerical flux
 real(p2)                , intent(out) :: wsn  !output: eigenvalue to be used for dtau

!Left and right states
 real(p2)                      , dimension(4  ) :: uL ,uR
 type(derivative_data_type_df4), dimension(4  ) :: uL_ddt, uR_ddt

!Dummy array to store Jacobian (computed but not used here).
 real(p2),                       dimension(4,4) :: dummy4x4

 real(p2) :: e_dot_n, kappa_loc
 real(p2)                      , dimension(4  ) :: dujk, duL, duR
 real(p2) :: nuL,nuR,Lr,Tr 

  e_dot_n = abs( ejk(ix)*njk(ix) + ejk(iy)*njk(iy) + ejk(iz)*njk(iz) )

  if (kappa < 0.0_p2) then

   if ( max(ar1,ar2) < abs(kappa) ) then
    kappa_loc = 0.0_p2
   else
    kappa_loc = 0.5_p2
   endif

  else

    kappa_loc = kappa

  endif

!-------------------------------------------------------------------------------
! Left and right solutions at the midpoint:

  if     (accuracy_order == 1) then

     uL = uj
     uR = uk

   if (hyp_scheme_type == 2) then

     call diff_parameters( diffusion_type,Lr_factor,uj(1), nuL,Lr,Tr )

     call diff_parameters( diffusion_type,Lr_factor,uk(1), nuR,Lr,Tr )

!     uL(1) = uL(1) + ( uL(2)*ejk(ix) + uL(3)*ejk(iy) + uL(4)*ejk(iz) )*ejk_mag*half
!     uR(1) = uR(1) - ( uR(2)*ejk(ix) + uR(3)*ejk(iy) + uR(4)*ejk(iz) )*ejk_mag*half

      dujk(1) = half*( uk(1) - uj(1) )
      duL(1)  =  ( uL(2)*ejk(ix) + uL(3)*ejk(iy) + uL(4)*ejk(iz) )/nuL*ejk_mag*half
      duR(1)  = -( uR(2)*ejk(ix) + uR(3)*ejk(iy) + uR(4)*ejk(iz) )/nuR*ejk_mag*half

        uL(1) = uj(1) + ( (one-kappa_loc)*(two*  duL(1) -dujk(1)) + (one+kappa_loc)*dujk(1) )*half
        uR(1) = uk(1) - ( (one-kappa_loc)*(two*(-duR(1))-dujk(1)) + (one+kappa_loc)*dujk(1) )*half

   endif

  elseif (accuracy_order == 2) then

   !Reconstruction to the edge midpoint for 2nd-order accuracy:
   !      See http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_aiaa2015-2451.pdf

!     uL = uj + ( graduj(:,ix)*ejk(ix) + graduj(:,iy)*ejk(iy) + graduj(:,iz)*ejk(iz) )*ejk_mag*half
!     uR = uk - ( graduk(:,ix)*ejk(ix) + graduk(:,iy)*ejk(iy) + graduk(:,iz)*ejk(iz) )*ejk_mag*half

      dujk = half*( uk - uj )
      duL  =  ( graduj(:,ix)*ejk(ix) + graduj(:,iy)*ejk(iy) + graduj(:,iz)*ejk(iz) )*ejk_mag*half
      duR  = -( graduk(:,ix)*ejk(ix) + graduk(:,iy)*ejk(iy) + graduk(:,iz)*ejk(iz) )*ejk_mag*half

        uL = uj + ( (one-kappa_loc)*(two*  duL -dujk) + (one+kappa_loc)*dujk )*half
        uR = uk - ( (one-kappa_loc)*(two*(-duR)-dujk) + (one+kappa_loc)*dujk )*half

  !This is SchemeII
   if (hyp_scheme_type == 2) then

     call diff_parameters( diffusion_type,Lr_factor,uj(1), nuL,Lr,Tr )
     call diff_parameters( diffusion_type,Lr_factor,uk(1), nuR,Lr,Tr )

   !Note: gradient variables are nu*(ux,uy,uz). Should be divided by nu.
!     uL(1) = uj(1) + ( uj(2)*ejk(ix) + uj(3)*ejk(iy) + uj(4)*ejk(iz) )*ejk_mag*half
!     uR(1) = uk(1) - ( uk(2)*ejk(ix) + uk(3)*ejk(iy) + uk(4)*ejk(iz) )*ejk_mag*half

      dujk(1) = half*( uk(1) - uj(1) )
      duL(1)  =  ( uj(2)*ejk(ix) + uj(3)*ejk(iy) + uj(4)*ejk(iz) )/nuL*ejk_mag*half
      duR(1)  = -( uk(2)*ejk(ix) + uk(3)*ejk(iy) + uk(4)*ejk(iz) )/nuR*ejk_mag*half

        uL(1) = uj(1) + ( (one-kappa_loc)*(two*  duL(1) -dujk(1)) + (one+kappa_loc)*dujk(1) )*half
        uR(1) = uk(1) - ( (one-kappa_loc)*(two*(-duR(1))-dujk(1)) + (one+kappa_loc)*dujk(1) )*half

   endif

  else

   write(*,*) " Invalid input: accuracy_order = ", accuracy_order
   write(*,*) "  accuracy_order must be 1 or 2. Stop..."
   stop

  endif

!-------------------------------------------------------------------------------
! Covert them to conservative variables in ddt.
! Here, ddt is not really used, but needed since numerical fluxes
! are written based on ddt variables for easy computation of Jacobian.

   uL_ddt = uL
   uR_ddt = uR

!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------
!  Compute a numerical flux by 3D flux subroutines.
!
!  Note: These flux subroutines are written based on automatic differentiation,
!        and thus return the flux derivative also. But the derivative is not
!        used here, and so it is received in 'dummy5x5', and not used.
!
!  Note: Input variables to each flux function must be derivative-data-type (ddt),
!        which contains derivative information as defined in the module
!        'derivative_data_df5'.
!
!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------

 !------------------------------------------------------------
 !  (1) Roe flux
 !------------------------------------------------------------
   if         (trim(flux_choice)=="hyp_upwind") then

     call upwind_ddt(uL_ddt,uR_ddt,njk, flux,dummy4x4,wsn)

 !------------------------------------------------------------
 !  Alpha damping
 !------------------------------------------------------------
   elseif     (trim(flux_choice)=="alpha") then

     call alpha_damp_ddt(uL_ddt,uR_ddt,njk,e_dot_n,ejk_mag, flux,dummy4x4,wsn)

 !------------------------------------------------------------
 !  Others...
 !------------------------------------------------------------
   else

    write(*,*) " Invalid input for flux_choice = ", trim(flux_choice)
    write(*,*) " Choose hyp_upwind or alpha, and try again."
    write(*,*) " ... Stop."
    stop

   endif

 end subroutine interface_flux
!********************************************************************************


!********************************************************************************
! Compute a numerical flux at a boundary node.
!
! In node-centered edge-based method, a boundary flux is computed at a node.
! This subroutine computes the flux based on the nodal value wj, and a boundary
! state given in wk (conceptually located at the same location as j).
! There is no edge here.
!
! The Roe flux is used to take into account characteristics at boundary.
!
!********************************************************************************
 subroutine interface_flux_b(uj,uk,njk,length, flux, wsn)

 use derivative_data_df4
 use data_module           , only : p2
 use flux_functions_ddt    , only : upwind_ddt, alpha_damp_ddt
 use input_parameter       , only : flux_choice

 implicit none

!Input
 real(p2), dimension(4)  , intent( in) :: uj  !primitive variables at j
 real(p2), dimension(4)  , intent( in) :: uk  !primitive variables at k
 real(p2), dimension(3)  , intent( in) :: njk !unit boundary element normal vector
 real(p2)                , intent( in) :: length !alternarive to the edge length

!Output
 real(p2), dimension(4)  , intent(out) :: flux !output
 real(p2)                , intent(out) :: wsn  !output

!Left and right states
 type(derivative_data_type_df4), dimension(4  ) :: uL_ddt, uR_ddt

!Dummy array to store Jacobian (computed but not used here).
 real(p2),                       dimension(4,4) :: dummy4x4

 !Covert them to conservative variables in ddt.
 !Here, ddt is not really used, but needed since numerical fluxes
 !are written based on ddt variables for easy computation of Jacobian.

   uL_ddt = uj
   uR_ddt = uk

!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------
! Use the Roe flux at boundary to implement a characteristic boundary flux.
!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------

 !------------------------------------------------------------
 !  Upwind
 !------------------------------------------------------------
   if         (trim(flux_choice)=="hyp_upwind") then

     call upwind_ddt(uL_ddt,uR_ddt,njk, flux,dummy4x4,wsn)

 !------------------------------------------------------------
 !  Alpha damping
 !------------------------------------------------------------
   elseif    (trim(flux_choice)=="alpha") then

     call alpha_damp_ddt(uL_ddt,uR_ddt,njk,1.0_p2,length, flux,dummy4x4,wsn)

 !------------------------------------------------------------
 !  Others...
 !------------------------------------------------------------
   else

    write(*,*) " Invalid input for flux_choice = ", trim(flux_choice)
    write(*,*) " Choose hyp_upwind or alpha, and try again."
    write(*,*) " ... Stop."
    stop

   endif

 end subroutine interface_flux_b
!********************************************************************************



!********************************************************************************
! Compute two Jacobians for an internal face:
!
!   dFnduL = d(numerical flux)/d(uj)
!   dFnduR = d(numerical flux)/d(uk)
!
! Flux Jacobians are computed at the edge midpoint between node j and k.
!
!               
!   ------------------> ejk(1:3)*ejk_mag:  This is the edge vector.
!
!   o--------x--------o
!   j       mid       k
!
! No reconstruction for Jacobian, whcih we compute exactly for 1st-order scheme.
!
!********************************************************************************
 subroutine interface_jac(uj,uk,njk,ejk,ejk_mag,ar1,ar2, dFnduL, dFnduR )

 use derivative_data_df4
 use data_module           , only : p2, half, ix, iy ,iz, two, one
 use input_parameter       , only : jacobian_choice, hyp_jac_type, kappa
 use flux_functions_ddt    , only : upwind_ddt, alpha_damp_ddt

 use input_parameter       , only : diffusion_type, Lr_factor
 use poisson_eqns          , only : diff_parameters_ddt

 implicit none

!Input
 real(p2), dimension(4)  , intent( in) :: uj, uk
 real(p2), dimension(3)  , intent( in) :: njk
 real(p2), dimension(3)  , intent( in) :: ejk     !unit edge vector
 real(p2)                , intent( in) :: ejk_mag !magnitude of the edge vector
 real(p2)                , intent( in) :: ar1, ar2

!Output
 real(p2), dimension(4,4), intent(out) :: dFnduL, dFnduR

!Left and right states

 real(p2), dimension(4,4) :: dfndu
 real(p2), dimension(4  ) :: dummy4
 real(p2)                 :: wsn

 integer :: i
 type(derivative_data_type_df4), dimension(4) :: uL_ddt, uR_ddt

 real(p2) :: e_dot_n, kappa_loc
 type(derivative_data_type_df4), dimension(4  ) :: dujk, duL, duR
 type(derivative_data_type_df4) :: nuL,nuR,Lr,Tr 

  e_dot_n = abs( ejk(ix)*njk(ix) + ejk(iy)*njk(iy) + ejk(iz)*njk(iz) )

  if (kappa < 0.0_p2) then

   if ( max(ar1,ar2) < abs(kappa) ) then
    kappa_loc = 0.0_p2
   else
    kappa_loc = 0.5_p2
   endif

  else

    kappa_loc = kappa

  endif

  jac_L_R : do i = 1, 2

  !Covert them to conservative variables in ddt.
  !Here, ddt is not really used, but needed since numerical fluxes
  !are written based on ddt variables for easy computation of Jacobian.

    uL_ddt = uj
    uR_ddt = uk

  !This step determines with respect to which state the flux derivative is computed.

    if (i==1) then

     !This sets the derivative of uL_ddt = 1.
      call ddt_seed(uL_ddt)

    else

     !This sets the derivative of uR_ddt = 1.
      call ddt_seed(uR_ddt)

    endif

 
 !This is SchemeII. If true, we will difrerentiate the kappa reconstruction scheme.
   if (hyp_jac_type == 2) then

     call diff_parameters_ddt( diffusion_type,Lr_factor,uL_ddt(1), nuL,Lr,Tr )
     call diff_parameters_ddt( diffusion_type,Lr_factor,uR_ddt(1), nuR,Lr,Tr )

!    uL_ddt(1) = uL_ddt(1) + ( uL_ddt(2)*ejk(ix) + uL_ddt(3)*ejk(iy) + uL_ddt(4)*ejk(iz) )*ejk_mag*half
!    uR_ddt(1) = uR_ddt(1) - ( uR_ddt(2)*ejk(ix) + uR_ddt(3)*ejk(iy) + uR_ddt(4)*ejk(iz) )*ejk_mag*half

      dujk(1) = half*( uR_ddt(1) - uL_ddt(1) )
      duL(1)  =  ( uL_ddt(2)*ejk(ix) + uL_ddt(3)*ejk(iy) + uL_ddt(4)*ejk(iz) )/nuL*ejk_mag*half
      duR(1)  = -( uR_ddt(2)*ejk(ix) + uR_ddt(3)*ejk(iy) + uR_ddt(4)*ejk(iz) )/nuR*ejk_mag*half

     uL_ddt(1) = uL_ddt(1) + ( (one-kappa_loc)*(two*  duL(1) -dujk(1)) + (one+kappa_loc)*dujk(1) )*half
     uR_ddt(1) = uR_ddt(1) - ( (one-kappa_loc)*(two*(-duR(1))-dujk(1)) + (one+kappa_loc)*dujk(1) )*half

   endif

!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------
!  Compute inviscid Jacobian by 3D flux subroutines.
!
!  Note: These flux subroutines are written based on automatic differentiation,
!        and thus return the flux and its derivative. Here, we only want the
!        derivative. So, the flux is received in 'dummy5', and not used.
!
!  Note: Input variables to each flux function must be derivative-data-type (ddt),
!        which contains derivative information as defined in the module
!        'derivative_data_df5'.
!
!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------

 !------------------------------------------------------------
 !  Upwind
 !------------------------------------------------------------
   if         (trim(jacobian_choice)=="hyp_upwind") then

     call upwind_ddt(uL_ddt,uR_ddt,njk, dummy4,dfndu,wsn)

 !------------------------------------------------------------
 !  Alpha damping
 !------------------------------------------------------------
   elseif     (trim(jacobian_choice)=="alpha") then

     call alpha_damp_ddt(uL_ddt,uR_ddt,njk,e_dot_n,ejk_mag, dummy4,dfndu,wsn)

 !------------------------------------------------------------
 !  Others...
 !------------------------------------------------------------
   else

    write(*,*) " Invalid input for jacobian_choice = ", trim(jacobian_choice)
    write(*,*) " Choose hyp_upwind or alpha, and try again."
    write(*,*) " ... Stop."
    stop

   endif

   if (i==1) then
    dFnduL = dfndu
   else
    dFnduR = dfndu
   endif

  end do jac_L_R

 end subroutine interface_jac
!********************************************************************************


!********************************************************************************
! Compute two Jacobians for the flux defined at at a boundary node:
!
!   dFnduL = d(numerical flux)/d(uj)
!   dFnduR = d(numerical flux)/d(uk)
!
! In node-centered edge-based method, a boundary flux is computed at a node.
! This subroutine computes the flux based on the nodal value wj, and a boundary
! state given in wk (conceptually located at the same location as j).
! There is no edge here.
!!
!********************************************************************************
 subroutine interface_jac_b(uj,uk,njk,length, dFnduL, dFnduR )

 use derivative_data_df4
 use data_module           , only : p2
 use flux_functions_ddt    , only : upwind_ddt, alpha_damp_ddt
 use input_parameter       , only : jacobian_choice

 implicit none

!Input
 real(p2), dimension(4)  , intent( in) :: uj  !primitive variables at j
 real(p2), dimension(4)  , intent( in) :: uk  !primitive variables at k
 real(p2), dimension(3)  , intent( in) :: njk !unit boundary element normal vector
 real(p2)                , intent( in) :: length

!Output
 real(p2), dimension(4,4), intent(out) :: dFnduL, dFnduR !Left and right Jacobians

!Left and right states

 real(p2), dimension(4,4) :: dfndu
 real(p2), dimension(4  ) :: dummy4
 real(p2)                 :: wsn

 integer :: i
 type(derivative_data_type_df4), dimension(4) :: uL_ddt, uR_ddt

  jac_L_R : do i = 1, 2

  !Here, ddt is not really used, but needed since numerical fluxes
  !are written based on ddt variables for easy computation of Jacobian.

    uL_ddt = uj
    uR_ddt = uk


    if (i==1) then
     !This sets the derivative of uL_ddt = 1.
     call ddt_seed(uL_ddt)
    else
     !This sets the derivative of uR_ddt = 1.
     call ddt_seed(uR_ddt)
    endif

!---------------------------------------------------------------------------------
!---------------------------------------------------------------------------------

 !------------------------------------------------------------
 !  Upwind
 !------------------------------------------------------------
   if         (trim(jacobian_choice)=="hyp_upwind") then

     call upwind_ddt(uL_ddt,uR_ddt,njk, dummy4,dfndu,wsn)

 !------------------------------------------------------------
 !  Alpha damping
 !------------------------------------------------------------
   elseif    (trim(jacobian_choice)=="alpha") then

     call alpha_damp_ddt(uL_ddt,uR_ddt,njk,1.0_p2,length, dummy4,dfndu,wsn)

 !------------------------------------------------------------
 !  Others...
 !------------------------------------------------------------
   else

    write(*,*) " Invalid input for jacobian_choice = ", trim(jacobian_choice)
    write(*,*) " Choose hyp_upwind or alpha, and try again."
    write(*,*) " ... Stop."
    stop

   endif


   if (i==1) then
    dFnduL = dfndu
   else
    dFnduR = dfndu
   endif

  end do jac_L_R

 end subroutine interface_jac_b
!********************************************************************************


 end module fluxjac_interface

