!********************************************************************************
! Educationally-Designed Unstructured 3D (EDU3D) Code
!
!  --- EDU3D Poisson
!
! This module defines input parameters, and set the default values.
! These parameters are specified in the file named 'input.nml', and read
! in the main program. 
!
!
!        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!
! the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!
! 
! This F90 program is written and made available for an educational purpose.
!
! This file may be updated in future.
!
! Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module input_parameter

  implicit none

 !This module contains the following subroutine:

  public :: read_nml_input_parameters ! read the input file
  public

  integer , parameter ::     p2 = selected_real_kind(P=15)

!----------------------------
! Default input values
!----------------------------

  character(80) :: project              = "default"   ! project name

  integer       ::   diffusion_type     = 1           !=1 Linear Poisson, =10 Nonlinear Poisson

  logical       :: accuracy_test        = .false.     !=.true. to perform accuracy test.
  logical       :: accuracy_test_te     = .false.     !=.true. to perform truncation errro test.
  integer       :: accuracy_test_ngrids = 2           !# of grids used in the test.

  logical                  :: bc_dirichlet_user_values = .false.      !=.true. use the user's values
  real(p2), dimension(100) :: bc_dirichlet             = 100*(1.0_p2) !user's values up to 100 boundary parts.

  logical                  :: bc_neumann_user_values   = .false.      !=.true. use the user's values
  real(p2), dimension(100) :: bc_neumann               = 100*(1.0_p2) !user's values up to 100 boundary parts.

  logical                  :: bc_neumann_read_uxuyuz   = .false.      !=.true. read the user-provided files
  character(80)            :: bc_neumann_file_ux       = "none"       !a file contains x-derivatives at all nodes.
  character(80)            :: bc_neumann_file_uy       = "none"       !a file contains y-derivatives at all nodes.
  character(80)            :: bc_neumann_file_uz       = "none"       !a file contains z-derivatives at all nodes.

  integer       :: single_dirichlet_node     = -1      !The node number where a solution is fixed.
  real(p2)      :: single_dirichlet_value    = 0.0_p2  !The value specified at the above node.


  character(80) :: flux_choice          = "hyp_upwind" ! numerical flux: hyp_upwind or alpha
  character(80) :: jacobian_choice      = "hyp_upwind" ! flux Jacobian : hyp_upwind or alpha

  real(p2)      ::               alpha  = 1.0_p2       ! Alpha in the alpha-damping scheme,

  integer       :: accuracy_order       = 2            ! 1=1st-order, 2=2nd-order
  real(p2)      :: lsq_weight_power     = 0.0_p2       ! lsq weight: 1/distance^lsq_weight_power


  real(p2)      ::          diss_factor = 1.0_p2       !may set diss_factor < 1.0 to reduce dissipation.
  integer       ::  hyp_scheme_type     = 2            !=1 SchemeI (more robust), =2 SchemeII (more accurate)
  integer       ::  hyp_jac_type        =-1            !=1 SchemeI (more robust), =2 SchemeII (more accurate)
  logical       ::  hyp_jac_hand_coded  = .false.      !=T use a hand-coded Jacobian (hyperbolic scheme only), =F not

  real(p2)      ::            kappa     = 0.5_p2       !Kappa in the kappa reconstruciton scheme (0.5 is recommended).
  real(p2)      ::          Lr_factor   = 1.0_p2       !Adjust as you like
  real(p2)      ::  ref_length_factor   = 1.0_p2       !Adjust as you like: ref_length * ref_length_factor
  real(p2)      ::  ref_length_in       = -1.0         !< -0.9 Lopt, = 0.0 use Lopt for rectangular, > 0 use this value.

! Parameters to modify the input grid's nodal coordinates (x,y,z):

  real(p2)      ::    scale_factor     = 1.0_p2 !(x,y,z) -> ( scale_factor*x, scale_factor*y, scale_factor*z )
  real(p2)      ::         x_shift     = 0.0_p2 !(x,y,z) -> ( x-x_shift,y        ,z        ))
  real(p2)      ::         y_shift     = 0.0_p2 !(x,y,z) -> ( x        ,y-y_shift,z        ))
  real(p2)      ::         z_shift     = 0.0_p2 !(x,y,z) -> ( x        ,y        ,z-z_shift))

! Parameters for the exact solution
  real(p2)      ::        exact_c1     = 1.0_p2 !c1 of exact sol = c4 * sin( pi*(c1*x + c2*y + c3*z) )
  real(p2)      ::        exact_c2     = 1.0_p2 !c2 of exact sol = c4 * sin( pi*(c1*x + c2*y + c3*z) )
  real(p2)      ::        exact_c3     = 1.0_p2 !c3 of exact sol = c4 * sin( pi*(c1*x + c2*y + c3*z) )
  real(p2)      ::        exact_c4     = 1.0_p2 !c4 of exact sol = c4 * sin( pi*(c1*x + c2*y + c3*z) )

! Nonlinear solver (steady solver)
  character(80) :: solver_type         = "implicit"   ! solver = "fwd_euler"
                                                      !          "tvdrk"
                                                      !          "implicit"
                                                      !          "jfnk"

  integer       :: perturb_initial     = 1            ! 0=unperturbed, 1=perturbed initial solution.

  real(p2)      :: solver_tolerance    = 1.0e-10_p2   ! solver: tolerance for nonlinear solve

  integer, dimension(10) :: solver_max_itr = 10*1     ! solver: max nonlinear iterations (1*10 -> 10 iterations for 10 grids).

  real(p2)      :: CFL                 = 1.0e+15_p2   ! CFL for pseudo time. Infinite for implicit and jfk solvers.

! JFNK-GCR paramters
  integer       :: gcr_projection      = 5       ! max GCR projections
  real(p2)      :: gcr_mu              = 0.5_p2  ! tolerance for GCR (reduction)
  integer       :: gcr_scale           = 1       !=1 Scale invariant implementation

! A parameter for nonlinear updates
  real(p2)      ::    nonlinear_omega  = 1.0_p2  ! nonlinear relaxation factor

! Linear relaxation in the implicit solver, which is a preconditioner solver in the JFNK solver.
  character(80) :: lrelax_scheme       = "gs"    ! preconditioner scheme type
  integer       :: lrelax_sweeps       = 500     ! preconditioner max relaxation
  real(p2)      :: lrelax_tolerance    = 0.5_p2  ! preconditioner tolerance (reduction)
  real(p2)      :: lrelax_omega        = 1.0_p2  ! relaxation factor (intitial)
  real(p2)      :: lrelax_omega_min    = 0.05_p2 ! minimum of relaxation factor
  real(p2)      :: lrelax_omega_max    = 1.0_p2  ! minimum of relaxation factor
  logical       :: lrelax_omega_fix    = .false. !=.true. keep the initial value.
  logical       :: lrelax_use_slope    = .true.  !=.true. adjust omega based on the linear residual history.

! Data alteration
  character(80) :: soln_normalization  = "none"  ! "min_2norm" to minimize the norm of sol.

! Data files
  logical       :: ugrid_unformatted   = .true.  ! ugrid_file_unformatted = T: unformatted, F: formatted
  logical       :: generate_tec_file_b = .true.  ! generate_tec_file_b = T
  logical       :: generate_tec_file_v = .false. ! generate_tec_file_v = T
  logical       :: generate_soln_file  = .false. ! generate_sol_file   = T

!----------------------------
! End of Default input values
!----------------------------

 namelist / input_parameters /   &
!
   project                 , &
   diffusion_type          , &
   accuracy_test           , &
   accuracy_test_te        , &
   accuracy_test_ngrids    , &
   bc_dirichlet_user_values, &
   bc_dirichlet            , &
   bc_neumann_user_values  , &
   bc_neumann              , &
   bc_neumann_read_uxuyuz  , &
   bc_neumann_file_ux      , &
   bc_neumann_file_uy      , &
   bc_neumann_file_uz      , &
   single_dirichlet_node   , &
   single_dirichlet_value  , &
   flux_choice             , &
   jacobian_choice         , &
   accuracy_order          , &
   lsq_weight_power        , &
   alpha                   , &
   diss_factor             , &
   hyp_scheme_type         , &
   hyp_jac_type            , &
   hyp_jac_hand_coded      , &
   kappa                   , &
   Lr_factor               , &
   ref_length_factor       , &
   ref_length_in           , &
   scale_factor            , &
   x_shift                 , &
   y_shift                 , &
   z_shift                 , &
   exact_c1                , &
   exact_c2                , &
   exact_c3                , &
   exact_c4                , &
   solver_type             , &
   perturb_initial         , &
   solver_tolerance        , &
   solver_max_itr          , &
   CFL                     , &
   gcr_projection          , &
   gcr_mu                  , &
   gcr_scale               , &
   nonlinear_omega         , &
   lrelax_scheme           , &
   lrelax_sweeps           , &
   lrelax_tolerance        , & 
   lrelax_omega            , &
   lrelax_omega_min        , &
   lrelax_omega_max        , &
   lrelax_omega_fix        , &
   lrelax_use_slope        , &
   soln_normalization      , &
   ugrid_unformatted       , &
   generate_tec_file_b     , &
   generate_tec_file_v     , &
   generate_soln_file

 contains

!*****************************************************************************
!
! This subroutine reads input_parameters specified in the input file:
!
!    file name = namelist_file
!
! prints the content on screen.
!
! In the main program, we set: namelist_file = "input.nml"
!
!*****************************************************************************

  subroutine read_nml_input_parameters(namelist_file)

  implicit none
  character(9), intent(in) :: namelist_file
  integer :: os

  write(*,*) "**************************************************************"
  write(*,*) " List of namelist variables and their values"
  write(*,*)

  open(unit=10,file=trim(namelist_file),form='formatted',status='old',iostat=os)
  read(unit=10,nml=input_parameters)

  if (hyp_jac_type < 0) hyp_jac_type = hyp_scheme_type

  write(*,nml=input_parameters) ! Print the namelist variables.
  close(10)

  end subroutine read_nml_input_parameters
!*****************************************************************************

 end module input_parameter
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!
!  End of input parameter module
