!********************************************************************************
!* Educationally-Designed Unstructured 3D (EDU3D) Code
!*
!*  --- EDU3D Poisson
!*
!* This module containes subroutines that comptutes the residual Jacobian
!* for the node-centered edge-based discretization. The Jacobian is exact
!* for the first-order node-centered edge-based discretization. So, Newton's
!* method can be constructed for the first-order scheme. This Jacobian can serve
!* as a preconditioning matrix for Jacobian-Free Newton-Krylov solver for
!* higher-order edge-based schemes.
!*
!*
!*        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!*
!* the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!*
!* 
!* This F90 program is written and made available for an educational purpose.
!*
!* This file may be updated in future.
!*
!* Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module jacobian

  implicit none

 !This module contains the following subroutine:

  public :: compute_jacobian ! compute the jacobian
  public :: kth_nghbr        ! find a nghbr index
  public :: gewp_solve       ! Gauss elimination to invert the diagonal blocks

 contains

!********************************************************************************
! Compute the residual Jacobian exactly for the first-order residual.
!
! - Jac = [V/dtau+dRes1/dU] for the linear system: Jac*dU = -Residual.
!
!    where Res1     = 1st-order accurate residual (compact stencil),
!          Residual = 1st/2nd-order residual.
!          V/dtau   = global diagonal matrix of dual_vol/dtau.
!
! - This is the linear system that needs to be solved per nonlinear iteration.
! - This is very similar to the residual subroutine.
! - Exact flux Jacobian will be computed by Automatic Differentiation.
! - Diagonal blocks are inverted and stored at the end.
!
! - Off-diagonal (4x4) blocks are stored as follows:
!
!     node1 with 5 nghbrs : jac_off( 1,:,:), jac_off( 2,:,:), ..., jac_off( 5,:,:),
!     node2 with 6 nghbrs : jac_off( 6,:,:), jac_off( 7,:,:), ..., jac_off(11,:,:),
!     node3 with 4 nghbrs : jac_off(12,:,:), jac_off(13,:,:), ..., jac_off(15,:,:),
!     etc.
!
!    Loop over nghbrs can be performed by using nnghrs array, which has the
!    number of nghbr nodes for each node: e.g.,
!
!      nnghbrs(node1)=5, nnghbrs(node2)=6, nnghbrs(node3)=4, etc,
!
!    and k0 array, which has the starting index for each node:
!
!      k0(node1)=0, k0(node2)=5, k0(node3)=11, etc.
! 
!    The off-diagonal blocks can be accessed at a node j as
!
!     do k = 1, nnghbrs(j)
!      jac_off(i,k0(j)+k,:,:)
!     end do
!
! - The Jacobian matirx is stored in two parts: diagonal and off-diagonal blocks.
!
!       For example, consider the residual at node j, Rj, which is a vector,
!       and suppose j has 5 neighbor nodes k=(1, 2, 3, 4, 5).
!       For implicit defect-correction solver, the residual is taken as 1st-order scheme,
!       so that Res_j is a function of Uj, U1, U2, U3, U4, and U5 (compact stencil).
!
!       Then, the j-th component of [V/dt+dR/dU]*dU = -Residual is
!
!       [        dRj/dU1]*dU1 + [dRj/dU2]*dU2 + [dRj/dU3]*dU3 +
!       [V/dtauj+dRj/dUj]*dUj + [dRj/dU4]*dU4 + [dRj/dU5]*dU5 = - Rj
!
!       where dtauj is the local time step at j, dRj/dUj is the derivative of Rj
!       with respect to the solution at j, i.e., Uj, and similarly for others.
!       The Gauss-Seidel relaxation can be written as
!
!       dUj = [V/dtj+dR/dUj]^{-1}*(- [dRj/dU1]*dU1 - [dRj/dU2]*dU2
!                                  - [dRj/dU3]*dU3 - [dRj/dU4]*dU4
!                                  - [dRj/dU5]*dU5 - Res_j )
!
!       To perform this operation, we store two types of data:
!
!        1.  5x5     Diagonal block:  jac_diag(j, :,:) = [V/dtauj+dRj/dUj]
!        2.  5x5 Off-Diagonal Block:  jac_off(j,k,:,:) = [        dRj/dUk], k = 1,2,3,4,5
!
!       so, we can perform the relaxation at every node by looping over the neighbors:
!
!         b = -Res_j
!        do k = 1, nnghbrs(j)
!         b = b - jac_off(i,k0(i)+k,:,:)*dU( nghbr(k0(i)+k), :)
!        end do
!         dUj = jac_diag(j, :,:)^{-1}*b,
!
!       where nghbr(k0(i)+k) stores the actual node number for kth nghbr.
!
!       Actual relaxation will be performed in edu3d_linear_solver.f90.
!
!       To compute dR/dUj and dR/dUk, just like the residual is computed as a sum of
!       numerical fluxes over edges, we compute the Jacobian blocks as a sum of
!       derivatives of the numerical flux over edges. So, we loop over edges,
!       compute the derivatives of the numerical flux with respect to each end node
!       of the edge, and accumulate the contribution at the two nodes.
!       At the end of the edge loop, at every node j, we have
!
!               dRj/dUj = sum_over_edges ( dflux/dUj )
!               dRj/dUk = dflux/dUk
!
!       which forms the j-th row of the global Jacobian. Here, 'flux' is the
!       numerical flux between nodes j and k.
!
! - The flux Jacobian, dflux/dUj, is computed by automatic differentiation.
!   Numerical flux subroutines in edu3d_flux_functions_ddt.f90 are all written
!   by automatic differentiation. They return the flux vector as well as
!   the derivative. This is very convenient, and it makes it so easy to try out
!   a new numerical flux without coding the Jacobian subroutine.
!   See edu3d_flux_functions_ddt.f90 and edu3d_fluxjac.f90 to learn
!   how it is implemented.
!
!********************************************************************************
 subroutine compute_jacobian

 use data_module     , only : p2, zero, one, third, fourth
 use data_module     , only : nnodes, x, y, z, nedges, edge, u, x, y, z
 use data_module     , only : n12, n12_mag, e12, e12_mag, jac_diag, jac_off, jac_dinv
 use data_module     , only : ntria, tria, nquad, quad, bn, bn_mag, bc_type
 use data_module     , only : k0, kth_nghbr_of_1, kth_nghbr_of_2
 use data_module     , only : vol, dtau, ar, nnghbrs

 use fluxjac_interface, only : interface_jac, interface_jac_b
 use bc_states        , only : set_bc_state, get_right_state
 use input_parameter  , only : flux_choice
 use poisson_eqns     , only : exact_sol, diff_parameters

 use input_parameter  , only : diffusion_type          , &
                               Lr_factor               , &
                               bc_dirichlet_user_values, &
                               bc_dirichlet            , &
                               bc_neumann_user_values  , &
                               bc_neumann              , &
                               bc_neumann_read_uxuyuz  , &
                               single_dirichlet_node   , &
                               single_dirichlet_value

 use input_parameter  , only : hyp_jac_hand_coded, hyp_jac_type

 use data_module      , only : ref_length

 implicit none

!Local variables

 real(p2), dimension(3)   :: da    , ev
 real(p2)                 :: da_mag, ev_mag

 real(p2), dimension(4)   :: uR
 real(p2), dimension(4,4) :: duRduL

 integer :: i, j, k, ib
 integer :: node1, node2
 integer :: nj, nl, nr

 integer , dimension(3,2) :: loc_nghbr
 real(p2), dimension(4,4) :: dFnduL, dFnduR

 integer                  :: idestat
 real(p2), dimension(4,4) :: inverse

 real(p2) :: nu, Lr, Tr, qnb, ub

 real(p2) :: ar_edge

!     1             3
!      o------------o
!       \         .
!        \       . 
!         \    .
!          \ .
!           o
!           2
!
!
! Neighbor nodes of a given node within a triangle (boundary elements).

  loc_nghbr(1,1) = 2 !nghbr1 of vertex 1.
  loc_nghbr(1,2) = 3 !nghbr2 of vertex 1.

  loc_nghbr(2,1) = 3 !nghbr1 of vertex 2.
  loc_nghbr(2,2) = 1 !nghbr2 of vertex 2.

  loc_nghbr(3,1) = 1 !nghbr1 of vertex 3.
  loc_nghbr(3,2) = 2 !nghbr2 of vertex 3.

!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
! Jacobian computation: interior edges

! Initialization

   jac_diag = zero
   jac_off  = zero

!--------------------------------------------------------------------------------
! Flux computation across internal edges (to be accumulated in res(:))
!
!   node2              1. Extrapolate the solutions to the edge-midpoint
!       o                 from the nodes, n1 and n2.
!        \             2. Compute the numerical flux
!         \            3. Add it to the residual for n1, and subtract it from
!          \              the residual for n2.
!           \
!            o         Directed area is the sum of all dual faces defined
!          node1       by the edge-midpoint, centroids of adjacent tetrahedra
!                      and the centroids of the adjacent triangular faces.
!                      Directed area is positive in n1 -> n2
!--------------------------------------------------------------------------------
  loop_edges : do i = 1, nedges


 ! Left and right nodes of the i-th edge

     node1 = edge(i,1)  ! Left node of the edge
     node2 = edge(i,2)  ! Right node of the edge
       da = n12(i,:)    ! This is the directed area vector (unit vector)
   da_mag = n12_mag(i)  ! Magnitude of the directed area vector
       ev = e12(i,:)    ! Edge vector (unit vector)
   ev_mag = e12_mag(i)  ! Magnitude of the edge vector

 !  Compute the flux Jacobians for given w1 and w2 (nodal values)

  !------------------------------------------------------------------------------------------
  ! Hand-coded Jaocobian is available only for the hyperbolic Poisson solver for now.

   if (hyp_jac_hand_coded) then

    ar_edge =  max(ar(node1),ar(node2))

    call hypdiff_3d_schemeII_jacobian( u(node1,:), u(node2,:), da, ev, ev_mag, ar_edge, &
                                           ref_length, hyp_jac_type==2, dFnduL, dFnduR  )

  !------------------------------------------------------------------------------------------
  ! Otherwise, the automatic differentiation is used to compute the Jacobian.

   else

   call interface_jac( u(node1,:), u(node2,:), da, ev, ev_mag, ar(node1), ar(node2), &
                                                                     dFnduL, dFnduR  )

   endif
  !------------------------------------------------------------------------------------------

 !  Add the Jacobian multiplied by the magnitude of the directed area vector to node1.

     jac_diag(node1,:,:)        = jac_diag(node1,:,:)        + (dFnduL) * da_mag
     k = kth_nghbr_of_1(i)
     jac_off( k0(node1)+k ,:,:) = jac_off( k0(node1)+k, :,:) + (dFnduR) * da_mag

 !  Subtract the Jacobian multiplied by the magnitude of the directed area vector from node2.
 !  NOTE: Subtract because the outward face normal is -da for the node2.

     jac_diag(node2,:,:)        = jac_diag(node2,:,:)        - (dFnduR) * da_mag
     k = kth_nghbr_of_2(i)
     jac_off( k0(node2)+k, :,:) = jac_off( k0(node2)+k, :,:) - (dFnduL) * da_mag

  end do loop_edges
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Triangular boundary faces:
!
! Close the residual by the 2nd-order boundary flux quadrature.
!
! Loop over triangular boundary elements and distribute a contribution to each node. 
! Node ordering (orientation) is not important here.
! For each node, j, we call the other nodes, l and r.
!
!     l              r
!      o------------o
!       \         .
!        \       .
!         \    .
!          \ .
!           o
!           j
!
!
!    Flux to be added to the residual at j is
!
!      6/8*Flux(uj,ub) + 1/8*Flux(ul,ub) + 1/8*Flux(ur,ub),
!
!    where ub is a BC state.
!
!
! - Second-order quadrature formulas for 3D tet/prs/pyr/hex grids can be found in:
!
! H. Nishikawa, "Beyond Interface Gradient: A General Principle for
! Constructing Diffusion Schemes", AIAA Paper 2010-5093, 40th AIAA Fluid
! Dynamics Conference and Exhibit, Chicago, 2010.
! http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_AIAA-2010-5093.pdf
! (NOTE: See Appendix B for boundary flux quadratures for 2D/3D elements.)
!
!--------------------------------------------------------------------------------

  if (ntria > 0) then

  boundary_element_loop : do i = 1, ntria

    ib = tria(i,4) !Boundary part number

    nodes_of_triangle : do j = 1, 3 ! Nodes: nj, nr, nl

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! Second-order accurate boundary flux quadratue for tetra (6/8,1/8,1/8)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
     !----------------------------------------------------------------
     ! Node j: Phi_jb(bn): BC is implemented here.

          nj = tria( i, j )

       call set_bc_state( u(nj,1),u(nj,2),u(nj,3),u(nj,4),ib,bn(i,:),nj,x(nj),y(nj),z(nj), ub,qnb) !Get ub,qnb
       call get_right_state(  bc_type(ib),u(nj,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL, uR
       call interface_jac_b( u(nj,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nj), dFnduL, dFnduR )

     !Contribution to node j (diagonal block):
      jac_diag(nj,:,:) = jac_diag(nj,:,:) &
                       + 6.0_p2/8.0_p2*(dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

     !----------------------------------------------------------------
     ! Flux between j and r: Phi_jr(bn): BC is implemented here.

          nr = tria( i, loc_nghbr(j,1) )

       call set_bc_state( u(nr,1),u(nr,2),u(nr,3),u(nr,4),ib,bn(i,:),nr,x(nr),y(nr),z(nr), ub,qnb) !Get ub,qnb
       call get_right_state(  bc_type(ib),u(nr,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL, uR
       call interface_jac_b( u(nr,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nr), dFnduL, dFnduR )

     !Contribution to node j (off-diagonal block):
      k = kth_nghbr(nj,nr)
      jac_off( k0(nj)+k ,:,:) = jac_off( k0(nj)+k, :,:) &
                              + 1.0_p2/8.0_p2*(dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

     !----------------------------------------------------------------
     ! Flux between j and l: Phi_jl(bn): BC is implemented here.

          nl = tria( i, loc_nghbr(j,2) )

       call set_bc_state( u(nl,1),u(nl,2),u(nl,3),u(nl,4),ib,bn(i,:),nl,x(nl),y(nl),z(nl), ub,qnb) !Get ub,qnb
       call get_right_state(  bc_type(ib),u(nl,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL, uR
       call interface_jac_b( u(nl,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nl), dFnduL, dFnduR )

     !Add the boundary contribution to the residual with the 2nd-order formula:
     !Contribution to node j (off-diagonal block):
      k = kth_nghbr(nj,nl)
      jac_off( k0(nj)+k ,:,:) = jac_off( k0(nj)+k, :,:) &
                              + 1.0_p2/8.0_p2*(dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! End of two other nodes
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

    end do nodes_of_triangle

  end do boundary_element_loop

  endif

!--------------------------------------------------------------------------------
! Quadrialteral boundary faces:
!
! Close the residual by the 2nd-order boundary flux quadrature.
!
! Loop over quad boundary elements and distribute a contribution to each node. 
!               
!      o------------o
!       \            \
!        \            \ 
!         \            \
!          \            \ 
!           o------------o
!           j
!
!
!    Flux to be added to the residual at j is just
!
!      Flux(uj,ub)
!
!    where ub is a BC state.
!
!
! - This is simple for hexa grids. We only need to compute the flux at the node j and
!   add it to the residual at j. This is second-order accurate for regular hexa grids
!   with general rectangular cells. If a hexahedral cell is irregular with non-planar
!   faces, then second-order will be lost. See the reference below for details.
!
! - Second-order quadrature formulas for 3D tet/prs/pyr/hex grids can be found in:
!
! H. Nishikawa, "Beyond Interface Gradient: A General Principle for
! Constructing Diffusion Schemes", AIAA Paper 2010-5093, 40th AIAA Fluid
! Dynamics Conference and Exhibit, Chicago, 2010.
! http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_AIAA-2010-5093.pdf
! (NOTE: See Appendix B for boundary flux quadratures for 2D/3D elements.)
!
!--------------------------------------------------------------------------------

  if (nquad > 0) then

  boundary_element_loop_quad : do i = 1, nquad

    ib = quad(i,5) !Boundary part number

    nodes_of_quad : do j = 1, 4

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! Second-order accurate boundary flux quadratue for hexa (1,0,0,0)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

     ! Node j: Phi_jb(bn): BC is implemented here.

      nj = quad( i, j )

      call set_bc_state( u(nj,1),u(nj,2),u(nj,3),u(nj,4),ib,bn(i,:),nj,x(nj),y(nj),z(nj), ub,qnb) !Get ub,qnb
      call get_right_state(  bc_type(ib),u(nj,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL
      call interface_jac_b( u(nj,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nj), dFnduL, dFnduR )

     !Contribution to node j (diagonal block):
      jac_diag(nj,:,:) = jac_diag(nj,:,:) + (dFnduL + matmul(dFnduR,duRduL)) * (fourth*bn_mag(i))

     ! That's it for hexa grids. Simple.

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! End of Second-order accurate boundary flux quadratue for hexa (1,0,0,0)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

    end do nodes_of_quad

  end do boundary_element_loop_quad

  endif

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Add pseudo time term vol/dtau to the diagonals of the diagonal blocks
! to form: Jacobian = P^{-1}*V/dtau + dRes1/dU, where P^{-1}*V/dtau is a global
! diagonal matrix having vol(i)/dtau(i) times 1 or Tr/nu for each node, and Res1
! is the first-order residual.

  do i = 1, nnodes
   call diff_parameters( diffusion_type,Lr_factor,u(i,1), nu,Lr,Tr )
    jac_diag(i,1,1) = jac_diag(i,1,1) + vol(i)/dtau(i)
    jac_diag(i,2,2) = jac_diag(i,2,2) + vol(i)/dtau(i) * Tr/nu !<- P^{-1}
    jac_diag(i,3,3) = jac_diag(i,3,3) + vol(i)/dtau(i) * Tr/nu !<- P^{-1}
    jac_diag(i,4,4) = jac_diag(i,4,4) + vol(i)/dtau(i) * Tr/nu !<- P^{-1}
  end do

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Add source terms: (ux,uy,uz)/nu.

  do i = 1, nnodes
   call diff_parameters( diffusion_type,Lr_factor,u(i,1), nu,Lr,Tr )
   do k = 2, 4
    jac_diag(i,k,k) = jac_diag(i,k,k) + vol(i)/nu
   end do
  end do

!--------------------------------------------------------------------------------


!--------------------------------------------------------------------------------
! Enforce a strong condition if requested.
! This assumes the exact solution. So, this is used only for verification.
!--------------------------------------------------------------------------------

 !-----------------------------------------------------------------------
 ! Triangular elements

   tria_strong : do i = 1, ntria

    ib = tria(i,4)

   !-----------------------------------------------------------
    if ( trim(bc_type(ib))=="dirichlet_strong") then

     do j = 1, 3
       nj = tria( i, j )
      !call exact_sol( exact_sol_type, x(nj), y(nj), z(nj), ue(1), ue(2), ue(3), ue(4) )
      !res(i,1) = u(i,1) - ue(1)
       jac_diag(nj,1,:) = zero
       jac_diag(nj,1,1) =  one
      do k = 1, nnghbrs(nj)
       jac_off( k0(nj)+k, 1,:) = zero
      end do
     end do

   !-----------------------------------------------------------
   ! This is for verification purpose only.
    elseif ( trim(bc_type(ib))=="dirichlet_strong_all") then

     do j = 1, 3

       nj = tria( i, j )
       jac_diag(nj,:,:) = zero

      do k = 1, 4
       jac_diag(nj,k,k) =  one
      end do

      do k = 1, nnghbrs(nj)
       jac_off( k0(nj)+k, :,:) = zero
      end do

     end do

   !-----------------------------------------------------------
   ! Fix the solution at a single node.

     elseif (single_dirichlet_node > 0 ) then

     do j = 1, 3
              nj = tria( i, j )

      if (single_dirichlet_node == nj) then

        jac_diag(nj,1,:) = zero
        jac_diag(nj,1,1) =  one

       do k = 1, nnghbrs(nj)
        jac_off( k0(nj)+k, 1,:) = zero
       end do

        exit tria_strong

      endif

      end do

     endif
   !-----------------------------------------------------------

   end do tria_strong


 !-----------------------------------------------------------------------
 ! Quad elements

   quad_strong : do i = 1, nquad

    ib = quad(i,5)

   !-----------------------------------------------------------
    if ( trim(bc_type(ib))=="dirichlet_strong") then

     do j = 1, 4
       nj = quad( i, j )
      !call exact_sol( exact_sol_type, x(nj), y(nj), z(nj), ue(1), ue(2), ue(3), ue(4) )
      !res(i,1) = u(i,1) - ue(1)
       jac_diag(nj,1,:) = zero
       jac_diag(nj,1,1) =  one
      do k = 1, nnghbrs(nj)
       jac_off( k0(nj)+k, 1,:) = zero
      end do
     end do

   !-----------------------------------------------------------
   ! This is for verification purpose only.
    elseif ( trim(bc_type(ib))=="dirichlet_strong_all") then

     do j = 1, 4

       nj = quad( i, j )
       jac_diag(nj,:,:) = zero

      do k = 1, 4
       jac_diag(nj,k,k) =  one
      end do

      do k = 1, nnghbrs(nj)
       jac_off( k0(nj)+k, :,:) = zero
      end do

     end do

   !-----------------------------------------------------------
   ! Fix the solution at a single node.

     elseif (single_dirichlet_node > 0 ) then

     do j = 1, 4
              nj = quad( i, j )

      if (single_dirichlet_node == nj) then

        jac_diag(nj,1,:) = zero
        jac_diag(nj,1,1) =  one

       do k = 1, nnghbrs(nj)
        jac_off( k0(nj)+k, 1,:) = zero
       end do

        exit quad_strong

      endif

      end do

     endif

   end do quad_strong

!----------------------------------------------------------------
! Conventional scheme has only 1 equation, and so the Jaociban
! has only 1x1 blocks. I know it is not
! an efficient way to implement it, but it is simpler this way
! to implement both the conventional and hyperbolic solvers...
!----------------------------------------------------------------
  if (trim(flux_choice)=="alpha") then

   do i = 1, nnodes

    do j = 2, 4
     jac_diag(i,1,j) = zero
     jac_diag(i,2,:) = zero
     jac_diag(i,3,:) = zero
     jac_diag(i,4,:) = zero
    end do

    do j = 2, 4
     jac_diag(i,j,j) =  one
    end do

    do k = 1, nnghbrs(nj)
     do j = 2, 4
      jac_off( k0(nj)+k, 1,j) = zero
      jac_off( k0(nj)+k, 2,:) = zero
      jac_off( k0(nj)+k, 3,:) = zero
      jac_off( k0(nj)+k, 4,:) = zero
     end do
    end do

   end do

  endif
!----------------------------------------------------------------
!----------------------------------------------------------------
!----------------------------------------------------------------

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
!
! Invert the diagonal blocks.
!
!  - I will invert the diagonal blocks here, and store them.
!  - Is it inefficient? Think about it.
!
!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------

  invert_at_node : do i = 1, nnodes

     inverse = zero
     idestat = 0

 !  Invert the diagonal block at node i by Gauss elimination with pivoting.
 !  Note: gewp_solve() actually solves a linear system, Ax=b, but here
 !        we use it only to obtain the inverse of A. So, b is a dummy.

 !                   A                dim  A^{-1}
    call gewp_solve( jac_diag(i,:,:), 4  , inverse, idestat )

 !  Report errors

    if (idestat/=0) then
     write(*,*) " Error in inverting the diagonal block... Stop"
     write(*,*) "  Node number = ", i
     write(*,*) "  Location    = ", x(i), y(i), z(i)
     do k = 1, 4
      write(*,'(12(es8.1))') ( jac_diag(i, k,j), j=1,4 )
     end do
     stop
    endif

 !  Save the inverse in jac_dinv.

    jac_dinv(i,:,:) = inverse

  end do invert_at_node

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------

 end subroutine compute_jacobian
!********************************************************************************


!********************************************************************************
!
! This function is useful to find 'k', such that 
! node 'n2' is the k-th nghbr of the node 'n1'.
!
!********************************************************************************
 function kth_nghbr(n1,n2)

 use data_module, only : nghbr, k0, nnghbrs

 implicit none

 integer, intent(in) :: n1, n2

 integer :: k
 integer :: kth_nghbr
 logical :: found

  kth_nghbr = 0
      found = .false.

 !Find k, such that n2 is the k-th neghbor of n1.

  n1_nghbr_loop : do k = 1, nnghbrs(n1)

   if ( nghbr(k0(n1)+k) == n2 ) then

    kth_nghbr = k
    found = .true.
    exit n1_nghbr_loop

   endif

  end do n1_nghbr_loop

  if (.not. found) then
   write(*,*) " Neighbor not found.. Error. Stop"
   stop
  endif

 end function kth_nghbr


!****************************************************************************
!* ------------------ GAUSS ELIMINATION WITH PIVOTING ---------------------
!*
!*  This computes the inverse of an (nm)x(nm) matrix "ai" and also
!*  computes the solution to a given lienar system.
!*
!*  IN :       ai = An (nm)x(nm) matrix whoise inverse is sought.
!*             bi = A vector of (nm): Right hand side of the linear sytem
!*             nm = The size of the matrix "ai"
!*
!* OUT :
!*            sol = Solution to the linear system: ai*sol=bi
!*        inverse = the inverse of "ai".
!*       idetstat = 0 -> inverse successfully computed
!*                  1 -> THE INVERSE DOES NOT EXIST (det=0).
!*                  2 -> No unique solutions exist.
!*****************************************************************************
  subroutine gewp_solve(ai,nm, inverse,idetstat)

  implicit none

 integer , parameter ::    p2 = selected_real_kind(15) ! Double precision
 real(p2), parameter ::  zero = 0.0_p2
 real(p2), parameter ::   one = 1.0_p2

  integer ,                   intent( in) :: nm
  real(p2), dimension(nm,nm), intent( in) :: ai

  real(p2), dimension(nm,nm), intent(out) :: inverse
  integer ,                   intent(out) :: idetstat

  real(p2), dimension(nm,nm+1) :: a
  real(p2), dimension(nm)      :: x
  integer , dimension(nm)      :: nrow
  integer                      :: I,J,K,pp,m

  do m = 1, nm
!*****************************************************************************
!* Set up the matrix a
!*****************************************************************************

       do J=1,nm
        do I=1,nm
          a(I,J) = ai(I,J)
        end do
       end do

       do k=1,nm
          a(k,nm+1)=zero; nrow(k)=k
       end do
          a(m,nm+1)=one

!*****************************************************************************
!* HONA IKOKA..... 
!*****************************************************************************
    do j=1,nm-1
!*****************************************************************************
!* FIND SMALLEST pp FOR a(pp,j) IS MAXIMUM IN JTH COLUMN.
!***************************************************************************** 
      call findmax(nm,j,pp,a,nrow)
!*****************************************************************************
!* IF a(nrow(p),j) IS zero, THERE'S NO UNIQUE SOLUTIONS      
!*****************************************************************************
!      if (abs(a(nrow(pp),j)) < epsilon(one)) then
!       write(6,*) 'THE INVERSE DOES NOT EXIST.'
!        idetstat = 1
!        return
!      endif
!*****************************************************************************
!* IF THE MAX IS NOT A DIAGONAL ELEMENT, SWITCH THOSE ROWS       
!*****************************************************************************
      if (nrow(pp) .ne. nrow(j)) then
      call switch(nm,j,pp,nrow)
      else
      endif  
!*****************************************************************************
!* ELIMINATE ALL THE ENTRIES BELOW THE DIAGONAL ONE
!***************************************************************************** 
      call eliminate_below(nm,j,a,nrow)

    end do
!*****************************************************************************
!* CHECK IF a(nrow(N),N)=0.0 .
!*****************************************************************************
!      if (abs(a(nrow(nm),nm)) < epsilon(one)) then
!        write(6,*) 'NO UNIQUE SOLUTION EXISTS!'
!        idetstat = 2
!        return
!      else
 !     endif
!*****************************************************************************
!* BACKSUBSTITUTION!
!*****************************************************************************
      call backsub(nm,x,a,nrow)
!*****************************************************************************
!* STORE THE SOLUTIONS, YOU KNOW THEY ARE INVERSE(i,m) i=1...
!*****************************************************************************
      do i=1,nm
         inverse(i,m)=x(i)
      end do
!*****************************************************************************
  end do

      idetstat = 0

    return

!*****************************************************************************
 end subroutine gewp_solve

!*****************************************************************************
!* Four subroutines below are used in gewp_solve() above.
!*****************************************************************************
!* FIND MAXIMUM ELEMENT IN jth COLUMN 
!***************************************************************************** 
      subroutine findmax(nm,j,pp,a,nrow)

      implicit none

      integer , parameter   :: p2 = selected_real_kind(15) ! Double precision
      integer , intent( in) :: nm
      real(p2), intent( in) :: a(nm,nm+1)
      integer , intent( in) :: j,nrow(nm)
      integer , intent(out) :: pp
      real(p2)              :: max
      integer               :: i

            max=abs(a(nrow(j),j)); pp=j

           do i=j+1,nm

             if (max < abs(a(nrow(i),j))) then

                  pp=i; max=abs(a(nrow(i),j))

             endif

           end do

      return

      end subroutine findmax
!*****************************************************************************
!* SWITCH THOSE ROWS       
!*****************************************************************************
      subroutine switch(nm,j,pp,nrow)

      implicit none

      integer, intent(   in) :: nm,j,pp
      integer, intent(inout) :: nrow(nm)
      integer                :: ncopy

      if (nrow(pp).ne.nrow(j)) then

         ncopy=nrow(j)
         nrow(j)=nrow(pp)
         nrow(pp)=ncopy

      endif

      return

      end subroutine switch
!*****************************************************************************
!* ELIMINATE ALL THE ENTRIES BELOW THE DIAGONAL ONE
!*(Give me j, the column you are working on now)
!***************************************************************************** 
      subroutine eliminate_below(nm,j,a,nrow)

      implicit none

      integer , parameter     :: p2 = selected_real_kind(15) ! Double precision
      real(p2), parameter     :: zero = 0.0_p2
      integer , intent(   in) :: nm
      real(p2), intent(inout) :: a(nm,nm+1)
      integer , intent(   in) :: j,nrow(nm)
      real(p2)                :: m
      integer                 :: k,i

      do i=j+1,nm

        m=a(nrow(i),j)/a(nrow(j),j)
        a(nrow(i),j)=zero

          do k=j+1,nm+1
            a(nrow(i),k)=a(nrow(i),k)-m*a(nrow(j),k)
          end do

      end do

      return

      end subroutine eliminate_below
!*****************************************************************************
!* BACKSUBSTITUTION!
!*****************************************************************************
      subroutine backsub(nm,x,a,nrow)

      implicit none

      integer , parameter   :: p2 = selected_real_kind(15) ! Double precision
      real(p2), parameter   :: zero = 0.0_p2

      integer , intent( in) :: nm
      real(p2), intent( in) :: a(nm,nm+1)
      integer , intent( in) :: nrow(nm)
      real(p2), intent(out) :: x(nm)
      real(p2)              :: sum
      integer               :: i,k

      x(nm)=a(nrow(nm),nm+1)/a(nrow(nm),nm)

      do i=nm-1,1,-1

         sum=zero

           do k=i+1,nm

              sum=sum+a(nrow(i),k)*x(k)

           end do

      x(i)=(a(nrow(i),nm+1)-sum)/a(nrow(i),i)

      end do

      return

      end subroutine backsub
!*********************************************************************


!********************************************************************************
!
! -- 3D Upwind Hyperbolic Diffusion Flux Jacobian --
!
! This is a hand coded Jacobian subroutine for the hyperbolic Poisson solver.
!
! Katate Masatsuka, June 2021. http://www.cfdbooks.com
!
!********************************************************************************
 subroutine hypdiff_3d_schemeII_jacobian(uj, uk, njk, ejk, ejk_mag, ar, &
                                      ref_length, SchemeII, dFduj, dFduk)

 use input_parameter, only : diffusion_type

 implicit none

 integer , parameter :: p2 = selected_real_kind(15) ! Double precision

!Input
 real(p2), dimension(4), intent( in) :: uj          !=[uj,uxj,uyj,uzj]
 real(p2), dimension(4), intent( in) :: uk          !=[uk,uxk,uyk,uzk]
 real(p2), dimension(3), intent( in) :: njk         !Unit face normal
 real(p2), dimension(3), intent( in) :: ejk         !Unit edge vector.
 real(p2)              , intent( in) :: ejk_mag     !Magnitude of the edge vector.
 real(p2)              , intent( in) :: ar          !Aspect ratio
 real(p2)              , intent( in) :: ref_length  !Reference length
 logical               , intent( in) :: SchemeII    !=T for Scheme-II

!Output
 real(p2), dimension(4,4), intent(out) :: dFduj !  Left Jacobian, d(Fn)/d(uj)
 real(p2), dimension(4,4), intent(out) :: dFduk ! Right Jacobian, d(Fn)/d(uk)

!Some constants
 real(p2), parameter :: zero = 0.00_p2
 real(p2), parameter :: half = 0.50_p2
 real(p2), parameter ::  one = 1.00_p2
 real(p2), parameter ::  two = 2.00_p2
 real(p2), parameter ::   pi = 3.141592653589793238_p2

!Local variables (as in the flux subroutine):
 real(p2), dimension(4)   ::  uL, uR     !Reconstructed values for Scheme-II.
 real(p2)                 :: pnL, pnR    !Projected gratient variables (no reconstruction for Jac).
 real(p2), dimension(4)   :: dujk        !For kappa reconstruction scheme.
 real(p2), dimension(4)   :: duL, duR    !For kappa reconstruction scheme.

 real(p2)                 :: Lr          !Relaxation length scale
 real(p2)                 :: nuj, nuk    !Diffusion coefficients: nuj=nu(uj), nuk=nu(uk)
 real(p2)                 :: nu          !Averaged diffusion coefficient
 real(p2)                 :: kappa       !Kappa in the kappa scheme.

 real(p2)                 :: ex, ey, ez  !Unit   edge vector components.
 real(p2)                 :: nx, ny, nz  !Unit normal vector components.

 integer                  :: iu, iux, iuy, iuz !=[1,2,3,4] as defined below.
 integer                  ::      ix,  iy,  iz !=[1,2,3]   as defined below.

!Local variables (used only in this Jacobian subroutine):
 real(p2), dimension(4,4) :: I4x4     !Identity matrix used for some initialization.

 real(p2), dimension(4,4) :: dFduL    !Derivative of averaged flux part w.r.t. uL
 real(p2), dimension(4,4) :: dFduR    !Derivative of averaged flux part w.r.t. uR

 real(p2), dimension(4,4) :: duLduj   !Derivative of reconstructed uL w.r.t. uj
 real(p2), dimension(4,4) :: duLduk   !Derivative of reconstructed uL w.r.t. uk

 real(p2), dimension(4,4) :: duRduj   !Derivative of reconstructed uR w.r.t. uj
 real(p2), dimension(4,4) :: duRduk   !Derivative of reconstructed uR w.r.t. uk

 real(p2), dimension(4,4) :: dDduL    !Derivative for linear part of dissipation
 real(p2), dimension(4,4) :: dDduR    !Derivative for linear part of dissipation

 real(p2), dimension(4,4) :: dDNduj   !Derivative for nonlinear part of dissipation
 real(p2), dimension(4,4) :: dDNduk   !Derivative for nonlinear part of dissipation

 real(p2)                 :: dnu_duj  !Derivative of nu w.r.t. uj.
 real(p2)                 :: dnu_duk  !Derivative of nu w.r.t. uk.

 real(p2)                 :: dnuj_duj !Derivative of nuj w.r.t. uj.
 real(p2)                 :: dnuk_duk !Derivative of nuk w.r.t. uk.

 real(p2)                 :: dduL_duj !Derivative of projected grad in kappa scheme.
 real(p2)                 :: dduR_duk !Derivative of projected grad in kappa scheme.

  ix  = 1
  iy  = 2
  iz  = 3

  iu  = 1
  iux = 2
  iuy = 3
  iuz = 4

! Identity matrix used for some initialization.

  I4x4      = zero
  I4x4(1,1) = one
  I4x4(2,2) = one
  I4x4(3,3) = one
  I4x4(4,4) = one

! Edeg vector (unit vector)

   ex = ejk(1)
   ey = ejk(2)
   ez = ejk(3)

! Relaxation length scale

   Lr = one / (two*pi)*ref_length

! Diffusion coefficient = 1 for Poisson.

   if     (diffusion_type==1 ) then

        nuj = one
        nuk = one
   dnuj_duj = zero
   dnuk_duk = zero

   elseif (diffusion_type==10) then

        nuj = one + uj(1)**2
        nuk = one + uk(1)**2
   dnuj_duj = two*uj(1)
   dnuk_duk = two*uk(1)

   else
        nuj = one
        nuk = one
   dnuj_duj = zero
   dnuk_duk = zero

   endif


! No reconstruction (first-order Scheme-I)

   uL = uj
   uR = uk

! Scheme-II recosntruction (first-order Scheme-II)

  if (SchemeII) then

   if (ar < 10.0_p2) then
    kappa = 0.0_p2
   else
    kappa = 0.5_p2
   endif

   dujk(1) = half*( uk(1) - uj(1) )
    duL(1) = ( uL(iux)*ejk(ix) + uL(iuy)*ejk(iy) + uL(iuz)*ejk(iz) )/nuj*ejk_mag*half
    duR(1) = ( uR(iux)*ejk(ix) + uR(iuy)*ejk(iy) + uR(iuz)*ejk(iz) )/nuk*ejk_mag*half

     uL(1) = uj(1) + ( (one-kappa)*(two*duL(1)-dujk(1)) + (one+kappa)*dujk(1) )*half
     uR(1) = uk(1) - ( (one-kappa)*(two*duR(1)-dujk(1)) + (one+kappa)*dujk(1) )*half

  endif

! Diffusion coefficient = 1 for Poisson.

         nu = half*(nuj+nuk)

    dnu_duj = half*dnuj_duj
    dnu_duk = half*dnuk_duk

! Face normal vector (unit vector)

    nx = njk(1)
    ny = njk(2)
    nz = njk(3)

! Projected gradient variables.

   pnL = uL(iux)*nx + uL(iuy)*ny + uL(iuz)*nz
   pnR = uR(iux)*nx + uR(iuy)*ny + uR(iuz)*nz

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Part 0. Compute the derivatives of the Scheme-II reconstruction
!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! Initialization with first-order Scheme-I.

   duLduj         =  I4x4
   duLduk         =  zero

   duRduj         =  zero
   duRduk         =  I4x4

  !-------------------------------------------------------------------------------
  ! Contribution from the first-order Scheme-II

  if (SchemeII) then

  !-------------------------------------------------------------------------------
  ! Derivatives of uL(1):
  ! uL(1) = uj(1) + ( (one-kappa)*(two*duL(1)-dujk(1)) + (one+kappa)*dujk(1) )*half

  ! (1) duL/duj

     duLduj(iu,iu ) =  one + ( (one-kappa)*(zero-(-half)) + (one+kappa)*(-half) )*half
     duLduj(iu,iux) =  ( (one-kappa)*( ejk(ix)/nuj*ejk_mag - zero)  )*half
     duLduj(iu,iuy) =  ( (one-kappa)*( ejk(iy)/nuj*ejk_mag - zero)  )*half
     duLduj(iu,iuz) =  ( (one-kappa)*( ejk(iz)/nuj*ejk_mag - zero)  )*half

    !Nonlinear contribution
     dduL_duj = ( uL(iux)*ejk(ix) + uL(iuy)*ejk(iy) + uL(iuz)*ejk(iz) )*(-one/nuj**2)*dnuj_duj*ejk_mag*half
     duLduj(iu,iu ) = duLduj(iu,iu ) + ( (one-kappa)*two*dduL_duj )*half

  ! (2) duL/duk

     duLduk(iu,iu ) = zero + ( (one-kappa)*(zero-( half)) + (one+kappa)*( half) )*half

  !-------------------------------------------------------------------------------
  ! Derivatives of uR(1):
  ! uR(1) = uk(1) - ( (one-kappa)*(two*duR(1)-dujk(1)) + (one+kappa)*dujk(1) )*half

  ! (1) duR/duj

    duRduj(iu,iu ) = zero - ( (one-kappa)*(zero-(-half)) + (one+kappa)*(-half) )*half

  ! (2) duR/duk

    duRduk(iu,iu ) =  one - ( (one-kappa)*(zero-( half)) + (one+kappa)*( half) )*half
    duRduk(iu,iux) =  -( (one-kappa)*( ejk(ix)/nuk*ejk_mag - zero)  )*half
    duRduk(iu,iuy) =  -( (one-kappa)*( ejk(iy)/nuk*ejk_mag - zero)  )*half
    duRduk(iu,iuz) =  -( (one-kappa)*( ejk(iz)/nuk*ejk_mag - zero)  )*half

    !Nonlinear contribution
     dduR_duk = ( uR(iux)*ejk(ix) + uR(iuy)*ejk(iy) + uR(iuz)*ejk(iz) )*(-one/nuk**2)*dnuk_duk*ejk_mag*half
     duRduk(iu,iu ) = duRduk(iu,iu ) - ( (one-kappa)*two*dduR_duk )*half

  endif

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Part 1. Compute dFn/duL, F = half*(fL + fR - diss)
!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------

 !----------------------------------------------------------------------
 ! 1.1 Compute the left flux Jacbian -> d( half*(fL+fR) )/duL

 ! 1st row: half*d(-pnL)/duL , pnL = uL(iux)*nx + uL(iuy)*ny + uL(iuz)*nz
   dFduL(1,1) =  zero
   dFduL(1,2) = -half*nx
   dFduL(1,3) = -half*ny
   dFduL(1,4) = -half*nz

 ! 2nd row: half*d(-uL(iu)*nx)/duL;
   dFduL(2,1) = -half*nx
   dFduL(2,2) =  zero
   dFduL(2,3) =  zero
   dFduL(2,4) =  zero

 ! 3rd row: half*d(-uL(iu)*ny)/duL;
   dFduL(3,1) = -half*ny
   dFduL(3,2) =  zero
   dFduL(3,3) =  zero
   dFduL(3,4) =  zero

 ! 4th row: half*d(-uL(iu)*nz)/duL;
   dFduL(4,1) = -half*nz
   dFduL(4,2) =  zero
   dFduL(4,3) =  zero
   dFduL(4,4) =  zero

 !----------------------------------------------------------------------
 ! 1.2 Compute the derivative of the dissipation -> d( -half*diss )/duL
 !     --> Nonlinear part, with nu as a function of u.

  ! 1st component: d( lambda * ( uR(iu) - uL(iu) ) )
    dDduL(1,:  ) = zero
    dDduL(1,iu ) = -half* nu/Lr*(zero-one)

  ! 2nd component: d( lambda * (   pnR  -    pnL )*nx * Tr/nu ) )
    dDduL(2,:  ) = zero
    dDduL(2,iux) = -half* Lr/nu*(zero-nx)*nx
    dDduL(2,iuy) = -half* Lr/nu*(zero-ny)*nx
    dDduL(2,iuz) = -half* Lr/nu*(zero-nz)*nx

  ! 3rd component: d( lambda * (   pnR  -    pnL )*ny * Tr/nu ) )
    dDduL(3,:) = zero
    dDduL(3,iux) = -half* Lr/nu*(zero-nx)*ny
    dDduL(3,iuy) = -half* Lr/nu*(zero-ny)*ny
    dDduL(3,iuz) = -half* Lr/nu*(zero-nz)*ny

  ! 4th component: d( lambda * (   pnR  -    pnL )*nz * Tr/nu ) )
    dDduL(4,:) = zero
    dDduL(4,iux) = -half* Lr/nu*(zero-nx)*nz
    dDduL(4,iuy) = -half* Lr/nu*(zero-ny)*nz
    dDduL(4,iuz) = -half* Lr/nu*(zero-nz)*nz

 !----------------------------------------------------------------------
 ! 1.3 Compute the derivative of the dissipation -> d( -half*diss )/duL
 !     --> Nonlinear part, with nu and Tr as functions of u.
 !     Skip this part if nu is constant.

    dDNduj(1,iu ) = zero

  ! 1st component: d( lambda * ( uR(iu) - uL(iu) ) ), lambda=nu/Lr
    dDNduj(1,iu ) = -half * ( uR(iu)-uL(iu) )       *dnu_duj/Lr

  ! 2nd component: d( (pnR-pnL)*nx * Lr/nu ) )
    dDNduj(2,iu ) = -half * (pnR-pnL)*nx*(-Lr/nu**2)*dnu_duj

  ! 3rd component: d( (pnR-pnL)*ny * Lr/nu ) )
    dDNduj(3,iu ) = -half * (pnR-pnL)*ny*(-Lr/nu**2)*dnu_duj

  ! 4th component: d( (pnR-pnL)*ny * Lr/nu ) )
    dDNduj(4,iu ) = -half * (pnR-pnL)*nz*(-Lr/nu**2)*dnu_duj

 !----------------------------------------------------------------------
 ! 1.4 Compute the final dF/duL

    dFduL = dFduL + dDduL

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Part 2. Compute dFn/duR, F = half*(fL + fR - diss)
!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------

 !----------------------------------------------------------------------
 ! 2.1 Compute the left flux Jacbian -> d( half*(fL+fR) )/duR

 ! 1st row: half*d(-pnR)/duR , pnR = uR(iux)*nx + uR(iuy)*ny + uR(iuz)*nz
   dFduR(1,1) =  zero
   dFduR(1,2) = -half*nx
   dFduR(1,3) = -half*ny
   dFduR(1,4) = -half*nz

 ! 2nd row: half*d(-uR(iu)*nx)/duR;
   dFduR(2,1) = -half*nx
   dFduR(2,2) =  zero
   dFduR(2,3) =  zero
   dFduR(2,4) =  zero

 ! 3rd row: half*d(-uR(iu)*ny)/duR;
   dFduR(3,1) = -half*ny
   dFduR(3,2) =  zero
   dFduR(3,3) =  zero
   dFduR(3,4) =  zero

 ! 4th row: half*d(-uR(iu)*nz)/duR;
   dFduR(4,1) = -half*nz
   dFduR(4,2) =  zero
   dFduR(4,3) =  zero
   dFduR(4,4) =  zero

 !----------------------------------------------------------------------
 ! 2.2 Compute the derivative of the dissipation -> d( -half*diss )/duR
 !     --> Only the linear part here, assuming nu and Tr are constants.

  ! 1st component: d( lambda * ( uR(iu) - uL(iu) ) )
    dDduR(1,:  ) = zero
    dDduR(1,iu ) = -half* nu/Lr*(one-zero)

  ! 2nd component: d( lambda * (   pnR  -    pnL )*nx * Tr/nu ) )
    dDduR(2,:  ) = zero
    dDduR(2,iux) = -half* Lr/nu*(nx-zero)*nx
    dDduR(2,iuy) = -half* Lr/nu*(ny-zero)*nx
    dDduR(2,iuz) = -half* Lr/nu*(nz-zero)*nx

  ! 3rd component: d( lambda * (   pnR  -    pnL )*ny * Tr/nu ) )
    dDduR(3,:) = zero
    dDduR(3,iux) = -half* Lr/nu*(nx-zero)*ny
    dDduR(3,iuy) = -half* Lr/nu*(ny-zero)*ny
    dDduR(3,iuz) = -half* Lr/nu*(nz-zero)*ny

  ! 4th component: d( lambda * (   pnR  -    pnL )*nz * Tr/nu ) )
    dDduR(4,:) = zero
    dDduR(4,iux) = -half* Lr/nu*(nx-zero)*nz
    dDduR(4,iuy) = -half* Lr/nu*(ny-zero)*nz
    dDduR(4,iuz) = -half* Lr/nu*(nz-zero)*nz

 !----------------------------------------------------------------------
 ! 2.3 Compute the derivative of the dissipation -> d( -half*diss )/duR
 !     --> Nonlinear part, with nu as a function of u.
 !     Skip this part if nu is constant.

    dDNduk = zero

  ! 1st component: d( lambda * ( uR(iu) - uL(iu) ) ), lambda=nu/Lr
    dDNduk(1,iu ) = -half * ( uR(iu)-uL(iu) )       *dnu_duk/Lr

  ! 2nd component: d( (pnR-pnL)*nx * Lr/nu ) )
    dDNduk(2,iu ) = -half * (pnR-pnL)*nx*(-Lr/nu**2)*dnu_duk

  ! 3rd component: d( (pnR-pnL)*ny * Lr/nu ) )
    dDNduk(3,iu ) = -half * (pnR-pnL)*ny*(-Lr/nu**2)*dnu_duk

  ! 4th component: d( (pnR-pnL)*ny * Lr/nu ) )
    dDNduk(4,iu ) = -half * (pnR-pnL)*nz*(-Lr/nu**2)*dnu_duk

 !----------------------------------------------------------------------
 ! 1.4 Compute the final dF/duR

    dFduR = dFduR + dDduR

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Part 3. Compute the final Jacobians by chain rule.
!         Note: uL and uR depend on both uj and uk for the kappa scheme (kappa>0).
!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------

   dFduj = matmul(dFduL,duLduj) + matmul(dFduR,duRduj) + dDNduj
   dFduk = matmul(dFduL,duLduk) + matmul(dFduR,duRduk) + dDNduk


 end subroutine hypdiff_3d_schemeII_jacobian
!--------------------------------------------------------------------------------



 end module jacobian
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
