!********************************************************************************
!* Educationally-Designed Unstructured 3D (EDU3D) Code
!*
!*  --- EDU3D Poisson
!*
!*
!* This module containes a linear relaxtion solver. Currently, the sequential
!* Gauss-Seidel scheme and the multi-color GS scheme are implemented. Other schemes
!* may be implemented and explored: e.g., symmetric GS, Jacobi, etc.
!*
!*
!*        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!*
!* the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!*
!* 
!* This F90 program is written and made available for an educational purpose.
!*
!* This file may be updated in future.
!*
!* Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module linear_solver

  implicit none

 !This module contains the following subroutine:

  public :: linear_relaxation_gen ! perform linear relaxations.

 contains

!********************************************************************************
!
! This subroutine relaxes the linear system by Sequential or multi-color
!  Gauss-Seidel method.
!
! - Other methods are possible: symmetric GS, etc.
! - This subroutine is used in the preconditioner for the JFNK solver.
! - Convergence results will be printed in the file: fort.1000.
!
!********************************************************************************
 subroutine linear_relaxation_gen(n,nq,nt, jac_off, jac_dinv, rhs, k0, nnghbrs, &
                                         nghbr, color, ncolors,   du,irelax,roc )

 use input_parameter, only : lrelax_omega    , lrelax_omega_min, &
                             lrelax_omega_max, lrelax_omega_fix, &
                             lrelax_use_slope, lrelax_sweeps   , &
                             lrelax_tolerance, lrelax_scheme

 implicit none

 integer , parameter ::   p2 = selected_real_kind(P=15) !Double precision
 real(p2), parameter :: zero = 0.0_p2
 real(p2), parameter ::  one = 1.0_p2

!Input
 integer ,                      intent( in) :: n, nq, nt
 real(p2), dimension(nt,nq,nq), intent( in) :: jac_off  !i=1,nnodes*nnghbrs
 real(p2), dimension(n ,nq,nq), intent( in) :: jac_dinv !i=1,nnodes
 real(p2), dimension(n ,nq   ), intent( in) :: rhs
 integer , dimension(n       ), intent( in) :: k0
 integer , dimension(n       ), intent( in) :: nnghbrs
 integer , dimension(nt      ), intent( in) :: nghbr
 integer , dimension(n       ), intent( in) :: color
 integer ,                      intent( in) :: ncolors

!Output
 real(p2), dimension(n,nq   ),  intent(out) :: du     !correction
 integer                     ,  intent(out) :: irelax !# of relaxations performed.
 real(p2)                    ,  intent(out) :: roc    !Residual reduction from the initial.

!Local variables
 real(p2), dimension(nq)   :: b
 integer                   :: i, k, icolor
 real(p2)                  :: omega

!Linear system residual: linear_res = A^{-1}*(b-Ax) at each cell.
 real(p2), dimension(n,nq) :: linear_res

 real(p2), dimension(lrelax_sweeps) :: roc_loc

  integer                  :: i_temp, stat
  real(p2)                 :: temp1, temp2

  real(p2)                 :: sumx, sumy, sumxy, sumx2, slope
  real(p2)                 :: x, y, ndata, slope_pre, roc_pre

 real(p2), dimension(nq)   :: linear_res_norm1
 real(p2), dimension(nq)   :: linear_res_norm1_initial

        sumx  = zero
        sumy  = zero
        sumxy = zero
        sumx2 = zero
        slope = zero
    slope_pre = zero
    roc_pre   = one

  open(unit=1001, file="delete_me2", status='unknown')

  roc_loc  = zero

!Note: write(1000,*) -> This writes in a file named 'fort.1000'.

  write(1000,*)
  write(1000,*) " GS linear relaxaton scheme "
  write(1000,*)

  omega = lrelax_omega !0.85_p2

!--------------------------------------------------------------------------------
! 0. Initialize the correction

     du = zero

!--------------------------------------------------------------------------------
! 1. Initial linear res with du=0.0.

    linear_res_norm1 = zero
    do i = 1, n
        b = -rhs(i,:)
        linear_res_norm1 = linear_res_norm1 + abs( matmul( jac_dinv(i,:,:), b ) )
    end do
    linear_res_norm1 = linear_res_norm1 / real(n, p2)
    linear_res_norm1_initial = linear_res_norm1

     write(1000,'(a,i10,a,es12.5,2x,2f8.3,2x,es10.2,a)')   "  after relax ", 0,          &
                " max(L1 norm) = ", maxval(linear_res_norm1), 1.0_p2, omega,0.0_p2,"          "

!    write(1001,*) 0, maxval(linear_res_norm1), omega
     write(1001,*) 0, 1.0_p2, omega

!--------------------------------------------------------------------------------
! 2. Linear Relaxation (Sweep)

  irelax = 0

 relax : do

   linear_res_norm1(:) = zero

 !Increment the relaxation counter.
   irelax = irelax + 1

 !--------------------------------------------------------------------------------
 ! Sequential GS
 !--------------------------------------------------------------------------------

   if (trim(lrelax_scheme) == "gs") then

 ! A sequential Gauss-Seidel relaxation(or called 'sweep')

    node_loop : do i = 1, n

    ! Form the right hand side of GS: [ sum( off_diagonal_block*du ) - residual ]!
        b = -rhs(i,:)
       nghbr_loop : do k = 1, nnghbrs(i)
        b = b - matmul( jac_off(k0(i)+k,:,:), du(nghbr(k0(i)+k),:) )
       end do nghbr_loop

    ! Update du by the GS relaxation:
    !
    ! e.g., for 3 nghbrs, perform the relaxation in the form:
    !
    !                     diagonal block        sum of off-diagonal block contributions
    ! duj = omega*{ [V/dtj+dR/duj]^{-1}*(-[dRj/du1]*du1 -[dRj/du2]*du2 -[dRj/du3]*du3 -Res_j) - duj }

         linear_res(i,:) = matmul( jac_dinv(i,:,:), b ) - du(i,:)
                 du(i,:) = du(i,:) + omega*linear_res(i,:)

        linear_res_norm1 = linear_res_norm1 + abs( linear_res(i,:) )

    end do node_loop

 !--------------------------------------------------------------------------------
 ! End of Sequential GS
 !--------------------------------------------------------------------------------

 !--------------------------------------------------------------------------------
 ! Multi-color GS
 !--------------------------------------------------------------------------------

   elseif (trim(lrelax_scheme) == "mcgs") then

   color_loop : do icolor = 1, ncolors

    node_loop_mc : do i = 1, n
    if ( color(i) /= icolor ) cycle node_loop_mc

        b = -rhs(i,:)

       nghbr_loop_mc : do k = 1, nnghbrs(i)
        b = b - matmul( jac_off(k0(i)+k,:,:), du(nghbr(k0(i)+k),:) )
       end do nghbr_loop_mc

         linear_res(i,:) = matmul( jac_dinv(i,:,:), b ) - du(i,:)
                 du(i,:) = du(i,:) + omega*linear_res(i,:)

        linear_res_norm1 = linear_res_norm1 + abs( linear_res(i,:) )

    end do node_loop_mc

   end do color_loop

 !--------------------------------------------------------------------------------
 ! End of Multi-color GS
 !--------------------------------------------------------------------------------

 else

  write(*,*) " Invalid input: trim(lrelax_scheme) = ", trim(lrelax_scheme)
  stop

 endif

    linear_res_norm1 = linear_res_norm1 / real(n, p2)

 ! End of A sequential Gauss-Seidel Relaxation(sweep)

    write(1001,*) irelax, maxval( linear_res_norm1/linear_res_norm1_initial ), omega

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! 3. Check the linear residual.

 !--------------------------------------------------------------------
 !  After the first relaxation
 !--------------------------------------------------------------------

    after_1st_relaxation : if (irelax > 0) then

!       roc = linear_res_norm/linear_res_norm_initial
        roc = maxval( linear_res_norm1/linear_res_norm1_initial )

     !-----------------------------------------------------------------
     ! Update the slope
     !-----------------------------------------------------------------

     !-----------------------------------------------------------------
     ! Print the current linear residual norm in the file 'fort.1000'.
     ! and adaptive relaxation factor, omega.
     !-----------------------------------------------------------------

     fix_omega : if (.not.lrelax_omega_fix) then

        if (irelax == 1) then
          sumx  = zero
          sumy  =  one   !y = residual/initial = 1.0 at at the beginning.
          sumxy = zero
          sumx2 = zero
          slope = zero
          slope_pre = zero
          roc_pre   = one
        endif

            x = real(irelax  ,p2) !x-coordinate = relaxation number
            y = roc               !y-coordinate = residual/initial
        ndata = real(irelax+1,p2) !the number of data
        sumx  = sumx  + x
        sumy  = sumy  + y
        sumxy = sumxy + x*y
        sumx2 = sumx2 + x**2
        slope = ( ndata*sumxy - sumx*sumy ) / ( ndata*sumx2 - sumx**2 )

       !------------------------------------------------------
       !------------------------------------------------------
       !(1)Converging. Keep or increase the current omega.
       !------------------------------------------------------
       !------------------------------------------------------
        converging : if (roc < one) then

          omega = min(1.5_p2*omega, lrelax_omega_max)

       !------------------------------------------------------
       !------------------------------------------------------
       !(2)Adaptive omega
       !------------------------------------------------------
       !------------------------------------------------------
        else

        !-------------------------------------------------
        !(2-1) Use the slope information
        !-------------------------------------------------

         if (lrelax_use_slope) then

          if (roc < 1.15_p2 .and. roc_pre < 1.15_p2) then

            write(1000,*) "  roc < 1.15_p2 .and. roc_pre < 1.15_p2:  Let's keep the current omega..."
            omega = omega

          else

            if     (slope < slope_pre) then

            write(1000,*) "  slope < slope_pre: Coming back. Increase omega..."

             omega = min(1.25_p2*omega, lrelax_omega_max)

            elseif (slope < zero) then

             omega = min(1.5_p2*omega, lrelax_omega_max)

            else

             omega = max(0.5_p2*omega, lrelax_omega_min)

            endif

          endif

        !-------------------------------------------------
        !(2-2) Simple reduction
        !-------------------------------------------------
         else
 
           omega = max(0.5_p2*omega, lrelax_omega_min)

         endif
        !-------------------------------------------------
        !-------------------------------------------------

        endif converging
       !------------------------------------------------------
       !------------------------------------------------------
       !------------------------------------------------------

        slope_pre = slope
          roc_pre = roc

      endif fix_omega

        write(1000,'(a,i10,a,es12.5,2x,2f8.3,2x,es10.2,a)')   "  after relax ", irelax,          &
                  " max(L1 norm) = ", maxval(linear_res_norm1), roc, omega,slope,"          "

     !-----------------------------------------------------------------
     ! Tolerance met: Exit
     !-----------------------------------------------------------------

      if (roc < lrelax_tolerance) then

       write(1000,*)
       write(1000,*) " Tolerance met. Exit GS relaxation. Total sweeps = ", irelax
       write(1000,*) " tolerance_linear = ", lrelax_tolerance

       exit relax

      endif

    endif after_1st_relaxation

  !-------------------------------------------------
  ! Stop if the maximum number of sweeps is reached.
  !---------------------------------------------

    max_reached : if (irelax == lrelax_sweeps) then

     write(1000,*)
     write(1000,*) " Tolerance not met... Max sweep reached = ", lrelax_sweeps

     exit relax

    endif max_reached

!--------------------------------------------------------------------
! End of 3. Check the linear residual.
!--------------------------------------------------------------------

 end do relax

!--------------------------------------------------------------------

  open(unit=2001, file = "delete_me_linear_res",  status="unknown", position="append")
   write(2001,*) irelax+1

   rewind(1001)
  do i = 0, irelax  !<- includes the initial residual.
   read( 1001,*) i_temp       , temp1, temp2
   write(2001,*) real(i_temp,p2)/real(irelax,p2), temp1, temp2
  end do

  close(1001)
  close(2001)

  open(unit=1002, iostat=stat, file="delete_me2", status='old')
  if (stat == 0) close(1002, status='delete')

!--------------------------------------------------------------------

 return

 end subroutine linear_relaxation_gen
!********************************************************************************

!********************************************************************************
!* This function returns a global vector of residual.
!********************************************************************************
 subroutine compute_linear_res(n,nq,nt,jac_off,jac_dinv,k0,nnghbrs,nghbr,x,rhs, &
                                                            linear_residual  )

 implicit none

 integer , parameter ::   p2 = selected_real_kind(P=15) !Double precision

!Input
 integer ,                      intent( in) :: n, nq, nt
 real(p2), dimension(nt,nq,nq), intent( in) :: jac_off  !i=1,nnodes*nnghbrs
 real(p2), dimension(n ,nq,nq), intent( in) :: jac_dinv !i=1,nnodes
 integer , dimension(n       ), intent( in) :: k0
 integer , dimension(n       ), intent( in) :: nnghbrs
 integer , dimension(nt      ), intent( in) :: nghbr
 real(p2), dimension(n,nq    ), intent( in) :: x
 real(p2), dimension(n,nq    ), intent( in) :: rhs

 real(p2), dimension(n,nq    ), intent(out) :: linear_residual

!Local variable
 real(p2), dimension(nq)   :: b
 integer                   :: i, k

    do i = 1, n

    ! Form the right hand side of GS: b = [ sum( off_diagonal_block*du ) - residual ]
    ! which is just b = -residual since du=0 initially.

        b = -rhs(i,:)
       nghbr_loop : do k = 1, nnghbrs(i)
        b = b - matmul( jac_off(k0(i)+k,:,:), x(nghbr(k0(i)+k),:) )
       end do nghbr_loop

       ! To compute the linear residual norm: A^{-1}*(b-Ax) at each cell.
        linear_residual(i,:) = matmul( jac_dinv(i,:,:), b ) - x(i,:)

    end do

 end subroutine compute_linear_res
!********************************************************************************

!********************************************************************************
!* This function returns rms
!********************************************************************************
 function my_rms(vec,n,nq)

 implicit none

 integer , parameter ::   p2 = selected_real_kind(P=15) !Double precision
 real(p2), parameter :: zero = 0.0_p2

!Input
 integer                  , intent(in) :: n
 integer                  , intent(in) :: nq
 real(p2), dimension(n,nq), intent(in) :: vec

!Output
 real(p2)                              :: my_rms

!Local variable
 integer :: i, k

   my_rms = zero

  do i = 1, n
   do k = 1, nq
    my_rms = my_rms + vec(i,k)**2
   end do
  end do

    my_rms = sqrt( my_rms / real(n*nq,p2) )

 end function my_rms
!********************************************************************************


 end module linear_solver
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
