!********************************************************************************
! Educationally-Designed Unstructured 3D (EDU3D) Code
!
!  --- EDU3D_Poisson
!
! This program reads an unstructured grid (.ugrid) and solves a Poisson equation
! in 3D.
!
!  - It currently has two main options for the Poisson equations: 
!
!    (1)    Linear Poisson: div(nu*grad(u)) = f(x,y,z), nu=1
!    (2) Nonlinear Poisson: div(nu*grad(u)) = f(x,y,z), nu=1+u^2
!
!    with a specific exact solution u = c4 * sin( pi*(c1*x + c2*y + c3*z) ),
!    where (c1,c3,c3,c4) are user specified constants.
!
!  - To solve the Poisson equation : div(nu*grad(u)) = f, with your own nu and f,
!    go to the file "edu3d_poisson_eqs.f90", and implement nu and f in the following
!    subroutines:
!
!     - nu in diff_parameters_ddt(diffusion_type, Lr_factor, u, nu, Lr, Tr )
!     - f in forcing_term( xp, yp, zp )
!
!
!  - Input parameters are controlled by namelist variables specified in 'input.nml'.
!
!
!  - This code discretizes the Poisson eq, by the node-centered edge-based discretization
!    -- Solutions stored at nodes, flux computed at edge-midpoints.
!    -- Numerical solutions are point values, not cell averages.
!    -- 2nd-order accurate on arbitrary tetrahedra, and regular prism/hex/mixed.
!    -- 3rd-order accurate on arbitrary tetrahedra (not implemented).
!    -- It can be considered as finite-volume based on edge-based flux quadrature.
!    -- It can be considered as finite-element based on edge-based flux quadrature.
!    -- It can be considered as conservative finite-difference scheme.
!
!  - Steady solver (nonlinear solver) can be chosen from
!    -- Explicit forward-Euler pseudo-time iteration (SLOW)
!    -- Explicit 2-stage RK pseudo-time iteration (SLOW)
!    -- Implicit defect-correction solver with exact first-order Jacobians (Good)
!    -- Jacobian-Free Newton-Krylov solver with GCR (Recommended)
!
!  - Boundary conditions: all implemented in the weak form through a numerical flux.
!    -- Dirichlet with an exact solution or user specified values.
!    -- Neumann with user specified normal gradient values.
!    -- Strong Dirichlet condition is available but not recomemended when the solution
!       is discontinuous across boundary parts (then, the solution is not unique at
!       an intersection node).
!
!
!  - Latest reference on the node-centered edge-based discretization:
!
!    H. Nishikawa and Y. Liu, " Accuracy-Preserving Source Term Quadrature for
!    Third-Order Edge-Based Discretization", Journal of Computational Physics,
!    Volume 344, September 2017, Pages 595-622.
!    https://doi.org/10.1016/j.jcp.2017.04.075
!    http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_liu_jcp2017_preprint.pdf
!
!  - Example cases are included in the package with grid generation codes:
!
!    (1)"case_accuracy_test_cube_hexa_nonuniform"
!        Discretization error test on nonuniform hexa gridss (for verification).
!       (1.1) "discretization_error_test_conventional"       : conventional +  weak Dirichlet
!       (1.2) "discretization_error_test_hyperbolic"         :   hyperbolic +  weak Dirichlet
!       (1.3) "discretization_error_test_conventional_strong": conventional + strong Dirichlet
!       (1.4) "discretization_error_test_hyperbolic_strong"  :   hyperbolic + strong Dirichlet

!    (2)"case_accuracy_test_cube_tetra_irregular"
!        Discretization error test on irregular tetra gridss (for verification).
!       (2.1) "discretization_error_test_conventional"       : conventional +  weak Dirichlet
!       (2.2) "discretization_error_test_hyperbolic"         :   hyperbolic +  weak Dirichlet
!       (2.3) "discretization_error_test_conventional_strong": conventional + strong Dirichlet
!       (2.4) "discretization_error_test_hyperbolic_strong"  :   hyperbolic + strong Dirichlet
!
!    (3)"case_cube_hexa_nonuniform"
!        A test case with the Dirichlet condition with user specified BC values on hex grids.
!       (3.1) "case_conventional"       : conventional +  weak Dirichlet
!       (3.2) "  case_hyperbolic"       :   hyperbolic +  weak Dirichlet
!
!    (4)"case_cube_tetra_irregular"
!        A test case with the Dirichlet condition with user specified BC values on tetra grids.
!       (4.1) "case_conventional"       : conventional +  weak Dirichlet
!       (4.2) "  case_hyperbolic"       :   hyperbolic +  weak Dirichlet
!
!    (5)"case_sphere_tetra"
!        A test case with the Dirichlet condition with user specified BC values on tetra grids.
!       (5.1) "case_conventional"       : conventional +  weak Dirichlet
!       (5.2) "  case_hyperbolic"       :   hyperbolic +  weak Dirichlet
!
!    Please ask Hiro (hiro@nianet.org) about any other example or any question.
!
!
!  - Things that are missing and you might want to implement by yourself.
!
!    -- CFL ramping (gradually increase CFL from the start of iteration).
!    -- Extend to unsteady problem (e.g., add BDF2 term).
!    -- Extend to other types of elements, prism and pyramid
!    -- HANIM for a better control of the JFNK solver.
!    -- Node reordering for efficiency
!    -- Grid coloring for multi-color Gauss-Seidel
!    -- Other grid file formats (other than just .ugrid).
!    -- Other output data file formats (other than just Tecplot)
!    -- Parallelization for faster computation
!    -- Other nonlinear solvers (e.g., multigrid)
!    -- 3rd-order edge-based method (quadratic-LSQ/extrapolation/source).
!    -- Grid adaptation capability
!
!  - This is a serial code. Could you parallelize it? How about OpenMP?
!    It'd be great if someone parallelized this code and shared it with everyone.
!
!
!        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!
! the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!
! This is Version 1.1 (2021).
!
! 06-21-2021: Corrected and simplified the BC implementation.
!             Added a capablity to write out a solution file.
!             Try by adding, generate_soln_file=T, to the input file.
!             Also added opitons to normalize the solution at the end,
!             and to fix the solution at a single node.
!
! 06-16-2021: Fixed some bugs. Now computes the Jacobian just once for linear Poisson.
!
!
! This F90 program is written and made available for an educational purpose.
!
! This file may be updated in future.
!
! Katate Masatsuka, June 2021. http://www.cfdbooks.com
!
! Notes:
!
!  The purpose of this code is to give a beginner an opportunity to learn how to
!  write an unstructured CFD code. Hence, the focus here is on the simplicity.
!  The code is not optimized for efficiency.
!
!  If the code is not simple enough to understand, please send questions to Hiro
!  at sunmasen(at)hotmail.com. He'll greatly appreciate it and revise the code.
!
!  If the code helps you understand how to write your own code that is more
!  efficient and has more features, it'll have served its purpose.
!
!********************************************************************************

 program edu3d_poisson

 ! This main program contains the following subroutines:
 !
 !  - set_filenames                ! define file names
 !  - allocate_nceb_arrays         ! allocate arrays for the edge-based scheme
 !  - set_initial_solution         ! set up an initial solution
 !  - read_grid                    ! read a grid file
 !  - write_tecplot_boundary_file  ! write a tecplot boundary grid file
 !  - write_tecplot_volume_file    ! write a tecplot volume grid file
 !  - big_endian_io                ! detect endian

  use input_parameter, only : read_nml_input_parameters   , &
                                    generate_tec_file_b   , &
                                    generate_tec_file_v   , &
                                    generate_soln_file    , &
                                    accuracy_test         , &
                                    solver_max_itr        , &
                                    accuracy_test_te      , &
                                    bc_neumann_read_uxuyuz, &
                                    soln_normalization

  use nceb_data      , only : construct_nceb_data
  use steady_solver  , only : solve_steady
  use poisson_eqns   , only : compute_ref_length

  implicit none

  integer , parameter      :: p2 = selected_real_kind(P=15)
  real(p2)                 :: wtime
  integer                  :: itr

   write(*,*)
   write(*,*) "----------------------------------------------------------------"
   write(*,*)
   write(*,*) "  This is EDU3D Poisson."
   write(*,*)
   write(*,*) "----------------------------------------------------------------"
   write(*,*)

!*******************************************************************************
!
! Read the input parameters, defined in the file named as 'input_coarsen.nml'.
!
!*******************************************************************************

   write(*,*) "Reading the input file: input.nml..... "
   write(*,*)
   call read_nml_input_parameters("input.nml")
   write(*,*)

!-------------------------------------------------------------------------------
! 1. Set file names

  if (accuracy_test) then
   call accuracy_test_main
  endif

!-------------------------------------------------------------------------------
! 1. Set file names

  if (accuracy_test_te) then
   call accuracy_test_main_te
  endif

!-------------------------------------------------------------------------------
! 1. Set file names
  call set_filenames

!-------------------------------------------------------------------------------
! 2. Read a grid file (.ugrid) and a bc file (.bcmap)

  call read_grid

!-------------------------------------------------------------------------------
! 3. Construct grid data required for the node-centered edge-based discretization.

  call construct_nceb_data

!-------------------------------------------------------------------------------
! 4. Allocate arrays: solution, residual, jacobian, etc.

  call allocate_nceb_arrays

!-------------------------------------------------------------------------------
! 5. Set up the initial solution with/without random perturbation.

  call set_initial_solution

!-------------------------------------------------------------------------------
! 6. Read gradient data files for Neumann conditions (special feature)

  if (bc_neumann_read_uxuyuz) call read_uxuyuz

!-------------------------------------------------------------------------------
! 7. Compute the reference length (required by a hyperbolic Poisson solver).

  call compute_ref_length

!-------------------------------------------------------------------------------
! 8. Compute the solution.

  !Compute the steady soution by solving Res(U)=0.

      call  solve_steady( solver_max_itr(1), itr, wtime )

  !Compute the unsteady soution by solving dU/dt+Res(U)=0:

  !   call  solve_unsteady (not implemented...)

!-------------------------------------------------------------------------------
! 9. Solution normalization: subtract the average from all if requested.

   if (trim(soln_normalization) == "min_2norm") call solution_minimum_2norm

!-------------------------------------------------------------------------------
! 10. Write output files for visualization

  !Tecplot boundary grid data file if requested:
   if (generate_tec_file_b) call write_tecplot_boundary_file

  !Tecplot volume grid data file if requested:
   if (generate_tec_file_v) call write_tecplot_volume_file

  !Write a solution file
   if (generate_soln_file)  call write_soln_file

   write(*,*) 
   write(*,*) " See fort.1000 for GS relaxation convergence history (implicit/jfnk)."
   write(*,*) " See fort.2000 for GCR convergence history (jfnk)."
   write(*,*)

   write(*,*)
   write(*,*) "----------------------------------------------------------------"
   write(*,*)
   write(*,*) "  End of EDU3D Poisson. "
   write(*,*)
   write(*,*) "----------------------------------------------------------------"
   write(*,*)

 stop

 contains


!********************************************************************************
!
! Special feature: Read the gradient data to be used in the Neumann BC.
!
!********************************************************************************
 subroutine read_uxuyuz

  use input_parameter, only : bc_neumann_file_ux , bc_neumann_file_uy, bc_neumann_file_uz
  use data_module    , only : nnodes, ux_bc, uy_bc, uz_bc

 implicit none

   write(*,*)
   write(*,*) " Neumann BC values are provided in the following data files:"
   write(*,*) "     bc_neumann_file_ux = ", trim(bc_neumann_file_ux)
   write(*,*) "     bc_neumann_file_uy = ", trim(bc_neumann_file_uy)
   write(*,*) "     bc_neumann_file_uz = ", trim(bc_neumann_file_uz)
   write(*,*)
   write(*,*) " Reading the files......."
   write(*,*)


   allocate( ux_bc(nnodes), uy_bc(nnodes), uz_bc(nnodes) )

 !-----------------------------------------------------------------------
 ! Assume binary for now...

  open( 1, file=bc_neumann_file_ux,form="unformatted",status="unknown",action="read",access='stream')
  read( 1) ux_bc
  close(1)

  open( 1, file=bc_neumann_file_uy,form="unformatted",status="unknown",action="read",access='stream')
  read( 1) uy_bc
  close(1)

  open( 1, file=bc_neumann_file_uz,form="unformatted",status="unknown",action="read",access='stream')
  read( 1) uz_bc
  close(1)

   write(*,*) " Finished reading the files......."
   write(*,*)

 end subroutine read_uxuyuz

!********************************************************************************
!
! Define file names.
!
!********************************************************************************
 subroutine set_filenames

  use input_parameter, only : project, ugrid_unformatted
  use data_module           , only : filename_ugrid    , filename_bcmap
  use data_module           , only : filename_tecplot_b, filename_tecplot_v
  use data_module           , only : filename_tecplot_hist

  use data_module           , only : filename_soln_u , filename_soln_ux, &
                                     filename_soln_uy, filename_soln_uz

 implicit none

 !-----------------------------------------------------------------------
 ! Input grid file (.ugrid)

   !(1) Binary file
    if (ugrid_unformatted) then

      if ( big_endian_io(9999) ) then
       filename_ugrid     = trim(project) // '.b8.ugrid'
      else
       filename_ugrid     = trim(project) // '.l8.ugrid'
      end if

   !(2) ASCII file
    else

       filename_ugrid     = trim(project) // '.ugrid'

    endif

 !-----------------------------------------------------------------------
 ! Input boundary condition file (ASCII file)

      filename_bcmap     = trim(project) // '.bcmap'

 !-----------------------------------------------------------------------
 ! Output: Tecplot boundary grid file (ASCII file)

      filename_tecplot_b = trim(project) // '_boundary_tec.dat'

 !-----------------------------------------------------------------------
 ! Output: Tecplot volume grid file (ASCII file)

      filename_tecplot_v = trim(project) // '_volume_tec.dat'

 !-----------------------------------------------------------------------
 ! Output: Iterative convergence history file (ASCII Tecplot file)

      filename_tecplot_hist = trim(project) // '_hist_tec.dat'

 !-----------------------------------------------------------------------
 ! Output: Soluton data file (ASCII file)

      filename_soln_u  = trim(project) // '_u.data'
      filename_soln_ux = trim(project) // '_ux.data'
      filename_soln_uy = trim(project) // '_uy.data'
      filename_soln_uz = trim(project) // '_uz.data'

 !-----------------------------------------------------------------------

 end subroutine set_filenames


!********************************************************************************
!
! Allocate arrays required for the node-centered edge-based discretization,
! and nonlinear solvers.
!
!********************************************************************************
 subroutine allocate_nceb_arrays

 use data_module, only : nnodes, nghbrs_total, u, gradu, dtau, wsn, res
 use data_module, only : du, u0, jac_diag, jac_dinv, jac_off

 use input_parameter, only : solver_type

 implicit none

! Solution, gradients, and residual arrays.

  allocate(     u(nnodes,4  ) )
  allocate(    du(nnodes,4  ) )
  allocate( gradu(nnodes,4,3) )
  allocate(   res(nnodes,4  ) )

  allocate(  dtau(nnodes    ) )
  allocate(   wsn(nnodes    ) )

! Extra arrays required by nonlinear solvers

  if (     trim(solver_type) == "fwd_euler" ) then

  ! No extra arrays are needed.

  elseif ( trim(solver_type) == "tvdrk2" ) then

   allocate( u0(nnodes,4) )               !to save the solution for 2-stage RK.

  elseif ( trim(solver_type) == "implicit" ) then

   allocate( jac_diag(      nnodes,4,4) ) !jacobian: diagonal blocks
   allocate( jac_dinv(      nnodes,4,4) ) !jacobian: inverse of diagonal blocks
   allocate( jac_off( nghbrs_total,4,4) ) !jacobian: off-diagonal blocks 

  elseif ( trim(solver_type) == "jfnk" ) then

   allocate( jac_diag(      nnodes,4,4) ) !jacobian: diagonal blocks
   allocate( jac_dinv(      nnodes,4,4) ) !jacobian: inverse of diagonal blocks
   allocate( jac_off( nghbrs_total,4,4) ) !jacobian: off-diagonal blocks 

  else

   write(*,*) " Invalid input: solver_type = ", trim(solver_type)
   write(*,*) "   Available options: fwd_euler, tvdrk2, implciit, jfnk. Stop..."
   stop

  endif

 end subroutine allocate_nceb_arrays

!********************************************************************************
!
! Allocate arrays required for the node-centered edge-based discretization,
! and nonlinear solvers.
!
!********************************************************************************
 subroutine deallocate_nceb_arrays

 use data_module, only : u, gradu, dtau, wsn, res
 use data_module, only : du, u0, jac_diag, jac_dinv, jac_off, ux_bc, uy_bc, uz_bc

 use input_parameter, only : solver_type

 implicit none

! Solution, gradients, and residual arrays.

  deallocate(     u )
  deallocate(    du )
  deallocate( gradu )
  deallocate(   res )

  deallocate(  dtau )
  deallocate(   wsn )
 
  if (allocated(ux_bc)) deallocate(ux_bc)
  if (allocated(uy_bc)) deallocate(uy_bc)
  if (allocated(uz_bc)) deallocate(uz_bc)

! Extra arrays required by nonlinear solvers

  if (     trim(solver_type) == "fwd_euler" ) then

  elseif ( trim(solver_type) == "tvdrk2" ) then

   deallocate( u0 )

  elseif ( trim(solver_type) == "implicit" ) then

   deallocate( jac_diag )
   deallocate( jac_dinv )
   deallocate( jac_off  )

  elseif ( trim(solver_type) == "jfnk" ) then

   deallocate( jac_diag )
   deallocate( jac_dinv )
   deallocate( jac_off  )

  else

   write(*,*) " Invalid input: solver_type = ", trim(solver_type)
   write(*,*) "   Available options: fwd_euler, tvdrk2, implciit, jfnk. Stop..."
   stop

  endif

 end subroutine deallocate_nceb_arrays

!********************************************************************************
!
! Deallocate grid data.
!
!********************************************************************************
 subroutine deallocate_grid_data

  use data_module, only : x, y, z, bc_type
  use data_module, only :  tet,  hex,  tria,  quad
  use data_module, only : ntet, nhex

  use data_module    , only : vol                   , & !nodal coords, dual vol
                              edge                  , & !edge data
                              n12, n12_mag          , & !directed area vector
                              k0, nnghbrs, nghbr    , & !node nghbr data
                              kth_nghbr_of_1        , & !nghbr index
                              kth_nghbr_of_2        , & !nghbr index
                              bn, bn_mag            , & !boundary normals
                              e12, e12_mag, bmark   , & !Edge vector, boundary mark
                              ar, arc               , & !Aspect ratios
                              e_dot_n_node, e_dot_n_edge, color

 use data_module, only : linear_lsq3x3_cx, linear_lsq3x3_cy, linear_lsq3x3_cz

 implicit none

  deallocate( x, y, z, bc_type )

  if (ntet > 0) deallocate (tet)
  if (nhex > 0) deallocate (hex)
  
  deallocate (tria)
  deallocate (quad)

  deallocate( vol, edge, n12, n12_mag , k0, nnghbrs, nghbr, kth_nghbr_of_1, kth_nghbr_of_2, bn, bn_mag )
  deallocate( e12, e12_mag, bmark, ar, arc,  e_dot_n_node, e_dot_n_edge, color )
 
  deallocate( linear_lsq3x3_cx, linear_lsq3x3_cy, linear_lsq3x3_cz )

 end subroutine deallocate_grid_data

!********************************************************************************
!
! Set the free stream values, and initial solution.
!
!********************************************************************************
 subroutine set_initial_solution

 use data_module           , only : p2, one, zero
 use data_module           , only : nnodes, u, bmark

 use poisson_eqns     , only : iu, iux, iuy, iuz

 use input_parameter, only : perturb_initial

 implicit none

 integer  :: i
 real(p2) :: rn

  write(*,*)
  write(*,*) "---------------------------------------"
  write(*,*) " Set initial solution"
  write(*,*)

  if (perturb_initial==1) then
    write(*,*) "     perturb_initial = ", perturb_initial
    write(*,*) "  -> Add random perturbation in the y-velocity. "
  endif

! Let the initial solution be the free streasome uniform values.

  loop_node : do i = 1, nnodes

   u(i,iu ) =  one
   u(i,iux) = zero !d(one)/dx=0
   u(i,iuy) = zero !d(one)/dy=0
   u(i,iuz) = zero !d(one)/dz=0

  !Add random perturbation in the solution if perturb_initial=1.
   if (perturb_initial==1) then

    call random_number(rn)

    if ( bmark(i) == 0 ) u(i,iu ) =  u(i,iu ) * ( one + 0.25_p2 * rn )! interior nodes only

   endif

  end do loop_node

 end subroutine set_initial_solution

!********************************************************************************
!* Read the grid and boundary condition file.
!*
!* Note: Currently, the EDU3D-Poisson works only for pure tetrahedral grids.
!*       I think you can modify the code to make it work for other elements.
!*
!********************************************************************************
!*
!* 1. "datafile_grid_in": .ugrid file name. 
!*
!*
!* 2. "datafile_bcmap_in" is assumed have been written in the following format:
!*
!*   -----------------------------------------------------------------------
!*    write(*,*) nb (# of boundaries)
!*    do i = 1, nb
!*     write(*,*) i, bc_type
!*    end do
!*   -----------------------------------------------------------------------
!*
!*   NOTE: bc_type is the name of the boundary condition, e.g.,
!*
!*         1. "freestream"
!*             Roe flux with freestream condition on the right state.
!*
!*         2. "outflow_supersonic"
!*             Roe flux with the interior state as the right state.
!*             (equivalent to the interior-extrapolation condition.)
!*
!*         3. "slip_wall"
!*             Right state with zero mass flux through the boundary.
!*
!*         4. "outflow_subsonic_p0"
!*             Fix the back pressure. This should work for subsonic flows in a
!*             large enough domain.
!*
!*         !You can implement more BCs.
!*
!*
!********************************************************************************
!* Data to be read and stored:
!*
!* 1. Some numbers
!*    nnodes        = number of nodes
!*    ntria         = number of triangular boundary elements
!*    nquad         = number of quadrilateral boundary elements
!*    ntet          = number of tetrahedra
!*    npyr          = number of pyramids
!*    nprs          = number of prims
!*    nhex          = number of hexahedra
!*
!* 2. Boundary element data:
!*    tria = list of vertices of each triangular boundary element
!*    quad = list of vertices of each quadrilateral boundary element
!*
!* 3. Volume element data:
!*    tet  = list of vertices of each tetrahedron
!*    pyr  = list of vertices of each pyramid
!*    prs  = list of vertices of each prism
!*    hex  = list of vertices of each hexehedron
!*
!* 4. x, y, z coordinates of ndoes
!*    x    = x-coordinate of the nodes
!*    y    = y-coordinate of the nodes
!*    z    = z-coordinate of the nodes
!*
!* 5. Boundary Data:
!*    nb      = number of boundary groups
!*    bc_type = boundary condition name
!*
!********************************************************************************
 subroutine read_grid

 !Input
  use input_parameter, only : ugrid_unformatted, scale_factor, & 
                                    x_shift, y_shift, z_shift, &
                                       bc_neumann_read_uxuyuz

 !Fill in these data:
  use data_module    , only : nnodes            , & !# of nodes
                              x, y, z           , & !nodal coords
                              ntet, tet         , & !tet elements
                              nhex, hex         , & !hex elements
                              ntria, tria       , & !boundary elements (tria)
                              nquad, quad       , & !boundary elements (quad)
                              nb, bc_type           !# of boundaries, and BCs

 !Input file names
  use data_module    , only : filename_ugrid
  use data_module    , only : filename_bcmap

 implicit none

 !Data not used in the code at the moment (tetrahedra and hexahedra only).

 integer                           :: nprs
 integer , dimension(:,:), pointer :: prs

 integer                           :: npyr
 integer , dimension(:,:), pointer :: pyr

 integer :: i, dummy_int, os

  write(*,*)
  write(*,*) "---------------------------------------"
  write(*,*) " Reading : ", trim(filename_ugrid)
  write(*,*)

 !------------------------------------------------------------------------------
 !(1)Read an unformatted file.

  unformatted : if ( ugrid_unformatted ) then

   open(unit=2, file=filename_ugrid, form='unformatted',access="stream", &
                                               status="unknown", iostat=os )
   read(2) nnodes, ntria, nquad, ntet, npyr, nprs, nhex

  write(*,*) "          nnodes = ", nnodes
  write(*,*) "          prisms = ", nprs
  write(*,*) "           tetra = ", ntet
  write(*,*) "            hexa = ", nhex
  write(*,*) "         pyramid = ", npyr
  write(*,*) "       triangles = ", ntria
  write(*,*) "  quadrilaterals = ", nquad

  if (ntria >0) allocate(tria(ntria,4))
  if (nquad >0) allocate(quad(nquad,5))
  if (ntet  >0) allocate(  tet(ntet,4))
  if (npyr  >0) allocate(  pyr(nprs,5))
  if (nprs  >0) allocate(  prs(nprs,6))
  if (nhex  >0) allocate(  hex(nhex,8))
                allocate(x(nnodes), y(nnodes), z(nnodes))

   do i = 1, nnodes

    read(2) x(i), y(i), z(i)

    x(i) = ( x(i)-x_shift ) * scale_factor
    y(i) = ( y(i)-y_shift ) * scale_factor
    z(i) = ( z(i)-z_shift ) * scale_factor

   end do

  ! Triangular boundary faces
    if (ntria>0) then
     do i = 1, ntria
      read(2) tria(i,1), tria(i,2), tria(i,3)
     end do
    endif

  ! Quad boundary faces
    if (nquad>0) then
     do i = 1, nquad
      read(2) quad(i,1), quad(i,2), quad(i,3), quad(i,4)
     end do
    endif

  ! Face tag: Boundary group number
    if (ntria>0) then
     do i = 1, ntria
      read(2) tria(i,4)
     end do
    endif

    if (nquad>0) then
     do i = 1, nquad
      read(2) quad(i,5)
     end do
    endif

  ! Tets
    if (ntet>0) then
     do i = 1, ntet
     read(2) tet(i,1), tet(i,2), tet(i,3), tet(i,4)
     end do
    endif

  ! Pyramids
    if (npyr>0) then

     write(*,*) " Well, sorry pyramids are not currently considered in this code... Stop"
     stop

     do i = 1, npyr
     read(2) pyr(i,1), pyr(i,2), pyr(i,3), pyr(i,4), pyr(i,5)
     end do
    endif

  ! Prisms
    if (nprs>0) then

     write(*,*) " Well, sorry prisms are not currently considered in this code... Stop"
     stop

     do i = 1, nprs
     read(2) prs(i,1), prs(i,2), prs(i,3), prs(i,4), prs(i,5), prs(i,6)
     end do
    endif

  ! Hexahedra
    if (nhex>0) then
     do i = 1, nhex
     read(2) hex(i,1), hex(i,2), hex(i,3), hex(i,4), hex(i,5), hex(i,6), hex(i,7), hex(i,8)
     end do
    endif

 !------------------------------------------------------------------------------
 !(2)Read a formatted file.

  else

   open(unit=2, file=filename_ugrid, status="unknown", iostat=os)
   read(2,*) nnodes, ntria, nquad, ntet, npyr, nprs, nhex

  write(*,*) "          nnodes = ", nnodes
  write(*,*) "          prisms = ", nprs
  write(*,*) "           tetra = ", ntet
  write(*,*) "            hexa = ", nhex
  write(*,*) "         pyramid = ", npyr
  write(*,*) "       triangles = ", ntria
  write(*,*) "  quadrilaterals = ", nquad

  if (ntria >0) allocate(tria(ntria,4))
  if (nquad >0) allocate(quad(nquad,5))
  if (ntet  >0) allocate(  tet(ntet,4))
  if (npyr  >0) allocate(  pyr(nprs,5))
  if (nprs  >0) allocate(  prs(nprs,6))
  if (nhex  >0) allocate(  hex(nhex,8))
                allocate(x(nnodes), y(nnodes), z(nnodes))

   do i = 1, nnodes

    read(2,*) x(i), y(i), z(i)

!    x(i) = x(i) * scale_factor
!    y(i) = y(i) * scale_factor
!    z(i) = z(i) * scale_factor

    x(i) = ( x(i)-x_shift ) * scale_factor
    y(i) = ( y(i)-y_shift ) * scale_factor
    z(i) = ( z(i)-z_shift ) * scale_factor

   end do

  ! Triangular boundary faces
    if (ntria>0) then
     do i = 1, ntria
      read(2,*) tria(i,1), tria(i,2), tria(i,3)
     end do
    endif

  ! Quad boundary faces
    if (nquad>0) then
     do i = 1, nquad
      read(2,*) quad(i,1), quad(i,2), quad(i,3), quad(i,4)
     end do
    endif

  ! Face tag: Boundary group number
    if (ntria>0) then
     do i = 1, ntria
      read(2,*) tria(i,4)
     end do
    endif

    if (nquad>0) then
     do i = 1, nquad
      read(2,*) quad(i,5)
     end do
    endif

  ! Tets
    if (ntet>0) then
     do i = 1, ntet
     read(2,*) tet(i,1), tet(i,2), tet(i,3), tet(i,4)
     end do
    endif

  ! Pyramids
    if (npyr>0) then

     write(*,*) " Well, sorry pyramids are not currently considered in this code... Stop"
     stop

     do i = 1, npyr
     read(2,*) pyr(i,1), pyr(i,2), pyr(i,3), pyr(i,4), pyr(i,5)
     end do
    endif

  ! Prisms
    if (nprs>0) then

     write(*,*) " Well, sorry prisms are not currently considered in this code... Stop"
     stop

     do i = 1, nprs
     read(2,*) prs(i,1), prs(i,2), prs(i,3), prs(i,4), prs(i,5), prs(i,6)
     end do
    endif

  ! Hexahedra
    if (nhex>0) then
     do i = 1, nhex
     read(2,*) hex(i,1), hex(i,2), hex(i,3), hex(i,4), hex(i,5), hex(i,6), hex(i,7), hex(i,8)
     end do
    endif

  endif unformatted
 !------------------------------------------------------------------------------

  close(2)

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! 2. Read the boundary condition data file

   write(*,*)
   write(*,*) "---------------------------------------"
   write(*,*) " Reading the boundary condition file....", trim(filename_bcmap)
   write(*,*)

! Open the input file.
  open(unit=2, file=filename_bcmap, status="unknown", iostat=os)

    read(2,*) nb

  if (ntria == 0) then
   allocate(tria(1,5))
   tria = 0
  endif

  if (nquad == 0) then
   allocate(quad(1,5))
   quad = 0
  endif

  if ( nb /= max( maxval(tria(:,4)), maxval(quad(:,5)) ) ) then
   write(*,*) " Error in bcmap file..." 
   stop
  endif

  allocate(bc_type(nb))

! READ: Read the boundary condition type. bc_type is character(80).
   do i = 1, nb
    read(2,*) dummy_int, bc_type(i)
    if (bc_neumann_read_uxuyuz) then
     bc_type(i) = "neumann" !Overwrite it if data are given.
     write(*,*) i, "  bc_type(i) is overwritten as ", trim(bc_type(i))
    endif
   end do

!  Print the data
    write(*,*)
   do i = 1, nb
    write(*,'(a10,i3,a12,a35)') " boundary", i, "  bc_type = ", trim(bc_type(i))
   end do

    write(*,*)

  close(2)

! End of Read the boundary condition data file
!--------------------------------------------------------------------------------

 end subroutine read_grid

!*******************************************************************************
!
! This subroutine writes a Tecplot file for boundaries.
!
!*******************************************************************************
 subroutine write_tecplot_boundary_file

  use data_module   , only : x, y, z     , & !nodal coords
                             u           , & !solution variables u=(u,ux,uy,uz)
                             ntria, tria , & !boundary elements
                             nquad, quad , & !boundary elements
                             nnodes      , & !number of nodes
                             nb          , & !number of boundaries
                             bc_type, gradu      !BC type name

 use poisson_eqns     , only : iu, iux, iuy, iuz, exact_sol

 use gradients        , only : compute_gradient

 use input_parameter    , only : diffusion_type, Lr_factor
 use poisson_eqns         , only : diff_parameters

! Output tecplot boundary filename
  use data_module               , only : filename_tecplot_b

 Implicit none

 integer , parameter :: p2 = selected_real_kind(15) !Double precision

!Input

 character(80)                         :: boundary_name
 integer                               :: nelms
 integer                               :: nnodes_loc, i, j, k, ib, os
 integer,  dimension(:,:), allocatable :: g2l
 integer,  dimension(:  ), allocatable :: l2g

  integer                :: ix=1, iy=2, iz=3
 real(p2), dimension(4) :: ue

 real(p2) :: nu,Lr,Tr 

 write(*,*)
 write(*,*) ' Tecplot boundary file = ', trim(filename_tecplot_b)
 write(*,*)

!Allocate global-to-local and local-to-global arrays.

 allocate(g2l(nnodes,2))
 allocate(l2g(nnodes  ))

 open(unit=7, file=filename_tecplot_b, status="unknown", iostat=os)
 write(7,*) 'TITLE = "GRID"'
 write(7,*) 'VARIABLES = "x","y","z","u","ux","uy","uz","uexact","uxexact","uyexact","uzexact",', &
            '"ux_lsq","uy_lsq","uz_lsq"'

!Let's recompute the LSQ gradients.

  call compute_gradient

!Write boundaries separately.

 if (ntria > 0) then

 tria_boundary_loop : do ib = 1, nb

!---------------------------------------------------------------------------
! Triangular boundaries.

       nelms = 0
  nnodes_loc = 0
         g2l = -1
         l2g = 0

 !To save memory, we write the boundary data based on the data
 !local to each boundary. Local numbers are assigned to each
 !node on the boundary, which is accessed by g2l (global to local).

  do i = 1, ntria

   if ( tria(i,4) == ib ) then

    nelms = nelms + 1

    do k = 1, 3
     if (g2l(tria(i,k),2) == -1) then
      nnodes_loc = nnodes_loc + 1
      g2l(tria(i,k),1) = nnodes_loc
      g2l(tria(i,k),2) = 100         !Flag to indicate the node is recorded.
      l2g(nnodes_loc)  = tria(i,k)
     endif
    end do

   endif

  end do

 !So, there are nnodes_loc nodes and nelms elements on this boundary (ib).

  if (nelms > 0) then
   boundary_name = trim(bc_type(ib)) // " (triangles)"
   write(7,*) 'ZONE T="', trim(boundary_name), '"  N=', nnodes_loc, &
              ', E=', nelms, ' , ET=quadrilateral, F=FEPOINT'

  !Write only the local nodes.
   do j = 1, nnodes_loc
     i = l2g(j)

     call exact_sol( x(i), y(i), z(i), ue(1),ue(2),ue(3),ue(4) )

     call diff_parameters( diffusion_type,Lr_factor,u(i,iu), nu,Lr,Tr )

     write(7,'(14es25.15)') x(i), y(i), z(i)                     , &
                            u(i,iu), u(i,iux), u(i,iuy), u(i,iuz), &
                            ue(1),ue(2),ue(3),ue(4)              , &
                            gradu(i,iu,ix), gradu(i,iu,iy), gradu(i,iu,iz)
   end do

  !Write only the local elements.
   do i = 1, ntria
    if ( tria(i,4) == ib ) write(7,'(4i10)') g2l(tria(i,1),1), g2l(tria(i,2),1), &
                                             g2l(tria(i,3),1), g2l(tria(i,3),1)
   end do
  endif

 end do tria_boundary_loop

 endif

!---------------------------------------------------------------------------


 if ( nquad > 0 ) then

 quad_boundary_loop : do ib = 1, nb

!---------------------------------------------------------------------------
! Triangular boundaries.

       nelms = 0
  nnodes_loc = 0
         g2l = -1
         l2g = 0

 !To save memory, we write the boundary data based on the data
 !local to each boundary. Local numbers are assigned to each
 !node on the boundary, which is accessed by g2l (global to local).

  do i = 1, nquad

   if ( quad(i,5) == ib ) then

    nelms = nelms + 1

    do k = 1, 4
     if (g2l(quad(i,k),2) == -1) then
      nnodes_loc = nnodes_loc + 1
      g2l(quad(i,k),1) = nnodes_loc
      g2l(quad(i,k),2) = 100         !Flag to indicate the node is recorded.
      l2g(nnodes_loc)  = quad(i,k)
     endif
    end do

   endif

  end do

 !So, there are nnodes_loc nodes and nelms elements on this boundary (ib).

  if (nelms > 0) then
   boundary_name = trim(bc_type(ib)) // " (quadrilaterals)"
   write(7,*) 'ZONE T="', trim(boundary_name), '"  N=', nnodes_loc, &
              ', E=', nelms, ' , ET=quadrilateral, F=FEPOINT'

  !Write only the local nodes.
   do j = 1, nnodes_loc
     i = l2g(j)

     call exact_sol( x(i), y(i), z(i), ue(1),ue(2),ue(3),ue(4) )

     call diff_parameters( diffusion_type,Lr_factor,u(i,iu), nu,Lr,Tr )

     write(7,'(14es25.15)') x(i), y(i), z(i)                     , &
                            u(i,iu), u(i,iux), u(i,iuy), u(i,iuz), &
                            ue(1),ue(2),ue(3),ue(4)              , &
                            gradu(i,iu,ix), gradu(i,iu,iy), gradu(i,iu,iz)
   end do

  !Write only the local elements.
   do i = 1, nquad
    if ( quad(i,5) == ib ) write(7,'(4i10)') g2l(quad(i,1),1), g2l(quad(i,2),1), &
                                             g2l(quad(i,3),1), g2l(quad(i,4),1)
   end do
  endif

 end do quad_boundary_loop

 endif

!---------------------------------------------------------------------------




 close(7)

  deallocate( g2l, l2g )

 end subroutine write_tecplot_boundary_file
!********************************************************************************

!*******************************************************************************
! This subroutine writes a Tecplot file for the volume grid.
! In this program, only tetrahedral grids are considered.
!*******************************************************************************
 subroutine write_tecplot_volume_file

  use data_module    , only : x, y, z     , & !nodal coords
                              u           , & !solution variables u=(u,ux,uy,uz)
                              ntet, tet   , & !number of tetra, and tetra info
                              nhex, hex   , & !number of tetra, and tetra info
                              nnodes, gradu          !number of nodes

  use poisson_eqns     , only : iu, iux, iuy, iuz, exact_sol

  use gradients        , only : compute_gradient

! Output tecplot volume filename
  use data_module               , only : filename_tecplot_v

  implicit none

  integer , parameter :: p2 = selected_real_kind(15) !Double precision

  integer :: i, os

  integer                :: ix=1, iy=2, iz=3
  real(p2), dimension(4) :: ue

 call compute_gradient

 write(*,*)
 write(*,*) ' Tecplot volume file = ', trim(filename_tecplot_v)
 write(*,*)

  open(unit=8, file=filename_tecplot_v, status="unknown", iostat=os)
  write(8,*) 'TITLE = "GRID"'
  write(8,*) 'VARIABLES = "x","y","z","u","ux","uy","uz","uexact","uxexact","uyexact","uzexact",', &
             '"ux_lsq","uy_lsq","uz_lsq"'

!---------------------------------------------------------------------------
! Tetrahedral zone

  if (ntet > 0) then

   write(8,*) 'zone  n=', nnodes,',e=', ntet,' , et=tetrahedron, f=fepoint'
   do i = 1, nnodes

     call exact_sol( x(i), y(i), z(i), ue(1),ue(2),ue(3),ue(4) )
     write(8,'(14es25.15)')           x(i),           y(i),           z(i),           &
                                   u(i,iu),       u(i,iux),       u(i,iuy), u(i,iuz), &
                                     ue(1),          ue(2),          ue(3),    ue(4), &
                            gradu(i,iu,ix), gradu(i,iu,iy), gradu(i,iu,iz)
   end do

   do i = 1, ntet
    write(8,'(4i10)') tet(i,1), tet(i,2), tet(i,3), tet(i,4)
   end do

  endif

!---------------------------------------------------------------------------

!---------------------------------------------------------------------------
! Hex zone

  if (nhex > 0) then

   write(8,*) 'zone  n=', nnodes,',e=', nhex,' , et=brick, f=fepoint'
   do i = 1, nnodes

     call exact_sol( x(i), y(i), z(i), ue(1),ue(2),ue(3),ue(4) )
     write(8,'(14es25.15)')           x(i),           y(i),           z(i),           &
                                   u(i,iu),       u(i,iux),       u(i,iuy), u(i,iuz), &
                                     ue(1),          ue(2),          ue(3),    ue(4), &
                            gradu(i,iu,ix), gradu(i,iu,iy), gradu(i,iu,iz)
   end do

   do i = 1, nhex
    write(8,'(8i10)') hex(i,1), hex(i,2), hex(i,3), hex(i,4), hex(i,5), hex(i,6), hex(i,7), hex(i,8)
   end do

  endif

!---------------------------------------------------------------------------

 close(8)

 end subroutine write_tecplot_volume_file
!********************************************************************************

!*******************************************************************************
!
! This subroutine alters the solution to minimize the 2-norm by subtracting
! the average from all values.
!
! Note: You might want to use this when you solve Laplace/Poisson equations
!       with Neumann BCs everywhere: e.g., add the following to inpu.nml,
!
!      single_dirichlet_node  = 1
!      single_dirichlet_value = 1.0
!
!      to get a unique solution (set u=1.0 at the node 1, in the above), and
!      then minimize the 2-norm by subtracting the mean, which can be requested
!      by adding the following to input.nml,
!
!      soln_normalization = "min_2norm"
! 
!*******************************************************************************

 subroutine solution_minimum_2norm

  use data_module    , only : p2, zero
  use data_module    , only : u, nnodes
  use poisson_eqns   , only : iu
  use input_parameter, only : soln_normalization

  implicit none

  integer  :: i
  real(p2) :: u_mean, u_mean0

  write(*,*) '----------------------------------------------------------------------------'
  write(*,*) ' Solution normalization requested: soln_normalization = ', trim(soln_normalization)
  write(*,*)
  write(*,*) '  nnodes = ', nnodes 

 !---------------------------
 ! Compute the mean:

   u_mean = zero

  do i = 1, nnodes

   u_mean = u_mean + u(i,iu)

  end do

   u_mean  = u_mean / real(nnodes,p2)
   u_mean0 = u_mean

  write(*,*)
  write(*,*) ' Mean before subtraction  = ', u_mean
 !---------------------------

 !---------------------------
 !---------------------------
 ! Subtract the mean.

  do i = 1, nnodes

    u(i,iu) = u(i,iu) - u_mean !Just subtract the mean!

  end do
 !---------------------------
 !---------------------------

 !---------------------------
 ! Compute the mean again:

   u_mean = zero

  do i = 1, nnodes

   u_mean = u_mean + u(i,iu)

  end do

   u_mean = u_mean / real(nnodes,p2)

  write(*,*)
  write(*,*) ' Mean after  subtraction  = ', u_mean
  write(*,*) ' Mean divided by old mean = ', u_mean/u_mean0
  write(*,*) '----------------------------------------------------------------------------'

 !---------------------------

 end subroutine solution_minimum_2norm
!********************************************************************************





!*******************************************************************************
! This subroutine writes a solution file.
!*******************************************************************************
 subroutine write_soln_file

  use data_module      , only : u, nnodes
  use poisson_eqns     , only : iu, iux, iuy, iuz
  use gradients        , only : compute_gradient

  use data_module      , only : filename_soln_u , filename_soln_ux, &
                                filename_soln_uy, filename_soln_uz

  implicit none

  integer :: i, os

 call compute_gradient

 !----------------------------------------------------------------------------
  write(*,*)
  write(*,*) ' Solution data file = ', trim(filename_soln_u)
  write(*,*)

  open(unit=8, file=filename_soln_u, status="unknown", iostat=os)

  do i = 1, nnodes

   write(8,*) u(i,iu)

  end do

  close(8)

 !----------------------------------------------------------------------------
  write(*,*)
  write(*,*) ' X-grad data file = ', trim(filename_soln_ux)
  write(*,*)

  open(unit=8, file=filename_soln_ux, status="unknown", iostat=os)

  do i = 1, nnodes

   write(8,*) u(i,iux)

  end do

  close(8)

 !----------------------------------------------------------------------------
  write(*,*)
  write(*,*) ' Y-grad data file = ', trim(filename_soln_ux)
  write(*,*)

  open(unit=8, file=filename_soln_uy, status="unknown", iostat=os)

  do i = 1, nnodes

   write(8,*) u(i,iuy)

  end do

  close(8)

 !----------------------------------------------------------------------------
  write(*,*)
  write(*,*) ' Z-grad data file = ', trim(filename_soln_ux)
  write(*,*)

  open(unit=8, file=filename_soln_uz, status="unknown", iostat=os)

  do i = 1, nnodes

   write(8,*) u(i,iuz)

  end do

  close(8)

 end subroutine write_soln_file
!********************************************************************************






!*******************************************************************************
!
! Compute error norms
!
!*******************************************************************************
 subroutine comute_error_norms(error_norm,heff)

  use data_module    , only : x, y, z     , & !nodal coords
                              u, gradu    , & !soluiton and gradient
                              nnodes, vol

  use poisson_eqns     , only : exact_sol
  use input_parameter, only : scale_factor

  use input_parameter    , only : diffusion_type, Lr_factor
  use poisson_eqns         , only : diff_parameters

  implicit none

  integer , parameter :: p2 = selected_real_kind(P=15)

  real(p2), dimension(7,3), intent(out) :: error_norm
  real(p2),                 intent(out) :: heff

  integer                  :: i, k, nnodes_actual
  real(p2), dimension(4)   :: uexact
  integer                  :: L1=1, L2=2, Li=3
  integer                  :: ix=1, iy=2, iz=3
  real(p2), dimension(7)   :: error
  real(p2)                 :: nu, Lr, Tr

  error_norm(:,L1) =  0.0_p2
  error_norm(:,L2) =  0.0_p2
  error_norm(:,Li) = -1.0_p2
              heff = 0.0_p2
     nnodes_actual = 0

  node_loop : do i = 1, nnodes

    nnodes_actual = nnodes_actual + 1
             heff = heff + vol(i)**(1.0_p2/3.0_p2)

    call exact_sol( x(i), y(i), z(i),          &
                    uexact(1), uexact(2), uexact(3), uexact(4) )

    call diff_parameters( diffusion_type,Lr_factor,u(i,1), nu,Lr,Tr )

        error(1)   = abs( u(i,1)        - uexact(1)   )
        error(2:4) = abs( u(i,2:4)/nu   - uexact(2:4) )

        error(5)   = abs( gradu(i,1,ix) - uexact(2)   )
        error(6)   = abs( gradu(i,1,iy) - uexact(3)   )
        error(7)   = abs( gradu(i,1,iz) - uexact(4)   )

    do k = 2, 7
     error(k) =  error(k) * scale_factor
    end do

    do k = 1, 7
      error_norm(k,L1) = error_norm(k,L1) + error(k)
      error_norm(k,L2) = error_norm(k,L2) + error(k)**2
     if ( error_norm(k,Li) < error(k) ) then
      error_norm(k,Li) = error(k)
     endif
    end do

  end do node_loop

  error_norm(:,L1) =       error_norm(:,L1) / real(nnodes_actual,p2)
  error_norm(:,L2) = sqrt( error_norm(:,L2) / real(nnodes_actual,p2) )

              heff = heff / real(nnodes_actual,p2)

  write(*,*)
  write(*,*) " Errors:       L1             L2             Li"

  write(*,'(a10,3es13.3)') "      u : ", error_norm(1,L1), error_norm(1,L2), error_norm(1,Li)
  write(*,'(a10,3es13.3)') "      ux: ", error_norm(2,L1), error_norm(2,L2), error_norm(2,Li)
  write(*,'(a10,3es13.3)') "      uy: ", error_norm(3,L1), error_norm(3,L2), error_norm(3,Li)
  write(*,'(a10,3es13.3)') "      uz: ", error_norm(4,L1), error_norm(4,L2), error_norm(4,Li)

  write(*,'(a10,3es13.3)') " LSQ(ux): ", error_norm(5,L1), error_norm(5,L2), error_norm(5,Li)
  write(*,'(a10,3es13.3)') " LSQ(uy): ", error_norm(6,L1), error_norm(6,L2), error_norm(6,Li)
  write(*,'(a10,3es13.3)') " LSQ(uz): ", error_norm(7,L1), error_norm(7,L2), error_norm(7,Li)

  write(*,*) "          heff = ", heff
  write(*,*) "        nnodes = ", nnodes
  write(*,*) " nnodes_actual = ", nnodes_actual
  write(*,*) 

!---------------------------------------------------------------------------

 close(8)

 end subroutine comute_error_norms
!********************************************************************************

!*******************************************************************************
!
! Compute truncation error norms
!
!*******************************************************************************
 subroutine comute_te_norms(error_norm,heff)

  use data_module    , only : nnodes, vol, res

  use poisson_eqns   , only : exact_sol
  use input_parameter, only : scale_factor

  use input_parameter, only : diffusion_type, Lr_factor
  use poisson_eqns   , only : diff_parameters

  implicit none

  integer , parameter :: p2 = selected_real_kind(P=15)

  real(p2), dimension(7,3), intent(out) :: error_norm
  real(p2),                 intent(out) :: heff

  integer                  :: i, k
  integer                  :: L1=1, L2=2, Li=3
  real(p2), dimension(7)   :: error

  error_norm(:,L1) =  0.0_p2
  error_norm(:,L2) =  0.0_p2
  error_norm(:,Li) = -1.0_p2
              heff = 0.0_p2
             error = 0.0_p2

   do i = 1, nnodes

    heff = heff + vol(i)**(1.0_p2/3.0_p2)

        error(1:4)   = abs( res(i,1:4) / vol(i) )

    do k = 1, 7
      error_norm(k,L1) = error_norm(k,L1) + error(k)
      error_norm(k,L2) = error_norm(k,L2) + error(k)**2
     if ( error_norm(k,Li) < error(k) ) then
      error_norm(k,Li) = error(k)
     endif
    end do

   end do

  error_norm(:,L1) =       error_norm(:,L1) / real(nnodes,p2)
  error_norm(:,L2) = sqrt( error_norm(:,L2) / real(nnodes,p2) )

              heff = heff / real(nnodes,p2)

  write(*,*)
  write(*,*) "       TE:       L1             L2             Li"

  write(*,'(a10,3es13.3)') "      u : ", error_norm(1,L1), error_norm(1,L2), error_norm(1,Li)
  write(*,'(a10,3es13.3)') "      ux: ", error_norm(2,L1), error_norm(2,L2), error_norm(2,Li)
  write(*,'(a10,3es13.3)') "      uy: ", error_norm(3,L1), error_norm(3,L2), error_norm(3,Li)
  write(*,'(a10,3es13.3)') "      uz: ", error_norm(4,L1), error_norm(4,L2), error_norm(4,Li)

  write(*,*) " heff = ", heff
  write(*,*)

!---------------------------------------------------------------------------

 close(8)

 end subroutine comute_te_norms
!*******************************************************************************
 

!********************************************************************************
!
! Find out big_endian_io.
!
!********************************************************************************
 function big_endian_io( opt_unit )

 integer, intent(in) :: opt_unit
 logical             :: big_endian_io

! one-byte integer(kind)
 integer, parameter  :: i1 = selected_int_kind(2)

! two-byte integer(kind)
 integer, parameter  :: i2 = selected_int_kind(4)
 integer(i1)         :: byte_one, byte_two

! 00000000 00000001 big-endian binary
 integer(i2)         :: two_byte_int = 1_i2

    open(opt_unit,status='scratch',form='unformatted')
      write( opt_unit) two_byte_int
      rewind(opt_unit)
      read(  opt_unit) byte_one, byte_two
    close(opt_unit)
    big_endian_io = ( byte_one == 0 .and. byte_two == 1 )

 end function big_endian_io
!********************************************************************************


!********************************************************************************
!
! Discretization error stest
!
!********************************************************************************
 subroutine accuracy_test_main

  use input_parameter, only : generate_tec_file_b, &
                              generate_tec_file_v, accuracy_test_ngrids, project

  use nceb_data      , only : construct_nceb_data
  use steady_solver  , only : solve_steady
  use poisson_eqns     , only : compute_ref_length

  use data_module    , only : nnodes

  integer , parameter         :: p2 = selected_real_kind(P=15)
  integer                     :: igrid, kvar
  character(80)               :: project_saved, igrid_char, filename_error
  real(p2), dimension(7,3,15) :: enorm
  real(p2), dimension(    15) :: heff
  real(p2), dimension(7,3,15) :: ooa
  real(p2)                    :: temp
  real(p2), dimension(    15) :: wtime
  integer , dimension(    15) :: itr
  integer , dimension(    15) :: nnodes_g

  integer                     :: L1 =1, L2 = 2, Li = 3, k

   project_saved = project

   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) " Accuracy test: accuracy_test_ngrids = ", accuracy_test_ngrids
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*)

  do igrid = 1, accuracy_test_ngrids

   write(*,*) "-------------------------------------------------------------------"
   write(*,*) " Grid = ", igrid
   write(*,*)

   write(igrid_char,'(i3)') igrid
   project = trim(project_saved) // '_' // trim(adjustl(igrid_char))
   write(*,*) "        igrid = ", trim(igrid_char)
   write(*,*) " project name = ", trim(project)

   call set_filenames
   call read_grid
   call construct_nceb_data
   call allocate_nceb_arrays
   call set_initial_solution
   call compute_ref_length
   call solve_steady( solver_max_itr(igrid),  itr(igrid), wtime(igrid) )
   call comute_error_norms(enorm(:,:,igrid), heff(igrid))

   nnodes_g(igrid) = nnodes

   if (igrid > 1) then

     temp = log10(heff(igrid-1))-log10(heff(igrid))

     do kvar = 1, 7
      ooa(kvar,L1,igrid) = (log10(enorm(kvar,L1,igrid-1))-log10(enorm(kvar,L1,igrid)))/temp
      ooa(kvar,L2,igrid) = (log10(enorm(kvar,L2,igrid-1))-log10(enorm(kvar,L2,igrid)))/temp
      ooa(kvar,Li,igrid) = (log10(enorm(kvar,Li,igrid-1))-log10(enorm(kvar,Li,igrid)))/temp
     end do

     write(*,'(a18,4es12.3)') "Order for     u  : ", ooa(1,L1,igrid), ooa(1,L2,igrid), ooa(1,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     ux : ", ooa(2,L1,igrid), ooa(2,L2,igrid), ooa(2,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uy : ", ooa(3,L1,igrid), ooa(3,L2,igrid), ooa(3,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uz : ", ooa(4,L1,igrid), ooa(4,L2,igrid), ooa(4,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for LSQ(ux): ", ooa(5,L1,igrid), ooa(5,L2,igrid), ooa(5,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for LSQ(uy): ", ooa(6,L1,igrid), ooa(6,L2,igrid), ooa(6,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for LSQ(uz): ", ooa(7,L1,igrid), ooa(7,L2,igrid), ooa(7,Li,igrid)

   endif

   if (generate_tec_file_v) call write_tecplot_volume_file
   if (generate_tec_file_b) call write_tecplot_boundary_file

   call deallocate_nceb_arrays
   call deallocate_grid_data

   write(*,*) "-------------------------------------------------------------------"

  end do


 !-------------------------------------------------------------------------------
 !-------------------------------------------------------------------------------

   write(*,*)
   write(*,*)
   write(*,*) "------------------------------------------------------------------"
   write(*,*) "--------------- Summary ---------------------"
   write(*,*) "------------------------------------------------------------------"
   write(*,*)

   filename_error = trim(project_saved) // '_error.data'
   open(unit=10, file=filename_error, status="unknown")

  do igrid = 1, accuracy_test_ngrids


   write(10,*) nnodes_g(igrid), heff(igrid), itr(igrid), &
               (enorm(k,1,igrid), k = 1,7)             , & !<- L1 error
               wtime(igrid)

   write(*,*) "------------------------------------------------------------------"
   write(*,*) "  Grid = ", igrid
   write(*,*)
   write(*,*) "    Iteration = ",  itr(igrid)
   write(*,*) "         heff = ", heff(igrid)
   write(*,*)

   write(*,'(a50)') "                       L1,      L2,       Linf"
   write(*,'(a18,3es12.3)') "ERRORS for     u  : ", enorm(1,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for     ux : ", enorm(2,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for     uy : ", enorm(3,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for     uz : ", enorm(4,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for LSQ(ux): ", enorm(5,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for LSQ(uy): ", enorm(6,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for LSQ(uz): ", enorm(7,:,igrid)
   write(*,*)

   if ( igrid > 1 ) then
     write(*,'(a18,4es12.3)') "Order for     u  : ", ooa(1,L1,igrid), ooa(1,L2,igrid), ooa(1,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     ux : ", ooa(2,L1,igrid), ooa(2,L2,igrid), ooa(2,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uy : ", ooa(3,L1,igrid), ooa(3,L2,igrid), ooa(3,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uz : ", ooa(4,L1,igrid), ooa(4,L2,igrid), ooa(4,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for LSQ(ux): ", ooa(5,L1,igrid), ooa(5,L2,igrid), ooa(5,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for LSQ(uy): ", ooa(6,L1,igrid), ooa(6,L2,igrid), ooa(6,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for LSQ(uz): ", ooa(7,L1,igrid), ooa(7,L2,igrid), ooa(7,Li,igrid)
   endif

   write(*,*)

  end do

   close(10)

 !-------------------------------------------------------------------------------
 !-------------------------------------------------------------------------------

   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) " End of Accuracy test"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*)

  stop

 end subroutine accuracy_test_main
!********************************************************************************

!********************************************************************************
!
! Truncation error test
!
!********************************************************************************
 subroutine accuracy_test_main_te

  use input_parameter, only : generate_tec_file_b, &
                              generate_tec_file_v, accuracy_test_ngrids, project

  use nceb_data      , only : construct_nceb_data
  use poisson_eqns     , only : compute_ref_length

  use data_module    , only : nnodes, u, x, y, z

  use residual       , only : compute_residual

  use poisson_eqns     , only : exact_sol

  use input_parameter, only : diffusion_type, Lr_factor
  use poisson_eqns     , only : diff_parameters

  integer , parameter         :: p2 = selected_real_kind(P=15)
  integer                     :: igrid, kvar
  character(80)               :: project_saved, igrid_char, filename_error
  real(p2), dimension(7,3,15) :: enorm
  real(p2), dimension(    15) :: heff
  real(p2), dimension(7,3,15) :: ooa
  real(p2)                    :: temp
  integer , dimension(    15) :: itr
  integer , dimension(    15) :: nnodes_g

  integer                     :: L1 =1, L2 = 2, Li = 3, k, i

  real(p2), dimension(4)   :: uexact
  real(p2) :: nu, Lr, Tr

   project_saved = project

   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) " Accuracy test: accuracy_test_ngrids = ", accuracy_test_ngrids
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*)

  do igrid = 1, accuracy_test_ngrids

   write(*,*) "-------------------------------------------------------------------"
   write(*,*) " Grid = ", igrid
   write(*,*)

   write(igrid_char,'(i3)') igrid
   project = trim(project_saved) // '_' // trim(adjustl(igrid_char))
   write(*,*) "        igrid = ", trim(igrid_char)
   write(*,*) " project name = ", trim(project)

   call set_filenames
   call read_grid
   call construct_nceb_data
   call allocate_nceb_arrays

   do i = 1, nnodes
    call exact_sol( x(i), y(i), z(i),          &
                    uexact(1), uexact(2), uexact(3), uexact(4) )

    call diff_parameters( diffusion_type,Lr_factor,uexact(1), nu,Lr,Tr )
    u(i,1  ) =    uexact(1)
    u(i,2:4) = nu*uexact(2:4)
   end do

   call compute_ref_length
   call compute_residual
   call comute_te_norms(enorm(:,:,igrid), heff(igrid))

   nnodes_g(igrid) = nnodes

   if (igrid > 1) then

     temp = log10(heff(igrid-1))-log10(heff(igrid))

     do kvar = 1, 7
      ooa(kvar,L1,igrid) = (log10(enorm(kvar,L1,igrid-1))-log10(enorm(kvar,L1,igrid)))/temp
      ooa(kvar,L2,igrid) = (log10(enorm(kvar,L2,igrid-1))-log10(enorm(kvar,L2,igrid)))/temp
      ooa(kvar,Li,igrid) = (log10(enorm(kvar,Li,igrid-1))-log10(enorm(kvar,Li,igrid)))/temp
     end do

     write(*,'(a18,4es12.3)') "Order for     u  : ", ooa(1,L1,igrid), ooa(1,L2,igrid), ooa(1,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     ux : ", ooa(2,L1,igrid), ooa(2,L2,igrid), ooa(2,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uy : ", ooa(3,L1,igrid), ooa(3,L2,igrid), ooa(3,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uz : ", ooa(4,L1,igrid), ooa(4,L2,igrid), ooa(4,Li,igrid)

   endif

   if (generate_tec_file_v) call write_tecplot_volume_file
   if (generate_tec_file_b) call write_tecplot_boundary_file

   call deallocate_nceb_arrays
   call deallocate_grid_data

   write(*,*) "-------------------------------------------------------------------"

  end do


 !-------------------------------------------------------------------------------
 !-------------------------------------------------------------------------------

   write(*,*)
   write(*,*)
   write(*,*) "------------------------------------------------------------------"
   write(*,*) "--------------- Summary ---------------------"
   write(*,*) "------------------------------------------------------------------"
   write(*,*)

   filename_error = trim(project_saved) // '_error.data'
   open(unit=10, file=filename_error, status="unknown")

  do igrid = 1, accuracy_test_ngrids


   write(10,*) nnodes_g(igrid), heff(igrid), itr(igrid), &
               (enorm(k,1,igrid), k = 1,7) !<- L1 error

   write(*,*) "------------------------------------------------------------------"
   write(*,*) "  Grid = ", igrid
   write(*,*)
   write(*,*) "    Iteration = ",  itr(igrid)
   write(*,*) "         heff = ", heff(igrid)
   write(*,*)

   write(*,'(a50)') "                       L1,      L2,       Linf"
   write(*,'(a18,3es12.3)') "ERRORS for     u  : ", enorm(1,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for     ux : ", enorm(2,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for     uy : ", enorm(3,:,igrid)
   write(*,'(a18,3es12.3)') "ERRORS for     uz : ", enorm(4,:,igrid)
   write(*,*)

   if ( igrid > 1 ) then
     write(*,'(a18,4es12.3)') "Order for     u  : ", ooa(1,L1,igrid), ooa(1,L2,igrid), ooa(1,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     ux : ", ooa(2,L1,igrid), ooa(2,L2,igrid), ooa(2,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uy : ", ooa(3,L1,igrid), ooa(3,L2,igrid), ooa(3,Li,igrid)
     write(*,'(a18,4es12.3)') "Order for     uz : ", ooa(4,L1,igrid), ooa(4,L2,igrid), ooa(4,Li,igrid)
   endif

   write(*,*)

  end do

   close(10)

 !-------------------------------------------------------------------------------
 !-------------------------------------------------------------------------------

   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) " End of Accuracy test"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*) "-------------------------------------------------------------------"
   write(*,*)

  stop

 end subroutine accuracy_test_main_te
!********************************************************************************

 end program edu3d_poisson

