!********************************************************************************
!* Educationally-Designed Unstructured 3D (EDU3D) Code
!*
!*  --- EDU3D Poisson
!*
!* This module containes data and functions relevant to the Poisson equations.
!*
!*
!*        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!*
!* the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!*
!* 
!* This F90 program is written and made available for an educational purpose.
!*
!* This file may be updated in future.
!*
!* Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module poisson_eqns

  implicit none

  integer , parameter :: p2 = selected_real_kind(P=15)

  public :: iu, iux, iuy, iuz

  public :: diff_parameters
  public :: diff_parameters_ddt

  public :: compute_ref_length

 !--------------------------------------------------------------------
 ! Indices of the variables.

  integer             :: iu  = 1
  integer             :: iux = 2
  integer             :: iuy = 3
  integer             :: iuz = 4

 contains

!********************************************************************************
!
! Parameter settng: Define the coefficient nu of div(nu*grad(u))=f, here.
!
!********************************************************************************
 subroutine diff_parameters_ddt(diffusion_type, Lr_factor, u, nu, Lr, Tr )

 use derivative_data_df4

 use data_module    , only : ref_length

 integer , parameter ::  p2 = selected_real_kind(P=15)
 real(p2), parameter :: one = 1.00_p2
 real(p2), parameter :: two = 2.00_p2
 real(p2), parameter ::  pi = 3.141592653589793238_p2

 integer                       , intent( in) :: diffusion_type
 real(p2)                      , intent( in) :: Lr_factor
 type(derivative_data_type_df4), intent( in) :: u

 type(derivative_data_type_df4), intent(out) :: nu, Lr, Tr

 !-------------------------------------------------------
 ! Diffusion coefficient nu of div(nu*grad(u))=f(x,y,z)
 !-------------------------------------------------------

 !--------------------------------------
 ! Linear Poisson (constant nu=1).

   if     (diffusion_type==1 ) then

    nu = one

 !--------------------------------------
 ! Your own; I set nu=1 here as an example.

   elseif (diffusion_type==2 ) then

    nu = one

 !--------------------------------------
 ! Nonlinear Poisson (nu = 1 + u^2).

   elseif (diffusion_type==10) then

    nu = one + u**2

 !--------------------------------------
 ! You can implement your own here, and
 ! activate it by adding diffusion_type=2
 ! to the input file, input.nml.

   elseif (diffusion_type==2) then

   ! nu = your nu here.

 !--------------------------------------
 ! Invalid choice

   else

    write(*,*) " Invalid input: diffusion_type = ", diffusion_type
    write(*,*) "  Currently it takes only 1 or 10..."
    stop

   endif
 !--------------------------------------

 !-------------------------------------------------------
 ! End of Diffusion coefficient nu of div(nu*grad(u))=f(x,y,z)
 !-------------------------------------------------------

 !
 ! No need to change parameters below for different nu.
 ! Keep the followings as they are.
 !

 !------------------------------------------------
 ! Hyperboloc-Poisson: Length scale
 !------------------------------------------------

    Lr = ref_length / (two*pi)

 !------------------------------------------------ 
 ! Hyperboloc-Poisson: Adjust it as you like.
 !------------------------------------------------ 

   Lr = Lr * Lr_factor

 !------------------------------------------------
 ! Hyperboloc-Poisson: Relaxation time: Tr = Lr^2/nu.
 !------------------------------------------------ 

    Tr = Lr**2 / nu

 !------------------------------------------------ 
 !------------------------------------------------ 

 end subroutine diff_parameters_ddt
!********************************************************************************

!********************************************************************************
!
! Source term (forcing term), which is a function of (x,y,z).
! This should be f = uxx+uyy+uzz with the exact u.
!
! You need to implement your own f(x,y,z). Here, f is defined such that the
! exact solution is given by c4 * sin( pi*(c1*x + c2*y + c3*z) ) as
! defined by the subroutine exact_sol(). Right, f is defined as f=div(nu*(grad(uexact))).
!
! So, if you change exact_sol(), you'll need to modify f accordingly.
! Or if you just want to solve div(nu*grad(u))=f without any exact solution,
! then, you can set f to be any function, and ignore exact_sol().
! The code will compute errors assuming exact solution is given by exact_sol(),
! but just ignore them.
!
!********************************************************************************
 function forcing_term( xp, yp, zp )

 use data_module    , only : p2, pi, one, zero
 use input_parameter, only : exact_c1, exact_c2, exact_c3, exact_c4, diffusion_type

 real(p2), intent(in) :: xp, yp, zp
 real(p2)             :: forcing_term
 real(p2)             :: x, y, z
 real(p2)             :: c1, c2, c3, c4, nu, nux, nuy, nuz

 real(p2)             :: ue, uxe, uye, uze, uxxe, uyye, uzze

    forcing_term = zero ! Initialization

    c1 = exact_c1
    c2 = exact_c2
    c3 = exact_c3
    c4 = exact_c4

     x = xp
     y = yp
     z = zp
 
  !------------------------------------------------------------------------------------------
  ! Linear Poisson: div(nu*grad(u))=f -> f=nu*div(grad(u))

   if     (diffusion_type==1  ) then

    forcing_term = - c4 * pi**2*( c1**2 + c2**2 + c3**2 )*sin( pi*(c1*x + c2*y + c3*z) )

  !------------------------------------------------------------------------------------------
  ! Nonlinear Poisson: div(nu*grad(u))=f -> f=grad(nu)*grad(u) + nu*div(grad(u))

   elseif (diffusion_type==10 ) then

    ue   = c4 *       sin( pi*(c1*x + c2*y + c3*z) )
    uxe  = c4 * c1*pi*cos( pi*(c1*x + c2*y + c3*z) )
    uye  = c4 * c2*pi*cos( pi*(c1*x + c2*y + c3*z) )
    uze  = c4 * c3*pi*cos( pi*(c1*x + c2*y + c3*z) )
 
    uxxe =-c4 * c1*c1*pi*pi*sin( pi*(c1*x + c2*y + c3*z) )
    uyye =-c4 * c2*c2*pi*pi*sin( pi*(c1*x + c2*y + c3*z) )
    uzze =-c4 * c3*c3*pi*pi*sin( pi*(c1*x + c2*y + c3*z) )

    nu  = one + ue**2
    nux = 2.0_p2*ue*uxe
    nuy = 2.0_p2*ue*uye
    nuz = 2.0_p2*ue*uze

    forcing_term = ( nux*uxe + nuy*uye + nuz*uze ) + nu*( uxxe + uyye + uzze )

  !------------------------------------------------------------------------------------------
  ! Your own f(x,y,z): Here, I set f = 0 as an example.

   elseif (diffusion_type==2 ) then

    forcing_term = zero ! your own function

   endif
  !------------------------------------------------------------------------------------------

 end function forcing_term
!********************************************************************************

!********************************************************************************
!
! Exat solution: u = c4 * sin( pi*(c1*x + c2*y + c3*z) ).
!
! In practical problems, there is no exact solutions, usually. This exact solution
! is used to verify the code. This is an exact solution of div(nu*grad(u))=f
! where f is defined in the subroutine above forcing_term().
!
! Note: the same exact solution is used for both linear and nonlinear Poisson
!       defined in the subroutine diff_parameters_ddt(diffusion_type, Lr_factor, u, nu, Lr, Tr )
!       at the top of this file.
!
! If you change the exact solution, then you'll need to modify forcing_term(),
! such that f=div(nu*(grad(uexact))).
!
! For practical problems, where there is no exact solutions, you just ignode
! this subroutine and any computations and resutls (e.g., errors).
!
!********************************************************************************
 subroutine exact_sol( xp, yp, zp, ue, uxe, uye, uze )

 use data_module    , only : p2, pi
 use input_parameter, only : exact_c1, exact_c2, exact_c3, exact_c4

 real(p2), intent( in) :: xp, yp, zp
 real(p2), intent(out) :: ue, uxe, uye, uze
 real(p2)              :: x, y, z
 real(p2)              :: c1, c2, c3, c4

    c1 = exact_c1
    c2 = exact_c2
    c3 = exact_c3
    c4 = exact_c4

     x = xp
     y = yp
     z = zp

   ue  = c4 * sin( pi*(c1*x + c2*y + c3*z) )
   uxe = c4 * c1*pi*cos( pi*(c1*x + c2*y + c3*z) )
   uye = c4 * c2*pi*cos( pi*(c1*x + c2*y + c3*z) )
   uze = c4 * c3*pi*cos( pi*(c1*x + c2*y + c3*z) )

 end subroutine exact_sol
!********************************************************************************


!********************************************************************************
! Relaxation parameters
!********************************************************************************
 subroutine diff_parameters(diffusion_type, Lr_factor, u, nu, Lr, Tr )

 use derivative_data_df4

 integer , parameter ::  p2 = selected_real_kind(P=15)

 integer , intent( in) :: diffusion_type
 real(p2), intent( in) :: Lr_factor, u
 real(p2), intent(out) :: nu, Lr, Tr

 type(derivative_data_type_df4) :: nu_ddt, Lr_ddt, Tr_ddt, u_ddt

  u_ddt = u
  call diff_parameters_ddt(diffusion_type, Lr_factor, u_ddt, nu_ddt, Lr_ddt, Tr_ddt )

  nu = nu_ddt%f
  Lr = Lr_ddt%f
  Tr = Tr_ddt%f

 end subroutine diff_parameters
!********************************************************************************


!*******************************************************************************
!
! Compute the reference length: Lopt as in Equation (52) of JCP2019 paper:
!
! H. Nishikawa, “A hyperbolic Poisson solver for tetrahedral grids”, Journal of
! Computational Physics, Volume 409, 15 May 2020, 109358, doi:10.1016/j.jcp.2020.109358
!
!*******************************************************************************
 subroutine compute_ref_length

  use data_module    , only : x, y, z    , & !nodal coords
                              ntria, tria, & !number of tria, and tria info
                              nquad, quad, & !number of quad, and quad info
                              nnodes         !number of nodes

  use data_module    , only : vol, ref_length, one
  use input_parameter, only : ref_length_in, ref_length_factor

  implicit none
  integer , parameter ::    p2 = selected_real_kind(15) ! Double precision
  real(p2), parameter ::  zero = 0.0_p2
  real(p2), parameter ::   two = 2.0_p2
  real(p2), parameter :: four  = 4.0_p2

  integer  :: i
  real(p2) :: volume, area
  real(p2) :: x1,x2,x3,x4, y1,y2,y3,y4, z1,z2,z3,z4

  real(p2), dimension(3) :: d_max


  real(p2)               :: inside_sqrt
  real(p2)               :: diag_estimate

  inside_sqrt = -1.0_p2

 !--------------------------------------------------------------
 ! Compute the volume of the domain

    volume = zero
   do i = 1, nnodes
    volume = volume + vol(i)
   end do

 !--------------------------------------------------------------
 ! Compute the total surface area of the domain

      area = zero

  !(1) Add contributions from triangular boundary faces.

   do i = 1, ntria

    x1 = x( tria(i,1) )
    x2 = x( tria(i,2) )
    x3 = x( tria(i,3) )

    y1 = y( tria(i,1) )
    y2 = y( tria(i,2) )
    y3 = y( tria(i,3) )

    z1 = z( tria(i,1) )
    z2 = z( tria(i,2) )
    z3 = z( tria(i,3) )

    area = area + triangle_area(x1,x2,x3, y1,y2,y3, z1,z2,z3)

   end do

  !(2) Add contributions from quadrialteral boundary faces.

   do i = 1, nquad

    x1 = x( quad(i,1) )
    x2 = x( quad(i,2) )
    x3 = x( quad(i,3) )
    x4 = x( quad(i,4) )

    y1 = y( quad(i,1) )
    y2 = y( quad(i,2) )
    y3 = y( quad(i,3) )
    y4 = y( quad(i,4) )

    z1 = z( quad(i,1) )
    z2 = z( quad(i,2) )
    z3 = z( quad(i,3) )
    z4 = z( quad(i,4) )
 
   !Quad area = area of tria 1-2-3 + area of tria 1-3-4

    area = area + triangle_area(x1,x2,x3, y1,y2,y3, z1,z2,z3) !area of tria 1-2-3
    area = area + triangle_area(x1,x3,x4, y1,y3,y4, z1,z3,z4) !area of tria 1-3-4

   end do

 !--------------------------------------------------------------
 ! Maximum distance in each coordinate direction.

        d_max(1) = maxval(x) - minval(x)
        d_max(2) = maxval(y) - minval(y)
        d_max(3) = maxval(z) - minval(z)

   diag_estimate = maxval( d_max ) !Basically use the max of dx,dy,dz.

    write(*,*)
    write(*,*) "  >>>>>                 V = ", volume
    write(*,*) "  >>>>>                 S = ", area
    write(*,*) "  >>>>>               S/V = ", area/volume
    write(*,*) "  >>>>>            Lx_max = ", d_max(1)
    write(*,*) "  >>>>>            Ly_max = ", d_max(2)
    write(*,*) "  >>>>>            Lz_max = ", d_max(3)
    write(*,*) "  >>>>>     diag_estimate = ", diag_estimate

   write(*,*)
   write(*,*) "-------------------------------------------------------"
   write(*,*) " Reference length scale for dimensional consistency"
   write(*,*) " for the hyperbolic solver."
   write(*,*)

 !----------------------------------------------------------------------------
 ! Compute the reference length by Lopt for rectangular domains: Eq.(51) in
 !  
 ! H. Nishikawa, “A hyperbolic Poisson solver for tetrahedral grids”, Journal of
 ! Computational Physics, Volume 409, 15 May 2020, 109358, doi:10.1016/j.jcp.2020.109358.
 !
 !----------------------------------------------------------------------------

   if     (ref_length_in < -1.0e-12_p2) then


    inside_sqrt = area**2/four - two*volume * sqrt (diag_estimate**2 + area )

     ref_length = volume/sqrt( inside_sqrt )

     write(*,*)
     write(*,*) " Lopt for a general domain is used for ref_length = ", ref_length
     write(*,*)

 !----------------------------------------------------------------------------
 ! Compute the reference length by Lopt: Eq.(52) in
 !  
 ! H. Nishikawa, “A hyperbolic Poisson solver for tetrahedral grids”, Journal of
 ! Computational Physics, Volume 409, 15 May 2020, 109358, doi:10.1016/j.jcp.2020.109358.
 !
 !----------------------------------------------------------------------------

   elseif (abs(ref_length_in)  < 1.0e-12_p2) then

     ref_length = one/sqrt( one/d_max(1)**2 + one/d_max(2)**2 + one/d_max(3)**2 )

     write(*,*)
     write(*,*) " Lopt for rectangular domain is used for ref_length = ", ref_length
     write(*,*)

 !----------------------------------------------------------------------------
 ! Use the value specified in the input file by "ref_length_in".
 !----------------------------------------------------------------------------

   elseif (ref_length_in > 0.0_p2) then

     ref_length = ref_length_in

    write(*,*)
    write(*,*) " >>> ref_length_in is positive = ", ref_length_in
    write(*,*) " Setting ref_length by the input value: ", ref_length
    write(*,*)

   else


     write(*,*)
     write(*,*) " Error in Lopt: ref_length = ", ref_length, " use < 0 , ==0, or > 0... Stop..."
     write(*,*)
     stop


   endif

 !----------------------------------------------------------------------------
 ! Adjust the ref_lenght as you like.
 !----------------------------------------------------------------------------

   ref_length = ref_length * ref_length_factor

 !----------------------------------------------------------------------------
 !----------------------------------------------------------------------------
 !----------------------------------------------------------------------------
 !-----------------------------------------------------

   write(*,*) "           volume = ", volume
   write(*,*) "             area = ", area
   write(*,*) "           ds_max = ", maxval( d_max )
   write(*,*) "    ref_length_in = ", ref_length_in
   write(*,*)
   write(*,*) "       ref_length = ", ref_length
   write(*,*)
   write(*,*) " This is not used for the conventional solver."
   write(*,*) "-------------------------------------------------------"
   write(*,*)

  !Check the sign of square root only when the general formula is requested.
   if (ref_length_in < -1.0e-12_p2) then
    if (inside_sqrt < 0.0_p2) then
     write(*,*)
     write(*,*) " Negative inside the sqrt in Lopt. Really?"
     write(*,*) "      inside_sqrt = ", inside_sqrt
     write(*,*) " That's why you get ref_length = NaN....."
     write(*,*)
     write(*,*) " Please report this to Hiro Nishikawa (hiro@nianet.org). Stop..."
     write(*,*)
     stop
    endif
   endif

 end subroutine compute_ref_length
!********************************************************************************

!*******************************************************************************
! Compute the area of a triangle in 3D defined by 3 vertices:
!
!       (x1,y1,z1), (x2,y2,z2), (x3,y3,z3),
!
! which is assumed to be ordered clockwise.
!
!     1             2
!      o------------o
!       \         .
!        \       . --------->
!         \    .
!          \ .
!           o
!           3
!
! Note: Area is a vector based on the right-hand rule: 
!       when wrapping the right hand around the triangle with the fingers in the
!       direction of the vertices [1,2,3], the thumb points in the positive
!       direction of the area.
!
! Note: Area vector is computed as the cross product of edge vectors [31] and [32].
!
!*******************************************************************************
 function triangle_area(x1,x2,x3, y1,y2,y3, z1,z2,z3) result(area)
 
 implicit none
 integer , parameter :: dp = selected_real_kind(15) !Double precision

 !Input
  real(dp), intent(in)   :: x1,x2,x3, y1,y2,y3, z1,z2,z3
 !Output
  real(dp)               :: area
  real(dp), dimension(3) :: area_vector

  integer :: ix=1, iy=2, iz=3

  !x-component of the area vector
   area_vector(ix) = 0.5_dp*( (y1-y3)*(z2-z3)-(z1-z3)*(y2-y3) )

  !y-component of the area vector
   area_vector(iy) = 0.5_dp*( (z1-z3)*(x2-x3)-(x1-x3)*(z2-z3) )

  !z-component of the area vector
   area_vector(iz) = 0.5_dp*( (x1-x3)*(y2-y3)-(y1-y3)*(x2-x3) )

   area = sqrt( area_vector(ix)**2 + &
                area_vector(iy)**2 + &
                area_vector(iz)**2   )

 end function triangle_area



 end module poisson_eqns
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
