!********************************************************************************
!* Educationally-Designed Unstructured 3D (EDU3D) Code
!*
!*  --- EDU3D Poisson
!*
!* This module containes a subroutine that comptutes the residuals at nodes for
!* the node-centered edge-based discretization. A special boundary flux quadrature
!* is required to preserve second- and third-order accuracy. Numerical flux
!* is computed at the edge-midpoint: edge-based quadrature.
!*
!*
!*        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!*
!* the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!*
!* 
!* This F90 program is written and made available for an educational purpose.
!*
!* This file may be updated in future.
!*
!* Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module residual

 !This module contains the following subroutine:
 
  public :: compute_residual ! compute the residual

 contains

!********************************************************************************
!
! This subtourine computes the residual at all nodes by the node-centered 
! edge-based discretization.
!
! The residual is computed in a loop over edges with a numerical flux computed
! at the edge midpoint, followed by the boundary closure.
!
! - The method is second-order accurate on arbitrary tetrahedral grids.
! - Second-order on regular prism/hex and mixed grids.
! - Requires a special boundary quadrature to preserve the design accuracy.
!
!********************************************************************************
 subroutine compute_residual

 use data_module      , only : p2, zero, third, fourth
 use data_module      , only : x, y, z, nedges, edge, u, gradu, nnodes
 use data_module      , only : res, wsn, n12, n12_mag, e12, e12_mag
 use data_module      , only : ntria, tria, nquad, quad, bn, bn_mag, bc_type
 use data_module      , only : n_residual_evaluation, vol, ar

 use fluxjac_interface, only : interface_flux, interface_flux_b
 use bc_states        , only : set_bc_state, get_right_state
 use gradients        , only : compute_gradient

 use input_parameter  , only : accuracy_order          , &
                               diffusion_type          , &
                               Lr_factor, flux_choice  , &
                               bc_dirichlet_user_values, &
                               bc_dirichlet            , &
                               bc_neumann_user_values  , &
                               bc_neumann              , &
                               bc_neumann_read_uxuyuz  , &
                               single_dirichlet_node   , &
                               single_dirichlet_value

 use poisson_eqns     , only : forcing_term, exact_sol, diff_parameters

 implicit none

!Local variables
 real(p2), dimension(4)   :: num_flux
 real(p2), dimension(4)   :: num_flux_jb, num_flux_jl, num_flux_jr

 real(p2), dimension(3)   :: da    , ev
 real(p2)                 :: da_mag, ev_mag

 real(p2)                 :: eig

 real(p2), dimension(4,4) :: dummy4x4

 integer :: i, j, ib
 integer :: node1, node2
 integer :: nj, nl, nr

 integer, dimension(3,2)  :: loc_nghbr

 real(p2) :: nu, Lr, Tr, ubb,qnbb
 real(p2), dimension(4) :: ue, uR

 logical  :: single_dirichlet_node_found

! Counting the number of residual evalutions.

   n_residual_evaluation = n_residual_evaluation + 1

!     1             3
!      o------------o
!       \         .
!        \       . 
!         \    .
!          \ .
!           o
!           2
!
! Neighbor nodes of a given node within a triangle (boundary elements).

  loc_nghbr(1,1) = 2 !nghbr1 of vertex 1.
  loc_nghbr(1,2) = 3 !nghbr2 of vertex 1.

  loc_nghbr(2,1) = 3 !nghbr1 of vertex 2.
  loc_nghbr(2,2) = 1 !nghbr2 of vertex 2.

  loc_nghbr(3,1) = 1 !nghbr1 of vertex 3.
  loc_nghbr(3,2) = 2 !nghbr2 of vertex 3.

!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
!-------------------------------------------------------------------------
! Residual computation: interior edges

! Initialization

   res = zero
   wsn = zero

! Compute gradients at all nodes for 2nd-order accuracy.

   if (accuracy_order == 2) call compute_gradient

!--------------------------------------------------------------------------------
! Flux computation across internal edges (to be accumulated in res(:))
!
!   node2              1. Extrapolate the solutions to the edge-midpoint
!       o                 from the nodes, n1 and n2.
!        \             2. Compute the numerical flux
!         \            3. Add it to the residual for n1, and subtract it from
!          \              the residual for n2.
!           \
!            o         Directed area is the sum of all dual faces defined
!          node1       by the edge-midpoint, centroids of adjacent tetrahedra
!                      and the centroids of the adjacent triangular faces.
!                      Directed area is positive in n1 -> n2
!--------------------------------------------------------------------------------
  loop_edges : do i = 1, nedges


 ! Left and right nodes of the i-th edge

     node1 = edge(i,1)  ! Left node of the edge
     node2 = edge(i,2)  ! Right node of the edge
       da = n12(i,:)    ! This is the directed area vector (unit vector)
   da_mag = n12_mag(i)  ! Magnitude of the directed area vector
       ev = e12(i,:)    ! Edge vector (unit vector)
   ev_mag = e12_mag(i)  ! Magnitude of the edge vector

 !  Compute the numerical flux for given w1 and u2.
 !  Reconstruction to the midpoint is performed in 'interface_flux' for 2nd-order.

   call interface_flux( u(node1,:), u(node2,:), gradu(node1,:,:), gradu(node2,:,:), &
                              da, ev, ev_mag, ar(node1), ar(node2), num_flux, eig   )

 !  Add the flux multiplied by the magnitude of the directed area vector to node1,
 !  and accumulate the max wave speed quantity for use in the pseudo-time step calculation.

     res(node1,:) = res(node1,:)  +  num_flux * da_mag
     wsn(node1  ) = wsn(node1  )  +       eig * da_mag

 !  Subtract the flux multiplied by the magnitude of the directed area vector from node2,
 !  and accumulate the max wave speed quantity for use in the pseudo-time step calculation.
 !  NOTE: Subtract because the outward face normal is -da for the node2.

     res(node2,:) = res(node2,:)  -  num_flux * da_mag
     wsn(node2  ) = wsn(node2  )  +       eig * da_mag


  end do loop_edges
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Triangular boundary faces:
!
! Close the residual by the 2nd-order boundary flux quadrature.
!
! Loop over triangular boundary elements and distribute a contribution to each node. 
! Node ordering (orientation) is not important here.
! For each node, j, we call the other nodes, l and r.
!
!     l              r
!      o------------o
!       \         .
!        \       .
!         \    .
!          \ .
!           o
!           j
!
!
!    Flux to be added to the residual at j is
!
!      6/8*Flux(uj,ub) + 1/8*Flux(ul,ub) + 1/8*Flux(ur,ub),
!
!    where ub is a BC state.
!
! - Second-order quadrature formulas for 3D tet/prs/pyr/hex grids can be found in:
!
! H. Nishikawa, "Beyond Interface Gradient: A General Principle for
! Constructing Diffusion Schemes", AIAA Paper 2010-5093, 40th AIAA Fluid
! Dynamics Conference and Exhibit, Chicago, 2010.
! http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_AIAA-2010-5093.pdf
! (NOTE: See Appendix B for boundary flux quadratures for 2D/3D elements.)
!
!--------------------------------------------------------------------------------

  if (ntria > 0) then

  boundary_element_loop : do i = 1, ntria

    ib = tria(i,4) !Boundary part number

    nodes_of_triangle : do j = 1, 3 ! Nodes: nj, nr, nl

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! Second-order accurate boundary flux quadratue for tetra (6/8,1/8,1/8)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

     !----------------------------------------------------------------
     ! Node j: Phi_jb(bn): BC is implemented here.

      nj = tria( i, j )

      call set_bc_state( u(nj,1),u(nj,2),u(nj,3),u(nj,4),ib,bn(i,:),nj,x(nj),y(nj),z(nj), ubb,qnbb) !Get ub,qnb
      call get_right_state(  bc_type(ib),u(nj,:),ubb,qnbb, bn(i,:),  dummy4x4,uR )               !Get uR
      call interface_flux_b( u(nj,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nj), num_flux_jb, eig )

      wsn(nj) = wsn(nj)  +  eig * bn_mag(i)

     !----------------------------------------------------------------
     ! Flux at r: Phi_jr(bn): BC is implemented here.

           nr = tria( i, loc_nghbr(j,1) )

      call set_bc_state( u(nr,1),u(nr,2),u(nr,3),u(nr,4),ib,bn(i,:),nr,x(nr),y(nr),z(nr), ubb,qnbb) !Get ub,qnb
      call get_right_state(  bc_type(ib),u(nr,:),ubb,qnbb, bn(i,:),  dummy4x4,uR )               !Get uR
      call interface_flux_b( u(nr,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nr), num_flux_jr, eig )

      wsn(nj) = wsn(nj)  +  eig * bn_mag(i)

     !----------------------------------------------------------------
     ! Flux at l: Phi_jl(bn): BC is implemented here.

          nl = tria( i, loc_nghbr(j,2) )

      call set_bc_state( u(nl,1),u(nl,2),u(nl,3),u(nl,4),ib,bn(i,:),nl,x(nl),y(nl),z(nl), ubb,qnbb) !Get ub,qnb
      call get_right_state(  bc_type(ib),u(nl,:),ubb,qnbb, bn(i,:),  dummy4x4,uR )               !Get uR
      call interface_flux_b( u(nl,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nl), num_flux_jl, eig )

      wsn(nj) = wsn(nj)  +  eig * bn_mag(i)

     ! Add the boundary contribution to the residual with the 2nd-order formula:

      res(nj,:) = res(nj,:) &
                + ( 6.0_p2*num_flux_jb + (num_flux_jl+num_flux_jr) )*(third*bn_mag(i))/8.0_p2

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! End of two other nodes
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

    end do nodes_of_triangle

  end do boundary_element_loop

  endif

!--------------------------------------------------------------------------------
! Quadrialteral boundary faces:
!
! Close the residual by the 2nd-order boundary flux quadrature.
!
! Loop over quad boundary elements and distribute a contribution to each node. 
!               
!      o------------o
!       \            \
!        \            \ 
!         \            \
!          \            \ 
!           o------------o
!           j
!
!
!    Flux to be added to the residual at j is just
!
!      Flux(uj,ub)
!
!    where ub is a BC state.
!
!
! - This is simple for hexa grids. We only need to compute the flux at the node j and
!   add it to the residual at j. This is second-order accurate for regular hexa grids
!   with general rectangular cells. If a hexahedral cell is irregular with non-planar
!   faces, then second-order will be lost. See the reference below for details.
!
! - Second-order quadrature formulas for 3D tet/prs/pyr/hex grids can be found in:
!
! H. Nishikawa, "Beyond Interface Gradient: A General Principle for
! Constructing Diffusion Schemes", AIAA Paper 2010-5093, 40th AIAA Fluid
! Dynamics Conference and Exhibit, Chicago, 2010.
! http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_AIAA-2010-5093.pdf
! (NOTE: See Appendix B for boundary flux quadratures for 2D/3D elements.)
!
!--------------------------------------------------------------------------------
 
  if (nquad > 0) then

  boundary_element_loop_quad : do i = 1, nquad

    ib = quad(i,5) !Boundary part number

    nodes_of_quad : do j = 1, 4

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! Second-order accurate boundary flux quadratue for hexa (1,0,0,0)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

     ! Node j: Phi_jb(bn): BC is implemented here.

      nj = quad( i, j )

      call set_bc_state( u(nj,1),u(nj,2),u(nj,3),u(nj,4),ib,bn(i,:),nj,x(nj),y(nj),z(nj), ubb,qnbb) !Get ub,qnb
      call get_right_state(  bc_type(ib),u(nj,:),ubb,qnbb, bn(i,:),  dummy4x4,uR )               !Get uR
      call interface_flux_b( u(nj,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nj), num_flux_jb, eig )

        wsn(nj) = wsn(nj)  +  eig * bn_mag(i)

      res(nj,:) = res(nj,:) + num_flux_jb*(fourth*bn_mag(i)) !multiplied by 1/4 of the quad area.

     ! That's it for hexa grids. Simple.

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! End of Second-order accurate boundary flux quadratue for hexa (1,0,0,0)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

    end do nodes_of_quad

  end do boundary_element_loop_quad

  endif


!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Add a forcing term to the diff eq: To solve Poisson eq.

  do i = 1, nnodes
   res(i,1) = res(i,1) + forcing_term( x(i), y(i), z(i) ) * vol(i)
  end do

!--------------------------------------------------------------------------------
!--------------------------------------------------------------------------------
! Add source terms in the hyperbolic diffusion.

  do i = 1, nnodes
   call diff_parameters( diffusion_type,Lr_factor,u(i,1), nu,Lr,Tr )
   res(i,2) = res(i,2) + u(i,2)/nu * vol(i)
   res(i,3) = res(i,3) + u(i,3)/nu * vol(i)
   res(i,4) = res(i,4) + u(i,4)/nu * vol(i)
  end do

!--------------------------------------------------------------------------------


!--------------------------------------------------------------------------------
! Enforce a strong condition if requested.
! This assumes the exact solution. So, this is used only for verification.
!--------------------------------------------------------------------------------
 
  single_dirichlet_node_found = .false.

 !-----------------------------------------------------------------------
 ! Triangular elements

   tria_strong : do i = 1, ntria

    ib = tria(i,4)

   !-----------------------------------------------------------
    if     ( trim(bc_type(ib))=="dirichlet_strong") then

     do j = 1, 3
       nj = tria( i, j )

       call exact_sol( x(nj), y(nj), z(nj), ue(1), ue(2), ue(3), ue(4) )

       if ( bc_dirichlet_user_values )  ue(1) = bc_dirichlet(ib)

       res(nj,1) = u(nj,1) - ue(1)

     end do

   !-----------------------------------------------------------
   ! This is for verification purpose only.
    elseif ( trim(bc_type(ib))=="dirichlet_strong_all") then

     do j = 1, 3
       nj = tria( i, j )

       call diff_parameters( diffusion_type,Lr_factor,u(nj,1),  nu,Lr,Tr )
       call exact_sol( x(nj), y(nj), z(nj), ue(1), ue(2), ue(3), ue(4) )

       res(nj,1) = u(nj,1) -    ue(1)
       res(nj,2) = u(nj,2) - nu*ue(2)
       res(nj,3) = u(nj,3) - nu*ue(3)
       res(nj,4) = u(nj,4) - nu*ue(4)

     end do

   !-----------------------------------------------------------
   ! Fix the solution at a single node.

     elseif (single_dirichlet_node > 0 ) then

     do j = 1, 3
              nj = tria( i, j )

      if (single_dirichlet_node == nj) then
       res(nj,1) = u(nj,1) - single_dirichlet_value
       single_dirichlet_node_found = .true.
       exit tria_strong
      endif

     end do

    endif

   end do tria_strong 

 !-----------------------------------------------------------------------
 ! Quad elements


   quad_strong : do i = 1, nquad

    ib = quad(i,5)

   !-----------------------------------------------------------
    if     ( trim(bc_type(ib))=="dirichlet_strong") then

     do j = 1, 4
       nj = quad( i, j )

       call exact_sol( x(nj), y(nj), z(nj), ue(1), ue(2), ue(3), ue(4) )

       if ( bc_dirichlet_user_values )  ue(1) = bc_dirichlet(ib)

       res(nj,1) = u(nj,1) - ue(1)

     end do

   !-----------------------------------------------------------
   ! This is for verification purpose only.
    elseif ( trim(bc_type(ib))=="dirichlet_strong_all") then

     do j = 1, 4
       nj = quad( i, j )

       call diff_parameters( diffusion_type,Lr_factor,u(nj,1),  nu,Lr,Tr )
       call exact_sol( x(nj), y(nj), z(nj), ue(1), ue(2), ue(3), ue(4) )

       res(nj,1) = u(nj,1) -    ue(1)
       res(nj,2) = u(nj,2) - nu*ue(2)
       res(nj,3) = u(nj,3) - nu*ue(3)
       res(nj,4) = u(nj,4) - nu*ue(4)

     end do

   !-----------------------------------------------------------
   ! Fix the solution at a single node.

     elseif (single_dirichlet_node > 0 ) then

      do j = 1, 4
               nj = quad( i, j )

       if (single_dirichlet_node == nj) then

        res(nj,1) = u(nj,1) - single_dirichlet_value

        single_dirichlet_node_found = .true.
        exit quad_strong

       endif

      end do

    endif

   end do quad_strong


   if (single_dirichlet_node > 0 .and. .not.single_dirichlet_node_found) then
    write(*,*) " The node cannot be found: single_dirichlet_node = ", single_dirichlet_node
    write(*,*) " This is not a boundary node. Correct it, and try again... Stop..."
    stop
   endif

!----------------------------------------------------------------
! Conventional scheme has only 1 equation. I know it is not
! an efficient way to implement it, but it is simpler this way
! to implement both the conventional and hyperbolic solvers...
!----------------------------------------------------------------
  if (trim(flux_choice)=="alpha") then

   do i = 1, nnodes
    res(i,2:4) = zero
   end do

  endif
!----------------------------------------------------------------
!----------------------------------------------------------------
!----------------------------------------------------------------


 end subroutine compute_residual
!********************************************************************************

 end module residual
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
