!********************************************************************************
! Educationally-Designed Unstructured 3D (EDU3D) Code
!
!  --- EDU3D Poisson
!
!
! This module containes subroutines that solve the global system of nonlinear
! residual equations:
!
!     Res(U) = 0.
!
! - This is a nonlinear system solver, which some people call a steady solver.
!
! - So, this can solve unsteady implicit-time-stepping-scheme residuals as well,
!   which incorporate physical time derivatives as source terms, or any nonlinear
!   system as long as a Jacobian is provided.
!
!
!        written by Dr. Katate Masatsuka (info[at]cfdbooks.com),
!
! the author of useful CFD books, "I do like CFD" (http://www.cfdbooks.com).
!
! 
! This F90 program is written and made available for an educational purpose.
!
! This file may be updated in future.
!
! Katate Masatsuka, June 2021. http://www.cfdbooks.com
!********************************************************************************

 module steady_solver

 !This module contains the following subroutines:

  public :: solve_steady                 ! solve residual equations (nonlinear system)

  public :: compute_residual_norm        ! compute residual norms
  public :: compute_local_time_step_dtau ! compute pseudo time step

 !Nonlinear solvers:

  public :: explicit_pseudo_time_forward_euler  !Explicit forward Euler
  public :: explicit_pseudo_time_tvdrk2         !Explicit 2-stage TVD RK
  public :: implicit                            !Implicit defect-correction
  public :: jfnk                                !Jacobian-Free Newton-Krylov

 contains

!*******************************************************************************
!
! This subroutine solves the global system of nonlinear residual eqs: Res(U) = 0
!
! Four nonlinear solvers are implemented:
!
! (1)Forward Euler  explicit time stepping. ! many iterations (or unstable)
! (2)2-stage TVD RK explicit time stepping. ! many iterations
! (3)Implicit defect-correction solver.     ! much less iterations
! (4)Jacobian-Free Newton-Krylov with GCR.  ! can be much much less iterations
!
! Note: In (4), the solver (3) is used as a preconditioner.
!
!*******************************************************************************
 subroutine solve_steady(solver_max_itr,solver_itr_actual,wtime)

 use input_parameter, only : solver_type, solver_tolerance, diffusion_type
 use input_parameter, only : flux_choice, jacobian_choice, CFL, accuracy_order
 use data_module    , only : i_iteration
 use data_module    , only : lrelax_sweeps_actual, lrelax_roc
 use residual       , only : compute_residual
 use data_module    , only : gcr_proj_actual, gcr_rkr0, n_residual_evaluation
 use data_module    , only : filename_tecplot_hist

 use input_parameter, only : bc_dirichlet_user_values
 use input_parameter, only : bc_neumann_user_values

 integer ,      parameter :: p2 = selected_real_kind(15) ! Double precision

 integer , intent( in) :: solver_max_itr
 integer , intent(out) :: solver_itr_actual
 real(p2), intent(out) :: wtime

 real(p2), dimension(4,3) :: res_norm
 real(p2), dimension(4,3) :: res_norm_initial
 integer                  :: L1 = 1, ii, jj
 real(p2) :: time_begin, time_end, walltime
 logical  :: iteration_exit

 !------------------------------------------------------------------------------------
 ! Print some input values on screen for checking.

  write(*,*)
  write(*,*) "---------------------------------------"
  write(*,*) " Begin nonlinear iteration"
  write(*,*)

  write(*,*)
  write(*,'(2a)')   "      solver_type =  ", trim(solver_type)
  write(*,'(a,es9.2)') "             CFL  = ", CFL
  write(*,'(a,i1)') "   accuracy_order =  ", accuracy_order
  write(*,'(2a)')   "      flux_choice =  ", trim(flux_choice)
  write(*,'(2a)')   "  jacobian_choice =  ", trim(jacobian_choice)

 !------------------------------------------------------------------------------------
 ! Equation type

  if     (diffusion_type==1) then
  write(*,'(2a)')   "   diffusion_type =  ", "linear Poisson with manufactured soln"
  elseif (diffusion_type==10) then
  write(*,'(2a)')   "   diffusion_type =  ", "nonlinear Poisson with manufactured soln"
  elseif (diffusion_type==2) then
  write(*,'(2a)')   "   diffusion_type =  ", "your choice (currently, Laplace: div(grad(u))=0)"
  else
  write(*,*) " Invalid input: diffusion_type = ", diffusion_type
  write(*,*) "  Try 1 (linear Poisson), 2 (custom), or 10 (nonlinear), and try again... Stop "
  stop
  endif

 !------------------------------------------------------------------------------------
 ! If Dirichlet BC is specified by a user: 

  if (bc_dirichlet_user_values) then
  write(*,'(1a)')      "               BC =  user specified Dirichlet BC  "
  endif

 !------------------------------------------------------------------------------------
 ! If Neumann BC is specified by a user: 

  if (bc_neumann_user_values) then
  write(*,'(1a)')      "               BC =  user specified Neumann BC  "
  endif

 !------------------------------------------------------------------------------------

  write(*,*)
 !------------------------------------------------------------------------------------

    call cpu_time( time_begin )

! Initialization

           i_iteration = 0
  lrelax_sweeps_actual = 0
            lrelax_roc = 0.0_p2
 n_residual_evaluation = 0

       gcr_proj_actual = 0
       gcr_rkr0        = 1.0_p2
              walltime = 0.0_p2

! Print the heading on screen.

   if ( trim(solver_type) == "jfnk" ) then

    write(*,'(a77)') "|   GCR solver"
    write(*,'(2a)') " Iteration    Poisson      x-grad       y-grad       z-grad", &
                   "    |   proj     reduction   prec-sweeps"

   elseif ( trim(solver_type) == "implicit" ) then

    write(*,'(a80)') "|   GS relaxation"
    write(*,'(2a)') " Iteration    Poisson      x-grad       y-grad       z-grad", &
                   "    |   sweeps   reduction"

   else

    write(*,*)
    write(*,'(2a)') " Iteration   continuity   x-momemtum   y-momentum   z-momentum"

   endif

!---------------------------------------------------------------------------------------------
!---------------------------------------------------------------------------------------------
! Begin nonlinear iteration.
!---------------------------------------------------------------------------------------------
!---------------------------------------------------------------------------------------------

   iteration_exit = .false.

  nonlinear_iteration : do

  !-----------------------------------------------------------------------
  ! Begin with the residual computation, and check if the iteration converged.

    call compute_residual
    call compute_residual_norm(res_norm)

  !-----------------------------------------------------------------------
  ! Define the "initial" norm by which the residual norm is normalized.

    if (i_iteration == 0 ) then !<- No iteration has been performed yet.

     ! Save off the initial residual norm.
       res_norm_initial = res_norm

    elseif (i_iteration == 1 ) then !<- This is after 1 iteration.

     ! Save off the initial residual norm. Let's normalize the residual norm by this one
     ! if this is larger than the initial at iteration 0.

       do ii = 1, 4
        do jj = 1, 3
         if ( res_norm(ii,jj) > res_norm_initial(ii,jj) ) res_norm_initial(ii,jj) = res_norm(ii,jj)
        end do
       end do

    endif

  !-----------------------------------------------------------------------
  ! Print the current residual norms on screen.

    if ( trim(solver_type) == "implicit" ) then

     write(*,'(i10,4es13.3,a,i6,es12.1,i6)') i_iteration, res_norm(:,L1),  " | ",   &
                                              lrelax_sweeps_actual, lrelax_roc

    elseif ( trim(solver_type) == "jfnk" ) then

     write(*,'(i10,4es13.3,a,i6,es12.1,i6)') i_iteration, res_norm(:,L1),  " | ",   &
                                              gcr_proj_actual, gcr_rkr0, lrelax_sweeps_actual
    else

     write(*,'(i10,4es13.3)') i_iteration, res_norm(:,L1)

    endif

  !-----------------------------------------------------------------------
  ! Check convergence and exit if converged.

    if (i_iteration > 1 ) then

     ! Exit if the residual is reduced by a user-specified order.
      if ( maxval( res_norm(:,1)/res_norm_initial(:,1) ) < solver_tolerance) then

        iteration_exit = .true.

        write(*,*)
        write(*,*) " Converged. Res reduction achieved = ", &
                   maxval( res_norm(:,1)/res_norm_initial(:,1) )
        write(*,*)

      endif

     ! Update the reference residual norm if increased.
     ! res_norm_initial(:,1) = max( res_norm_initial(:,1), res_norm(:,1) )

    endif

  !-----------------------------------------------------------------------
  ! Exit if the iteration reaches a user-specified maximum number.

    if (i_iteration == solver_max_itr) then
     
       write(*,*)
       write(*,*) " Max iteration reached...  "
       write(*,*)

       iteration_exit = .true.

    endif

  !-----------------------------------------------------------------------
  ! Write the residual history.

      call cpu_time( time_end )
      walltime = time_end - time_begin
         wtime = time_end - time_begin

      call write_residual_hist_tec(i_iteration,res_norm,res_norm_initial, &
      lrelax_sweeps_actual,gcr_proj_actual, filename_tecplot_hist,iteration_exit,walltime )

  !-----------------------------------------------------------------------
  ! Exit iteration if converged or max is reached.

    if (iteration_exit) then
     solver_itr_actual = i_iteration
     exit nonlinear_iteration
    endif

  !-----------------------------------------------------------------------
  ! OK, not finished. Let's perform one iteration.

  !----------------------------------------------------------
  ! First, compute the pseudo time steps at all nodes.

    call compute_local_time_step_dtau

  !----------------------------------------------------------
  ! Perform one iteration: u <- u + du

     solver_method : if ( trim(solver_type) == "fwd_euler") then

      ! Forward Euler explicit iteration
        call explicit_pseudo_time_forward_euler

     elseif ( trim(solver_type) == "tvdrk2" ) then

      ! 2-stage TVD RK explicit iteration
        call explicit_pseudo_time_tvdrk2
 
     elseif ( trim(solver_type) == "implicit" ) then

      ! Implicit Defect-Correction solver
        call implicit(i_iteration)

     elseif ( trim(solver_type) == "jfnk" ) then

      ! Jacobian-Free Newton-Krylov with GCR
        call jfnk(i_iteration)

     endif solver_method
  !----------------------------------------------------------

   i_iteration = i_iteration + 1

  end do nonlinear_iteration

!---------------------------------------------------------------------------------------------
!---------------------------------------------------------------------------------------------
! Enbd of nonlinear iteration.
!---------------------------------------------------------------------------------------------
!---------------------------------------------------------------------------------------------

  write(*,*)
  write(*,*) "                      Total iteration = ", i_iteration
  write(*,*) " Total number of residual evaluations = ", n_residual_evaluation
  write(*,*) "                       Total CPU time = ", walltime
  write(*,*)

 end subroutine solve_steady


!********************************************************************************
!
! This subroutine computes the residual norms: L1, L2, L_infty
!
!********************************************************************************
 subroutine compute_residual_norm(res_norm)

 use data_module, only : nnodes, res

 implicit none

 integer , parameter ::   p2 = selected_real_kind(15) ! Double precision
 real(p2), parameter :: zero = 0.0_p2
 real(p2), parameter ::  one = 1.0_p2

 real(p2), dimension(4,3), intent(out) :: res_norm

!Local variables
 real(p2), dimension(4) :: residual
 integer                :: i

   res_norm(:,1) =  zero
   res_norm(:,2) =  zero
   res_norm(:,3) = - one


  loop_nodes : do i = 1, nnodes

   residual = abs( res(i,:) )

   res_norm(:,1) = res_norm(:,1)    + residual    !L1   norm
   res_norm(:,2) = res_norm(:,2)    + residual**2 !L2   norm
   res_norm(:,3) = max(res_norm(:,3), residual)   !Linf norm

  end do loop_nodes


   res_norm(:,1) =      res_norm(:,1)/real(nnodes,p2)  !L1 norm
   res_norm(:,2) = sqrt(res_norm(:,2)/real(nnodes,p2)) !L2 norm

 end subroutine compute_residual_norm
!********************************************************************************

!********************************************************************************
!
! This subroutine computes the local pseudo-time step at each node.
!
!********************************************************************************
 subroutine compute_local_time_step_dtau

 use data_module    , only : nnodes, dtau, wsn, vol
 use input_parameter, only : CFL

 implicit none

 integer , parameter ::   p2 = selected_real_kind(P=15)
 real(p2), parameter :: half = 0.5_p2
 integer             :: i


  loop_nodes : do i = 1, nnodes

  ! Local time step: dtau = CFL*volume/sum(0.5*max_wave_speed*face_area).

   dtau(i) = CFL*vol(i) / ( half*wsn(i) )

  end do loop_nodes


 end subroutine compute_local_time_step_dtau
!********************************************************************************


!********************************************************************************
!
! Perform 1 nonlinear iteration by the forward-Euler pseudo-time stepping.
!
!   u^{n+1} = u^n - (dt/vol)*Res(u^n)
!
! Note: This is very slow to converge to steady state.
!
!*******************************************************************************
 subroutine explicit_pseudo_time_forward_euler

 use data_module, only : nnodes, u, du, dtau, res, vol

 implicit none

 integer :: i


  loop_nodes : do i = 1, nnodes

  !This is the correction suggested by the forward Euler scheme:

   du(i,:) = - dtau(i)/vol(i) * res(i,:)

  !Update the solution by applying the correction.

    u(i,:) = u(i,:) + du(i,:)

  end do loop_nodes


 end subroutine explicit_pseudo_time_forward_euler
!********************************************************************************

!********************************************************************************
!
! Perform 1 nonlinear iteration by an explicit 2-stage RK pseudo-time stepping.
!
!  Two-stage Runge-Kutta scheme: u^n is saved as u0(:,:)
!   1. u^*     = u^n - (dt/vol)*Res(u^n)
!   2. u^{n+1} = 1/2*u^n + 1/2*[u^* - (dt/vol)*Res(u^*)]
!
! Note: This is very slow to converge to steady state.
!
!*******************************************************************************
 subroutine explicit_pseudo_time_tvdrk2

 use data_module, only : nnodes, u0, u, du, dtau, res, vol
 use residual   , only : compute_residual

 implicit none

 integer , parameter :: p2 = selected_real_kind(P=15)
 real(p2), parameter :: half = 0.5_p2
 integer             :: i

 !-----------------------------
 !- 1st stage of Runge-Kutta:
 !-----------------------------

   u0 = u

  loop_nodes : do i = 1, nnodes

  !This is the correction suggested by the forward Euler scheme:

   du(i,:) = - dtau(i)/vol(i) * res(i,:)

  !Update the solution by applying the correction.

    u(i,:) = u(i,:) + du(i,:)

  end do loop_nodes

 !-----------------------------
 !- 2nd stage of Runge-Kutta:
 !-----------------------------

  call compute_residual

  u = half*( u + u0 )

  loop_nodes2 : do i = 1, nnodes

  !This is the correction suggested by the forward Euler scheme:

   du(i,:) = - half*dtau(i)/vol(i) * res(i,:)

  !Update the solution by applying the correction.

    u(i,:) = u(i,:) + du(i,:)

  end do loop_nodes2

 end subroutine explicit_pseudo_time_tvdrk2
!********************************************************************************

!********************************************************************************
!
! Perform 1 nonlinear iteration by the implicit defect-correction solver:
!
!   U^{n+1} = U^n + dU,  dRes/dU*dU = - Res(U^n),
!
! where 
! 
!
! (1)Compute the residual Jacobian:
!
!    - Exact derivative of 1st-order residual (no LSQ gradients).
!    - This is Newton's method for 1st-order scheme (with a large CFL).
!
! (2)Solve the linear system: (V/dtau+dRes1/du)*du = -Res, where dRes1/du is
!    the exact derivative of the 1st-order residual (no LSQ gradients).
!
!    - Gauss-Seidel relaxation is implemented.
!    - Max relaxations and tolerance are specified by input parameters.
!
! (3)Update the solution: u <- u + du.
!
!********************************************************************************
 subroutine implicit(itr)

 use data_module    , only : nnodes, u, du
 use jacobian       , only : compute_jacobian
 use linear_solver  , only : linear_relaxation_gen
 use input_parameter, only : nonlinear_omega, diffusion_type

 use data_module    , only : k0, nnghbrs, nghbr, color, ncolors, nghbrs_total
 use data_module    , only : res, jac_off, jac_dinv
 use data_module    , only : lrelax_sweeps_actual, lrelax_roc

 implicit none

 integer, intent(in) :: itr

 integer  :: i

  !Compute the residual Jacobian

   if (itr == 0)                          call compute_jacobian !Compute Jacobian at the beginning.
   if (itr >  0 .and. diffusion_type/=1 ) call compute_jacobian !Recompute for nonlinear Poisson

  !Compute du by relaxing (approximately solving) the linear system.

    call linear_relaxation_gen(nnodes,4,nghbrs_total,jac_off, jac_dinv, &
                                res,k0, nnghbrs, nghbr, color, ncolors, &
                                     du,lrelax_sweeps_actual,lrelax_roc )

  !Update the solution by applying the correction with a safety factor.

   loop_nodes : do i = 1, nnodes

    !Update the conservative variables
     u(i,:) = u(i,:) + nonlinear_omega*du(i,:)

   end do loop_nodes


 end subroutine implicit
!********************************************************************************


!********************************************************************************
!
! Perform 1 nonlinear iteration by the implicit defect-correction solver.
!
! (1)Compute the residual Jacobian:
!
!    - Exact derivative of 1st-order residual.
!    - This is used as a preconditioner matrix, Jac.
!
! (2)Solve the exact linearized system: (V/dtau+dRes/du)*du = -Res:
!
!    - Frechet derivative is used to evaluate dRes/dU*du.
!    - Gauss-Seidel relaxation is implemented to approximately invert Jac.
!    - Max relaxations and tolerance are specified by input parameters.
!    - Max GCR projection and tolerance are specified by input parameters.
!
! (3)Update the solution: u <- u + du.
!
! Note: This is Newton's method for both 1st- and 2nd-order schemes if
!       CFL is sufficiently large.
!
! Note: Jacobian-Free means that the exact Jacobian dRes/du is not stored.
!       But it requires a preconditioner matrix, which is given by
!       the exact derivative of the 1st-order scheme.
!
!********************************************************************************
 subroutine jfnk(itr)

 use input_parameter, only : gcr_projection !Max GCR projection
 use input_parameter, only : gcr_mu         !GCR tolerance (reduction)
 use input_parameter, only : nonlinear_omega, diffusion_type

 use data_module    , only : gcr_proj_actual      !Actual # of projections
 use data_module    , only : lrelax_sweeps_actual !Last GS relaxations
 use data_module    , only : gcr_rkr0             !Actual GCR-res reduction
 use data_module    , only : status_pc            !0=success in preconditioner
 use data_module    , only : status_gcr           !0=success in GCR

 use data_module    , only : nnodes, u, du, i_iteration
 use jacobian       , only : compute_jacobian
 use gcr            , only : gcr_solver

 implicit none

 integer, intent(in) :: itr

 integer  :: i

  !Compute the residual Jacobian used as a preconditioner matrix.
  !This matrix is unchanged during the GCR solve.

   if (itr == 0)                          call compute_jacobian !Compute Jacobian at the beginning.
   if (itr >  0 .and. diffusion_type/=1 ) call compute_jacobian !Recompute for nonlinear Poisson

  !Compute du by JFNK-GCR solver by solving the exactly linearized system.

   call gcr_solver(gcr_projection, gcr_mu, gcr_proj_actual, lrelax_sweeps_actual, &
                                      gcr_rkr0, status_pc, status_gcr,i_iteration+1 )

  !Update the solution by applying the correction with a safety factor.

   loop_nodes : do i = 1, nnodes

    !Update the conservative variables
     u(i,:) = u(i,:) + nonlinear_omega*du(i,:)

   end do loop_nodes


 end subroutine jfnk
!********************************************************************************

!********************************************************************************
!* Write a Tecplot file for convergence history.
!*
!********************************************************************************
 subroutine write_residual_hist_tec(i_iteration,res_norm,res_norm_initial, &
                                 sweeps, gcr_proj, datafile_hist,last,time )

 use data_module, only : p2

 implicit none

 integer                      , intent(in) :: i_iteration
 real(kind=p2), dimension(4,3), intent(in) :: res_norm
 real(kind=p2), dimension(4,3), intent(in) :: res_norm_initial
 integer                      , intent(in) :: sweeps, gcr_proj
 character(80)                , intent(in) :: datafile_hist
 logical                      , intent(in) :: last
 real(kind=p2),                 intent(in) :: time

 integer       :: k, i, tmp_itr, tmp_sweeps, tmp_gcr
 real(kind=p2) :: tmp1(4), tmp1_scaled(4), tmp_time
 real(kind=p2) :: tmp2(4), tmp2_scaled(4)
 real(kind=p2) :: tmpi(4), tmpi_scaled(4)

   if (i_iteration == 0) then
    open(unit=110, file = "delete_me",  status="unknown")
   else
    open(unit=110, file = "delete_me", position='append')
   endif

   write(110,'(i13,es27.15,i13, 24es27.15,i13)')                           &
             i_iteration,   time,  sweeps                                , &
             (res_norm(k,1),k=1,4)                                       , &
             (res_norm(k,1)/( res_norm_initial(k,1) + 1.0e-25_p2), k=1,4), &
             (res_norm(k,2),k=1,4)                                       , &
             (res_norm(k,2)/( res_norm_initial(k,2) + 1.0e-25_p2), k=1,4), &
             (res_norm(k,3),k=1,4)                                       , &
             (res_norm(k,3)/( res_norm_initial(k,3) + 1.0e-25_p2), k=1,4), &
             gcr_proj

   if (last) then

    open(unit=120, file = datafile_hist, status="unknown")

    write(120,*) "title = ",'"Convergence History from datafile_history"'
    write(120,*) "variables = ",'"Iteration","CPU time","Sweep",', &
    '"R1_L1" , "R2_L1" , "R3_L1" , "R4_L1", "Rc1_L1", "Rc2_L1", "Rc3_L1", "Rc4_L1",', &
    '"R1_L2" , "R2_L2" , "R3_L2" , "R4_L2", "Rc1_L2", "Rc2_L2", "Rc3_L2", "Rc4_L2",', &
    '"R1_Li" , "R2_Li" , "R3_Li" , "R4_Li", "Rc1_Li", "Rc2_Li", "Rc3_Li", "Rc4_Li",', &
    '"maxRc_L1","maxRc_L2","maxRc_Li","Krylov vector"'
    write(120,*) "zone", " i = ",i_iteration+1, "f = point"

    rewind(110)

    do i = 1, i_iteration+1

     read(110,*) tmp_itr, tmp_time, tmp_sweeps           , &
                (tmp1(k), k=1,4), (tmp1_scaled(k), k=1,4), &
                (tmp2(k), k=1,4), (tmp2_scaled(k), k=1,4), &
                (tmpi(k), k=1,4), (tmpi_scaled(k), k=1,4), &
                tmp_gcr

     write(120,'(i13,es27.15,i13, 24es27.15, 3es27.15,i13)') &
                 tmp_itr, tmp_time, tmp_sweeps           ,   &
                (tmp1(k), k=1,4), (tmp1_scaled(k), k=1,4),   &
                (tmp2(k), k=1,4), (tmp2_scaled(k), k=1,4),   &
                (tmpi(k), k=1,4), (tmpi_scaled(k), k=1,4),   &
                maxval( tmp1_scaled ), &
                maxval( tmp2_scaled ), &
                maxval( tmpi_scaled ), &
                tmp_gcr
    end do

    close(120)

   endif

  close(110)

 end subroutine write_residual_hist_tec
!--------------------------------------------------------------------------------

 end module steady_solver
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------

